@@ -1,6 +1,6 @@
 /*
  * BlueALSA - a2dp-aac.c
- * Copyright (c) 2016-2022 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -10,10 +10,7 @@
 
 #include "a2dp-aac.h"
 
-#if ENABLE_AAC
-
 #include <errno.h>
-#include <endian.h>
 #include <pthread.h>
 #include <stdbool.h>
 #include <stddef.h>
@@ -63,7 +60,7 @@ struct a2dp_codec a2dp_aac_sink = {
 	.dir = A2DP_SINK,
 	.codec_id = A2DP_CODEC_MPEG24,
 	.capabilities.aac = {
-		/* NOTE: AAC Long Term Prediction and AAC Scalable are
+		/* NOTE: AAC Long Term Prediction and AAC Scalable might be
 		 *       not supported by the FDK-AAC library. */
 		.object_type =
 			AAC_OBJECT_TYPE_MPEG2_AAC_LC |
@@ -99,7 +96,7 @@ struct a2dp_codec a2dp_aac_source = {
 	.dir = A2DP_SOURCE,
 	.codec_id = A2DP_CODEC_MPEG24,
 	.capabilities.aac = {
-		/* NOTE: AAC Long Term Prediction and AAC Scalable are
+		/* NOTE: AAC Long Term Prediction and AAC Scalable might be
 		 *       not supported by the FDK-AAC library. */
 		.object_type =
 			AAC_OBJECT_TYPE_MPEG2_AAC_LC |
@@ -133,6 +130,21 @@ struct a2dp_codec a2dp_aac_source = {
 
 void a2dp_aac_init(void) {
 
+	LIB_INFO info[16] = { 0 };
+	info[ARRAYSIZE(info) - 1].module_id = ~FDK_NONE;
+
+	aacDecoder_GetLibInfo(info);
+	aacEncGetLibInfo(info);
+
+	unsigned int caps_dec = FDKlibInfo_getCapabilities(info, FDK_AACDEC);
+	unsigned int caps_enc = FDKlibInfo_getCapabilities(info, FDK_AACENC);
+	debug("FDK-AAC lib capabilities: dec:%#x enc:%#x", caps_dec, caps_enc);
+
+	if (caps_dec & CAPF_ER_AAC_SCAL)
+		a2dp_aac_sink.capabilities.aac.object_type |= AAC_OBJECT_TYPE_MPEG4_AAC_SCA;
+	if (caps_enc & CAPF_ER_AAC_SCAL)
+		a2dp_aac_source.capabilities.aac.object_type |= AAC_OBJECT_TYPE_MPEG4_AAC_SCA;
+
 	if (config.a2dp.force_mono)
 		a2dp_aac_source.capabilities.aac.channels = AAC_CHANNELS_1;
 	if (config.a2dp.force_44100)
@@ -173,12 +185,13 @@ static unsigned int a2dp_aac_get_fdk_vbr_mode(
 	return 5;
 }
 
-static void *a2dp_aac_enc_thread(struct ba_transport_thread *th) {
+void *a2dp_aac_enc_thread(struct ba_transport_thread *th) {
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
 
 	struct ba_transport *t = th->t;
+	struct ba_transport_pcm *t_pcm = th->pcm;
 	struct io_poll io = { .timeout = -1 };
 
 	HANDLE_AACENCODER handle;
@@ -187,8 +200,8 @@ static void *a2dp_aac_enc_thread(struct ba_transport_thread *th) {
 
 	const a2dp_aac_t *configuration = &t->a2dp.configuration.aac;
 	const unsigned int bitrate = AAC_GET_BITRATE(*configuration);
-	const unsigned int channels = t->a2dp.pcm.channels;
-	const unsigned int samplerate = t->a2dp.pcm.sampling;
+	const unsigned int channels = t_pcm->channels;
+	const unsigned int samplerate = t_pcm->sampling;
 
 	/* create AAC encoder without the Meta Data module */
 	if ((err = aacEncOpen(&handle, 0x07, channels)) != AACENC_OK) {
@@ -284,7 +297,7 @@ static void *a2dp_aac_enc_thread(struct ba_transport_thread *th) {
 	pthread_cleanup_push(PTHREAD_CLEANUP(ffb_free), &pcm);
 
 	const unsigned int aac_frame_size = aacinf.inputChannels * aacinf.frameLength;
-	const size_t sample_size = BA_TRANSPORT_PCM_FORMAT_BYTES(t->a2dp.pcm.format);
+	const size_t sample_size = BA_TRANSPORT_PCM_FORMAT_BYTES(t_pcm->format);
 	if (ffb_init(&pcm, aac_frame_size, sample_size) == -1 ||
 			ffb_init_uint8_t(&bt, RTP_HEADER_LEN + aacinf.maxOutBufBytes) == -1) {
 		error("Couldn't create data buffers: %s", strerror(errno));
@@ -324,13 +337,22 @@ static void *a2dp_aac_enc_thread(struct ba_transport_thread *th) {
 	AACENC_OutArgs out_args = { 0 };
 
 	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_set_state_running(th);;) {
+	for (ba_transport_thread_state_set_running(th);;) {
 
-		ssize_t samples;
-		if ((samples = io_poll_and_read_pcm(&io, &t->a2dp.pcm,
-						pcm.tail, ffb_len_in(&pcm))) <= 0) {
-			if (samples == -1)
-				error("PCM poll and read error: %s", strerror(errno));
+		ssize_t samples = ffb_len_in(&pcm);
+		switch (samples = io_poll_and_read_pcm(&io, t_pcm, pcm.tail, samples)) {
+		case -1:
+			if (errno == ESTALE) {
+				in_args.numInSamples = -1;
+				/* flush encoder internal buffers */
+				while (aacEncEncode(handle, NULL, &out_buf, &in_args, &out_args) == AACENC_OK)
+					continue;
+				ffb_rewind(&pcm);
+				continue;
+			}
+			error("PCM poll and read error: %s", strerror(errno));
+			/* fall-through */
+		case 0:
 			ba_transport_stop_if_no_clients(t);
 			continue;
 		}
@@ -389,7 +411,7 @@ static void *a2dp_aac_enc_thread(struct ba_transport_thread *th) {
 			rtp_state_update(&rtp, pcm_frames);
 
 			/* update busy delay (encoding overhead) */
-			t->a2dp.pcm.delay = asrsync_get_busy_usec(&io.asrs) / 100;
+			t_pcm->delay = asrsync_get_busy_usec(&io.asrs) / 100;
 
 			/* If the input buffer was not consumed, we have to append new data to
 			 * the existing one. Since we do not use ring buffer, we will simply
@@ -402,8 +424,6 @@ static void *a2dp_aac_enc_thread(struct ba_transport_thread *th) {
 
 fail:
 	debug_transport_thread_loop(th, "EXIT");
-	ba_transport_thread_set_state_stopping(th);
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -414,12 +434,14 @@ fail_open:
 	return NULL;
 }
 
-static void *a2dp_aac_dec_thread(struct ba_transport_thread *th) {
+__attribute__ ((weak))
+void *a2dp_aac_dec_thread(struct ba_transport_thread *th) {
 
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
 
 	struct ba_transport *t = th->t;
+	struct ba_transport_pcm *t_pcm = th->pcm;
 	struct io_poll io = { .timeout = -1 };
 
 	HANDLE_AACDECODER handle;
@@ -432,8 +454,8 @@ static void *a2dp_aac_dec_thread(struct ba_transport_thread *th) {
 
 	pthread_cleanup_push(PTHREAD_CLEANUP(aacDecoder_Close), handle);
 
-	const unsigned int channels = t->a2dp.pcm.channels;
-	const unsigned int samplerate = t->a2dp.pcm.sampling;
+	const unsigned int channels = t_pcm->channels;
+	const unsigned int samplerate = t_pcm->sampling;
 
 #ifdef AACDECODER_LIB_VL0
 	if ((err = aacDecoder_SetParam(handle, AAC_PCM_MIN_OUTPUT_CHANNELS, channels)) != AAC_DEC_OK) {
@@ -472,7 +494,7 @@ static void *a2dp_aac_dec_thread(struct ba_transport_thread *th) {
 	int markbit_quirk = -3;
 
 	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_set_state_running(th);;) {
+	for (ba_transport_thread_state_set_running(th);;) {
 
 		ssize_t len = ffb_blen_in(&bt);
 		if ((len = io_poll_and_read_bt(&io, th, bt.data, len)) <= 0) {
@@ -489,7 +511,7 @@ static void *a2dp_aac_dec_thread(struct ba_transport_thread *th) {
 		int missing_rtp_frames = 0;
 		rtp_state_sync_stream(&rtp, rtp_header, &missing_rtp_frames, NULL);
 
-		if (!ba_transport_pcm_is_active(&t->a2dp.pcm)) {
+		if (!ba_transport_pcm_is_active(t_pcm)) {
 			rtp.synced = false;
 			continue;
 		}
@@ -540,8 +562,8 @@ static void *a2dp_aac_dec_thread(struct ba_transport_thread *th) {
 				warn("AAC channels mismatch: %u != %u", aacinf->numChannels, channels);
 
 			const size_t samples = (size_t)aacinf->frameSize * channels;
-			io_pcm_scale(&t->a2dp.pcm, pcm.data, samples);
-			if (io_pcm_write(&t->a2dp.pcm, pcm.data, samples) == -1)
+			io_pcm_scale(t_pcm, pcm.data, samples);
+			if (io_pcm_write(t_pcm, pcm.data, samples) == -1)
 				error("FIFO write error: %s", strerror(errno));
 
 			/* update local state with decoded PCM frames */
@@ -556,8 +578,6 @@ static void *a2dp_aac_dec_thread(struct ba_transport_thread *th) {
 
 fail:
 	debug_transport_thread_loop(th, "EXIT");
-	ba_transport_thread_set_state_stopping(th);
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -571,14 +591,12 @@ fail_open:
 
 int a2dp_aac_transport_start(struct ba_transport *t) {
 
-	if (t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE)
+	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE)
 		return ba_transport_thread_create(&t->thread_enc, a2dp_aac_enc_thread, "ba-a2dp-aac", true);
 
-	if (t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SINK)
+	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SINK)
 		return ba_transport_thread_create(&t->thread_dec, a2dp_aac_dec_thread, "ba-a2dp-aac", true);
 
 	g_assert_not_reached();
 	return -1;
 }
-
-#endif
