@@ -1,6 +1,6 @@
 /*
  * BlueALSA - a2dp-lc3plus.c
- * Copyright (c) 2016-2022 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -9,15 +9,14 @@
  */
 
 #include "a2dp-lc3plus.h"
+/* IWYU pragma: no_include "config.h" */
 
-#if ENABLE_LC3PLUS
-
-#include <endian.h>
 #include <errno.h>
 #include <pthread.h>
 #include <stdbool.h>
 #include <stddef.h>
 #include <stdint.h>
+#include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 
@@ -28,7 +27,6 @@
 #include "a2dp.h"
 #include "audio.h"
 #include "bluealsa-config.h"
-#include "codec-sbc.h"
 #include "io.h"
 #include "rtp.h"
 #include "utils.h"
@@ -169,7 +167,7 @@ static int a2dp_lc3plus_get_frame_dms(const a2dp_lc3plus_t *conf) {
 	}
 }
 
-static void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
+void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 
 	/* Cancellation should be possible only in the carefully selected place
 	 * in order to prevent memory leaks and resources not being released. */
@@ -177,12 +175,13 @@ static void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
 
 	struct ba_transport *t = th->t;
+	struct ba_transport_pcm *t_pcm = th->pcm;
 	struct io_poll io = { .timeout = -1 };
 
 	const a2dp_lc3plus_t *configuration = &t->a2dp.configuration.lc3plus;
 	const int lc3plus_frame_dms = a2dp_lc3plus_get_frame_dms(configuration);
-	const unsigned int channels = t->a2dp.pcm.channels;
-	const unsigned int samplerate = t->a2dp.pcm.sampling;
+	const unsigned int channels = t_pcm->channels;
+	const unsigned int samplerate = t_pcm->sampling;
 	const unsigned int rtp_ts_clockrate = 96000;
 
 	/* check whether library supports selected configuration */
@@ -236,6 +235,7 @@ static void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 
 	int32_t *pcm_ch1 = malloc(lc3plus_ch_samples * sizeof(int32_t));
 	int32_t *pcm_ch2 = malloc(lc3plus_ch_samples * sizeof(int32_t));
+	int32_t *pcm_ch_buffers[2] = { pcm_ch1, pcm_ch2 };
 	pthread_cleanup_push(PTHREAD_CLEANUP(free), pcm_ch1);
 	pthread_cleanup_push(PTHREAD_CLEANUP(free), pcm_ch2);
 
@@ -257,13 +257,23 @@ static void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 	rtp_state_init(&rtp, samplerate, rtp_ts_clockrate);
 
 	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_set_state_running(th);;) {
+	for (ba_transport_thread_state_set_running(th);;) {
 
-		ssize_t samples;
-		if ((samples = io_poll_and_read_pcm(&io, &t->a2dp.pcm,
-						pcm.tail, ffb_len_in(&pcm))) <= 0) {
-			if (samples == -1)
-				error("PCM poll and read error: %s", strerror(errno));
+		ssize_t samples = ffb_len_in(&pcm);
+		switch (samples = io_poll_and_read_pcm(&io, t_pcm, pcm.tail, samples)) {
+		case -1:
+			if (errno == ESTALE) {
+				int encoded = 0;
+				memset(pcm_ch1, 0, lc3plus_ch_samples * sizeof(*pcm_ch1));
+				memset(pcm_ch2, 0, lc3plus_ch_samples * sizeof(*pcm_ch2));
+				/* flush encoder internal buffers by feeding it with silence */
+				lc3plus_enc24(handle, pcm_ch_buffers, rtp_payload, &encoded);
+				ffb_rewind(&pcm);
+				continue;
+			}
+			error("PCM poll and read error: %s", strerror(errno));
+			/* fall-through */
+		case 0:
 			ba_transport_stop_if_no_clients(t);
 			continue;
 		}
@@ -289,9 +299,8 @@ static void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 				lc3plus_frames < ((1 << 4) - 1)) {
 
 			int encoded = 0;
-			int32_t *in_buffers[2] = { pcm_ch1, pcm_ch2 };
 			audio_deinterleave_s24_4le(input, lc3plus_ch_samples, channels, pcm_ch1, pcm_ch2);
-			if ((err = lc3plus_enc24(handle, in_buffers, bt.tail, &encoded)) != LC3PLUS_OK) {
+			if ((err = lc3plus_enc24(handle, pcm_ch_buffers, bt.tail, &encoded)) != LC3PLUS_OK) {
 				error("LC3plus encoding error: %s", lc3plus_strerror(err));
 				break;
 			}
@@ -361,7 +370,7 @@ static void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 			rtp_state_update(&rtp, pcm_frames);
 
 			/* update busy delay (encoding overhead) */
-			t->a2dp.pcm.delay = asrsync_get_busy_usec(&io.asrs) / 100;
+			t_pcm->delay = asrsync_get_busy_usec(&io.asrs) / 100;
 
 			/* If the input buffer was not consumed (due to codesize limit), we
 			 * have to append new data to the existing one. Since we do not use
@@ -375,8 +384,6 @@ static void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th) {
 
 fail:
 	debug_transport_thread_loop(th, "EXIT");
-	ba_transport_thread_set_state_stopping(th);
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -389,7 +396,8 @@ fail_init:
 	return NULL;
 }
 
-static void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
+__attribute__ ((weak))
+void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 
 	/* Cancellation should be possible only in the carefully selected place
 	 * in order to prevent memory leaks and resources not being released. */
@@ -397,11 +405,12 @@ static void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
 
 	struct ba_transport *t = th->t;
+	struct ba_transport_pcm *t_pcm = th->pcm;
 	struct io_poll io = { .timeout = -1 };
 
 	const a2dp_lc3plus_t *configuration = &t->a2dp.configuration.lc3plus;
-	const unsigned int channels = t->a2dp.pcm.channels;
-	const unsigned int samplerate = t->a2dp.pcm.sampling;
+	const unsigned int channels = t_pcm->channels;
+	const unsigned int samplerate = t_pcm->sampling;
 	const unsigned int rtp_ts_clockrate = 96000;
 
 	/* check whether library supports selected configuration */
@@ -440,6 +449,7 @@ static void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 
 	int32_t *pcm_ch1 = malloc(lc3plus_ch_samples * sizeof(int32_t));
 	int32_t *pcm_ch2 = malloc(lc3plus_ch_samples * sizeof(int32_t));
+	int32_t *pcm_ch_buffers[2] = { pcm_ch1, pcm_ch2 };
 	pthread_cleanup_push(PTHREAD_CLEANUP(free), pcm_ch1);
 	pthread_cleanup_push(PTHREAD_CLEANUP(free), pcm_ch2);
 
@@ -460,7 +470,7 @@ static void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 	bool rtp_media_fragment_skip = false;
 
 	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_set_state_running(th);;) {
+	for (ba_transport_thread_state_set_running(th);;) {
 
 		ssize_t len = ffb_blen_in(&bt);
 		if ((len = io_poll_and_read_bt(&io, th, bt.data, len)) <= 0) {
@@ -488,7 +498,7 @@ static void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 			ffb_rewind(&bt_payload);
 		}
 
-		if (!ba_transport_pcm_is_active(&t->a2dp.pcm)) {
+		if (!ba_transport_pcm_is_active(t_pcm)) {
 			rtp.synced = false;
 			continue;
 		}
@@ -502,15 +512,14 @@ static void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 
 		while (missing_pcm_frames > 0) {
 
-			int32_t *out_buffers[2] = { pcm_ch1, pcm_ch2 };
-			lc3plus_dec24(handle, bt_payload.data, 0, out_buffers, 1);
+			lc3plus_dec24(handle, bt_payload.data, 0, pcm_ch_buffers, 1);
 			audio_interleave_s24_4le(pcm_ch1, pcm_ch2, lc3plus_ch_samples, channels, pcm.data);
 
 			warn("Missing LC3plus data, loss concealment applied");
 
 			const size_t samples = lc3plus_frame_samples;
-			io_pcm_scale(&t->a2dp.pcm, pcm.data, samples);
-			if (io_pcm_write(&t->a2dp.pcm, pcm.data, samples) == -1)
+			io_pcm_scale(t_pcm, pcm.data, samples);
+			if (io_pcm_write(t_pcm, pcm.data, samples) == -1)
 				error("FIFO write error: %s", strerror(errno));
 
 			missing_pcm_frames -= lc3plus_ch_samples;
@@ -557,8 +566,7 @@ static void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 		/* Decode retrieved LC3plus frames. */
 		while (lc3plus_frames--) {
 
-			int32_t *out_buffers[2] = { pcm_ch1, pcm_ch2 };
-			err = lc3plus_dec24(handle, lc3plus_payload, lc3plus_frame_len, out_buffers, 0);
+			err = lc3plus_dec24(handle, lc3plus_payload, lc3plus_frame_len, pcm_ch_buffers, 0);
 			audio_interleave_s24_4le(pcm_ch1, pcm_ch2, lc3plus_ch_samples, channels, pcm.data);
 
 			if (err == LC3PLUS_DECODE_ERROR)
@@ -571,8 +579,8 @@ static void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 			lc3plus_payload += lc3plus_frame_len;
 
 			const size_t samples = lc3plus_frame_samples;
-			io_pcm_scale(&t->a2dp.pcm, pcm.data, samples);
-			if (io_pcm_write(&t->a2dp.pcm, pcm.data, samples) == -1)
+			io_pcm_scale(t_pcm, pcm.data, samples);
+			if (io_pcm_write(t_pcm, pcm.data, samples) == -1)
 				error("FIFO write error: %s", strerror(errno));
 
 			/* update local state with decoded PCM frames */
@@ -587,8 +595,6 @@ static void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th) {
 
 fail:
 	debug_transport_thread_loop(th, "EXIT");
-	ba_transport_thread_set_state_stopping(th);
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -604,14 +610,12 @@ fail_init:
 
 int a2dp_lc3plus_transport_start(struct ba_transport *t) {
 
-	if (t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE)
+	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE)
 		return ba_transport_thread_create(&t->thread_enc, a2dp_lc3plus_enc_thread, "ba-a2dp-lc3p", true);
 
-	if (t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SINK)
+	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SINK)
 		return ba_transport_thread_create(&t->thread_dec, a2dp_lc3plus_dec_thread, "ba-a2dp-lc3p", true);
 
 	g_assert_not_reached();
 	return -1;
 }
-
-#endif
