@@ -1,6 +1,6 @@
 /*
  * bluealsa-pcm.c
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2022 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -12,11 +12,13 @@
 # include <config.h>
 #endif
 
+#include <alloca.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <poll.h>
 #include <pthread.h>
 #include <signal.h>
+#include <stdatomic.h>
 #include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -25,7 +27,7 @@
 #include <strings.h>
 #include <sys/eventfd.h>
 #include <sys/ioctl.h>
-#include <time.h>
+#include <sys/time.h>
 #include <unistd.h>
 
 #include <alsa/asoundlib.h>
@@ -43,9 +45,9 @@
 #define BA_PAUSE_STATE_PAUSED  (1 << 0)
 #define BA_PAUSE_STATE_PENDING (1 << 1)
 
-#if SND_LIB_VERSION >= 0x010104 && SND_LIB_VERSION <= 0x010205
+#if SND_LIB_VERSION >= 0x010104
 /**
- * alsa-lib releases 1.1.4 to 1.2.5.1 inclusive have a bug in the rate plugin
+ * alsa-lib releases from 1.1.4 have a bug in the rate plugin
  * which, when combined with the hw params refinement algorithm used by the
  * ioplug, can cause snd_pcm_avail() to return bogus values. This, in turn,
  * can trigger deadlock in applications built on the portaudio library
@@ -69,23 +71,28 @@ struct bluealsa_pcm {
 
 	/* requested BlueALSA PCM */
 	struct ba_pcm ba_pcm;
-	size_t ba_pcm_buffer_size;
+
+	/* PCM FIFO */
 	int ba_pcm_fd;
+	/* PCM control socket */
 	int ba_pcm_ctrl_fd;
 
+	/* Indicates that the server is connected. */
+	atomic_bool connected;
+
 	/* event file descriptor */
 	int event_fd;
 
 	/* virtual hardware - ring buffer */
-	char *io_hw_buffer;
+	char * _Atomic io_hw_buffer;
 	/* The IO thread is responsible for maintaining the hardware pointer
 	 * (pcm->io_hw_ptr), the application is responsible for the application
-	 * pointer (io->appl_ptr). These are both volatile as they are both
+	 * pointer (io->appl_ptr). These pointers should be atomic as they are
 	 * written in one thread and read in the other. */
-	volatile snd_pcm_sframes_t io_hw_ptr;
-	snd_pcm_uframes_t io_hw_boundary;
+	_Atomic snd_pcm_sframes_t io_hw_ptr;
+	_Atomic snd_pcm_uframes_t io_hw_boundary;
 	/* Permit the application to modify the frequency of poll() events. */
-	volatile snd_pcm_uframes_t io_avail_min;
+	_Atomic snd_pcm_uframes_t io_avail_min;
 	pthread_t io_thread;
 	bool io_started;
 
@@ -96,7 +103,7 @@ struct bluealsa_pcm {
 	snd_pcm_uframes_t delay_hw_ptr;
 	unsigned int delay_pcm_nread;
 	/* In the capture mode, delay_running indicates that frames are being
-	 * transfered to the FIFO by the server. In playback mode it indicates
+	 * transferred to the FIFO by the server. In playback mode it indicates
 	 * that the IO thread is transferring frames to the FIFO. */
 	bool delay_running;
 
@@ -134,32 +141,29 @@ static snd_pcm_uframes_t snd_pcm_ioplug_hw_avail(const snd_pcm_ioplug_t * const
 		diff = io->buffer_size - hw_ptr + appl_ptr;
 	if (diff < 0)
 		diff += pcm->io_hw_boundary;
-	return diff <= io->buffer_size ? (snd_pcm_uframes_t) diff : 0;
+	snd_pcm_uframes_t diff_ = diff;
+	return diff_ <= io->buffer_size ? diff_ : 0;
 }
 #endif
 
 /**
- * Helper function for closing PCM transport. */
-static int close_transport(struct bluealsa_pcm *pcm) {
-	int rv = 0;
-	pthread_mutex_lock(&pcm->mutex);
-	if (pcm->ba_pcm_fd != -1) {
-		rv |= close(pcm->ba_pcm_fd);
-		pcm->ba_pcm_fd = -1;
-	}
-	if (pcm->ba_pcm_ctrl_fd != -1) {
-		rv |= close(pcm->ba_pcm_ctrl_fd);
-		pcm->ba_pcm_ctrl_fd = -1;
-	}
-	pthread_mutex_unlock(&pcm->mutex);
-	return rv;
+ * Helper function for terminating IO thread. */
+static void io_thread_cancel(struct bluealsa_pcm *pcm) {
+
+	if (!pcm->io_started)
+		return;
+
+	pthread_cancel(pcm->io_thread);
+	pthread_join(pcm->io_thread, NULL);
+	pcm->io_started = false;
+
 }
 
 /**
- * Helper function for IO thread termination. */
+ * Helper function for logging IO thread termination. */
 static void io_thread_cleanup(struct bluealsa_pcm *pcm) {
 	debug2("IO thread cleanup");
-	pcm->io_started = false;
+	(void)pcm;
 }
 
 /**
@@ -201,16 +205,15 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 	pthread_cleanup_push(PTHREAD_CLEANUP(io_thread_cleanup), pcm);
 
 	sigset_t sigset;
-	sigemptyset(&sigset);
-
-	/* Block signal, which will be used for pause/resume actions. */
-	sigaddset(&sigset, SIGIO);
-	/* Block SIGPIPE, so we could receive EPIPE while writing to the pipe
-	 * whose reading end has been closed. This will allow clean playback
-	 * termination. */
-	sigaddset(&sigset, SIGPIPE);
-
-	if ((errno = pthread_sigmask(SIG_BLOCK, &sigset, NULL)) != 0) {
+	/* Block all signals in the IO thread.
+	 * Especially, we need to block SIGPIPE, so we could receive EPIPE while
+	 * writing to the pipe which reading end was closed by the server. This
+	 * will allow clean playback termination. Also, we need to block SIGIO,
+	 * which is used for pause/resume actions. The rest of the signals are
+	 * blocked because we are using thread cancellation and we do not want
+	 * any interference from signal handlers. */
+	sigfillset(&sigset);
+	if ((errno = pthread_sigmask(SIG_SETMASK, &sigset, NULL)) != 0) {
 		SNDERR("Thread signal mask error: %s", strerror(errno));
 		goto fail;
 	}
@@ -226,14 +229,18 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 	debug2("Starting IO loop: %d", pcm->ba_pcm_fd);
 	for (;;) {
 
-		if (pcm->pause_state & BA_PAUSE_STATE_PENDING ||
+		pthread_mutex_lock(&pcm->mutex);
+		unsigned int is_pause_pending = pcm->pause_state & BA_PAUSE_STATE_PENDING;
+		pthread_mutex_unlock(&pcm->mutex);
+
+		if (is_pause_pending ||
 				pcm->io_hw_ptr == -1) {
 			debug2("Pausing IO thread");
 
 			pthread_mutex_lock(&pcm->mutex);
 			pcm->pause_state = BA_PAUSE_STATE_PAUSED;
-			pthread_cond_signal(&pcm->pause_cond);
 			pthread_mutex_unlock(&pcm->mutex);
+			pthread_cond_signal(&pcm->pause_cond);
 
 			int tmp;
 			sigwait(&sigset, &tmp);
@@ -246,8 +253,6 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 
 			if (pcm->io_hw_ptr == -1)
 				continue;
-			if (pcm->ba_pcm_fd == -1)
-				goto fail;
 
 			asrsync_init(&asrs, io->rate);
 			io_hw_ptr = pcm->io_hw_ptr;
@@ -277,7 +282,7 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 		/* When used with the rate plugin the buffer might contain a fractional
 		 * number of periods. So if the leftover in the buffer is less than a
 		 * whole period size, adjust the number of frames which should be
-		 * transfered.  */
+		 * transferred.  */
 		if (io->buffer_size - offset < frames)
 			frames = io->buffer_size - offset;
 
@@ -300,14 +305,17 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 					if (errno == EINTR)
 						continue;
 					SNDERR("PCM FIFO read error: %s", strerror(errno));
+					pcm->connected = false;
 					goto fail;
 				}
 				head += ret;
 				len -= ret;
 			}
 
-			if (ret == 0)
+			if (ret == 0) {
+				pcm->connected = false;
 				goto fail;
+			}
 
 			io_thread_update_delay(pcm, io_hw_ptr);
 
@@ -321,6 +329,7 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 						continue;
 					if (errno != EPIPE)
 						SNDERR("PCM FIFO write error: %s", strerror(errno));
+					pcm->connected = false;
 					goto fail;
 				}
 				head += ret;
@@ -343,11 +352,20 @@ static void *io_thread(snd_pcm_ioplug_t *io) {
 	}
 
 fail:
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
-	pthread_cleanup_pop(1);
-	close_transport(pcm);
-	eventfd_write(pcm->event_fd, 0xDEAD0000);
+
+	/* make sure we will not get stuck in the pause sync loop */
+	pthread_mutex_lock(&pcm->mutex);
+	pcm->pause_state = BA_PAUSE_STATE_PAUSED;
+	pthread_mutex_unlock(&pcm->mutex);
 	pthread_cond_signal(&pcm->pause_cond);
+
+	eventfd_write(pcm->event_fd, 0xDEAD0000);
+
+	/* wait for cancellation from main thread */
+	while (true)
+		sleep(3600);
+
+	pthread_cleanup_pop(1);
 	return NULL;
 }
 
@@ -377,7 +395,7 @@ static int bluealsa_start(snd_pcm_ioplug_t *io) {
 	/* start the IO thread */
 	pcm->io_started = true;
 	if ((errno = pthread_create(&pcm->io_thread, NULL,
-					PTHREAD_ROUTINE(io_thread), io)) != 0) {
+					PTHREAD_FUNC(io_thread), io)) != 0) {
 		debug2("Couldn't create IO thread: %s", strerror(errno));
 		pcm->io_started = false;
 		return -errno;
@@ -391,11 +409,7 @@ static int bluealsa_stop(snd_pcm_ioplug_t *io) {
 	struct bluealsa_pcm *pcm = io->private_data;
 	debug2("Stopping");
 
-	if (pcm->io_started) {
-		pcm->io_started = false;
-		pthread_cancel(pcm->io_thread);
-		pthread_join(pcm->io_thread, NULL);
-	}
+	io_thread_cancel(pcm);
 
 	pcm->delay_running = false;
 	pcm->delay_pcm_nread = 0;
@@ -421,14 +435,14 @@ static snd_pcm_sframes_t bluealsa_pointer(snd_pcm_ioplug_t *io) {
 	/* Any error returned here is translated to -EPIPE, SND_PCM_STATE_XRUN,
 	 * by ioplug; and that prevents snd_pcm_readi() and snd_pcm_writei()
 	 * from returning -ENODEV to the application on device disconnection.
-`	 * Instead, when the device is disconnected, we update the PCM state
+	 * Instead, when the device is disconnected, we update the PCM state
 	 * directly here but we do not return an error code. This ensures that
 	 * ioplug does not undo that state change. Both snd_pcm_readi() and
 	 * snd_pcm_writei() return -ENODEV when the PCM state is
 	 * SND_PCM_STATE_DISCONNECTED after their internal call to
 	 * snd_pcm_avail_update(), which will be the case when we set it here.
 	 */
-	if (pcm->ba_pcm_fd == -1)
+	if (!pcm->connected)
 		snd_pcm_ioplug_set_state(io, SND_PCM_STATE_DISCONNECTED);
 
 #ifndef SND_PCM_IOPLUG_FLAG_BOUNDARY_WA
@@ -520,7 +534,7 @@ static int bluealsa_fix_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *par
 	if ((ret = snd_pcm_hw_params_set_buffer_size(io->pcm, refined_params, buffer_size)) < 0)
 		return ret;
 
-	memcpy(params, refined_params, snd_pcm_hw_params_sizeof());
+	snd_pcm_hw_params_copy(params, refined_params);
 
 	return ret;
 }
@@ -528,12 +542,14 @@ static int bluealsa_fix_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *par
 
 static int bluealsa_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params) {
 	struct bluealsa_pcm *pcm = io->private_data;
-	(void)params;
+
 	debug2("Initializing HW");
 
 #if	BLUEALSA_HW_PARAMS_FIX
 	if (bluealsa_fix_hw_params(io, params) < 0)
 		debug2("Warning - unable to fix incorrect buffer size in hw parameters");
+#endif
+
 	snd_pcm_uframes_t period_size;
 	int ret;
 	if ((ret = snd_pcm_hw_params_get_period_size(params, &period_size, 0)) < 0)
@@ -541,7 +557,6 @@ static int bluealsa_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params)
 	snd_pcm_uframes_t buffer_size;
 	if ((ret = snd_pcm_hw_params_get_buffer_size(params, &buffer_size)) < 0)
 		return ret;
-#endif
 
 	pcm->frame_size = (snd_pcm_format_physical_width(io->format) * io->channels) / 8;
 
@@ -553,6 +568,8 @@ static int bluealsa_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params)
 		return -EBUSY;
 	}
 
+	pcm->connected = true;
+
 	if (pcm->io.stream == SND_PCM_STREAM_PLAYBACK)
 		/* By default, the size of the pipe buffer is set to a too large value for
 		 * our purpose. On modern Linux system it is 65536 bytes. Large buffer in
@@ -566,7 +583,6 @@ static int bluealsa_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params)
 
 	debug2("FIFO buffer size: %zd frames", pcm->delay_fifo_size);
 
-#if BLUEALSA_HW_PARAMS_FIX
 	/* ALSA default for avail min is one period. */
 	pcm->io_avail_min = period_size;
 
@@ -574,15 +590,6 @@ static int bluealsa_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params)
 			buffer_size / period_size, pcm->frame_size * period_size,
 			period_size * (buffer_size / period_size) == buffer_size ? '=' : '<',
 			buffer_size * pcm->frame_size);
-#else
-	/* ALSA default for avail min is one period. */
-	pcm->io_avail_min = io->period_size;
-
-	debug2("Selected HW buffer: %zd periods x %zd bytes %c= %zd bytes",
-			io->buffer_size / io->period_size, pcm->frame_size * io->period_size,
-			io->period_size * (io->buffer_size / io->period_size) == io->buffer_size ? '=' : '<',
-			io->buffer_size * pcm->frame_size);
-#endif
 
 	return 0;
 }
@@ -590,16 +597,31 @@ static int bluealsa_hw_params(snd_pcm_ioplug_t *io, snd_pcm_hw_params_t *params)
 static int bluealsa_hw_free(snd_pcm_ioplug_t *io) {
 	struct bluealsa_pcm *pcm = io->private_data;
 	debug2("Freeing HW");
-	if (close_transport(pcm) == -1)
-		return -errno;
-	return 0;
+
+	/* Before closing PCM transport make sure that
+	 * the IO thread is terminated. */
+	io_thread_cancel(pcm);
+
+	int rv = 0;
+	if (pcm->ba_pcm_fd != -1)
+		rv |= close(pcm->ba_pcm_fd);
+	if (pcm->ba_pcm_ctrl_fd != -1)
+		rv |= close(pcm->ba_pcm_ctrl_fd);
+
+	pcm->ba_pcm_fd = -1;
+	pcm->ba_pcm_ctrl_fd = -1;
+	pcm->connected = false;
+
+	return rv == 0 ? 0 : -errno;
 }
 
 static int bluealsa_sw_params(snd_pcm_ioplug_t *io, snd_pcm_sw_params_t *params) {
 	struct bluealsa_pcm *pcm = io->private_data;
 	debug2("Initializing SW");
 
-	snd_pcm_sw_params_get_boundary(params, &pcm->io_hw_boundary);
+	snd_pcm_uframes_t boundary;
+	snd_pcm_sw_params_get_boundary(params, &boundary);
+	pcm->io_hw_boundary = boundary;
 
 	snd_pcm_uframes_t avail_min;
 	snd_pcm_sw_params_get_avail_min(params, &avail_min);
@@ -615,7 +637,7 @@ static int bluealsa_prepare(snd_pcm_ioplug_t *io) {
 	struct bluealsa_pcm *pcm = io->private_data;
 
 	/* if PCM FIFO is not opened, report it right away */
-	if (pcm->ba_pcm_fd == -1) {
+	if (!pcm->connected) {
 		snd_pcm_ioplug_set_state(io, SND_PCM_STATE_DISCONNECTED);
 		return -ENODEV;
 	}
@@ -642,7 +664,49 @@ static int bluealsa_prepare(snd_pcm_ioplug_t *io) {
 
 static int bluealsa_drain(snd_pcm_ioplug_t *io) {
 	struct bluealsa_pcm *pcm = io->private_data;
-	bluealsa_dbus_pcm_ctrl_send_drain(pcm->ba_pcm_ctrl_fd, NULL);
+	debug2("Draining");
+
+	if (!pcm->connected) {
+		snd_pcm_ioplug_set_state(io, SND_PCM_STATE_DISCONNECTED);
+		return -ENODEV;
+	}
+
+	/* A bug in the ioplug drain implementation means that snd_pcm_drain()
+	 * always either finishes in state SND_PCM_STATE_SETUP or returns an error.
+	 * It is not possible to finish in state SND_PCM_STATE_DRAINING and return
+	 * success; therefore is is impossible to correctly implement capture
+	 * drain logic. So for capture PCMs we do nothing and return success. */
+
+	if (io->stream == SND_PCM_STREAM_PLAYBACK) {
+
+		/* We must ensure that all remaining frames in the ring buffer are
+		 * flushed to the FIFO by the I/O thread. It is possible that the
+		 * client has called snd_pcm_drain() without the start_threshold
+		 * having been reached, or while paused, so we must first ensure that
+		 * the IO thread is running. */
+		if (bluealsa_start(io) < 0)
+			return 0;
+
+		/* Block until the drain is complete. */
+		struct pollfd pfd = { pcm->event_fd, POLLIN, 0 };
+		while (bluealsa_pointer(io) >= 0 && io->state == SND_PCM_STATE_DRAINING) {
+			if (poll(&pfd, 1, -1) == -1) {
+				if (errno == EINTR)
+					continue;
+				break;
+			}
+			if (pfd.revents & POLLIN) {
+				eventfd_t event;
+				eventfd_read(pcm->event_fd, &event);
+				if (event & 0xDEAD0000)
+					break;
+			}
+		}
+
+		bluealsa_dbus_pcm_ctrl_send_drain(pcm->ba_pcm_ctrl_fd, NULL);
+
+	}
+
 	/* We cannot recover from an error here. By returning zero we ensure that
 	 * ioplug stops the pcm. Returning an error code would be interpreted by
 	 * ioplug as an incomplete drain and would it leave the pcm running. */
@@ -662,10 +726,6 @@ static snd_pcm_sframes_t bluealsa_calculate_delay(snd_pcm_ioplug_t *io) {
 
 	snd_pcm_sframes_t delay = 0;
 
-	/* if PCM is not started there should be no capture delay */
-	if (!pcm->delay_running && io->stream == SND_PCM_STREAM_CAPTURE)
-		return 0;
-
 	struct timespec now;
 	gettimestamp(&now);
 
@@ -686,6 +746,12 @@ static snd_pcm_sframes_t bluealsa_calculate_delay(snd_pcm_ioplug_t *io) {
 
 	pthread_mutex_lock(&pcm->mutex);
 
+	/* if PCM is not started there should be no capture delay */
+	if (!pcm->delay_running && io->stream == SND_PCM_STREAM_CAPTURE) {
+		pthread_mutex_unlock(&pcm->mutex);
+		return 0;
+	}
+
 	struct timespec diff;
 	timespecsub(&now, &pcm->delay_ts, &diff);
 
@@ -757,11 +823,16 @@ static int bluealsa_pause(snd_pcm_ioplug_t *io, int enable) {
 		 * the server will not be paused while we are processing a transfer. */
 		pthread_mutex_lock(&pcm->mutex);
 		pcm->pause_state |= BA_PAUSE_STATE_PENDING;
-		while (!(pcm->pause_state & BA_PAUSE_STATE_PAUSED) && pcm->ba_pcm_fd != -1)
+		while (!(pcm->pause_state & BA_PAUSE_STATE_PAUSED) && pcm->connected)
 			pthread_cond_wait(&pcm->pause_cond, &pcm->mutex);
 		pthread_mutex_unlock(&pcm->mutex);
 	}
 
+	if (!pcm->connected) {
+		snd_pcm_ioplug_set_state(io, SND_PCM_STATE_DISCONNECTED);
+		return -ENODEV;
+	}
+
 	if (!bluealsa_dbus_pcm_ctrl_send(pcm->ba_pcm_ctrl_fd,
 				enable ? "Pause" : "Resume", NULL))
 		return -errno;
@@ -785,11 +856,11 @@ static void bluealsa_dump(snd_pcm_ioplug_t *io, snd_output_t *out) {
 	struct bluealsa_pcm *pcm = io->private_data;
 	snd_output_printf(out, "BlueALSA PCM: %s\n", pcm->ba_pcm.pcm_path);
 	snd_output_printf(out, "BlueALSA BlueZ device: %s\n", pcm->ba_pcm.device_path);
-	snd_output_printf(out, "BlueALSA Bluetooth codec: %s\n", pcm->ba_pcm.codec);
+	snd_output_printf(out, "BlueALSA Bluetooth codec: %s\n", pcm->ba_pcm.codec.name);
 	/* alsa-lib commits the PCM setup only if bluealsa_hw_params() returned
 	 * success, so we only dump the ALSA PCM parameters if the BlueALSA PCM
 	 * connection is established. */
-	if (pcm->ba_pcm_fd >= 0) {
+	if (pcm->connected) {
 		snd_output_printf(out, "Its setup is:\n");
 		snd_pcm_dump_setup(io->pcm, out);
 	}
@@ -798,7 +869,7 @@ static void bluealsa_dump(snd_pcm_ioplug_t *io, snd_output_t *out) {
 static int bluealsa_delay(snd_pcm_ioplug_t *io, snd_pcm_sframes_t *delayp) {
 	struct bluealsa_pcm *pcm = io->private_data;
 
-	if (pcm->ba_pcm_fd == -1) {
+	if (!pcm->connected) {
 		snd_pcm_ioplug_set_state(io, SND_PCM_STATE_DISCONNECTED);
 		return -ENODEV;
 	}
@@ -841,6 +912,9 @@ static int bluealsa_poll_descriptors(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 		unsigned int nfds) {
 	struct bluealsa_pcm *pcm = io->private_data;
 
+	if (nfds < 1)
+		return -EINVAL;
+
 	nfds_t dbus_nfds = nfds - 1;
 	if (!bluealsa_dbus_connection_poll_fds(&pcm->dbus_ctx, &pfd[1], &dbus_nfds))
 		return -EINVAL;
@@ -859,12 +933,15 @@ static int bluealsa_poll_revents(snd_pcm_ioplug_t *io, struct pollfd *pfd,
 	*revents = 0;
 	int ret = 0;
 
+	if (nfds < 1)
+		return -EINVAL;
+
 	bluealsa_dbus_connection_poll_dispatch(&pcm->dbus_ctx, &pfd[1], nfds - 1);
 	while (dbus_connection_dispatch(pcm->dbus_ctx.conn) == DBUS_DISPATCH_DATA_REMAINS)
 		continue;
 	gettimestamp(&pcm->dbus_dispatch_ts);
 
-	if (pcm->ba_pcm_fd == -1)
+	if (!pcm->connected)
 		goto fail;
 
 	if (pfd[0].revents & POLLIN) {
@@ -1335,17 +1412,6 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 	pcm->io.callback = &bluealsa_callback;
 	pcm->io.private_data = pcm;
 
-#if SND_LIB_VERSION >= 0x010102 && SND_LIB_VERSION <= 0x010103
-	/* ALSA library thread-safe API functionality does not play well with ALSA
-	 * IO-plug plug-ins. It causes deadlocks which often make our PCM plug-in
-	 * unusable. As a workaround we are going to disable this functionality. */
-	if (setenv("LIBASOUND_THREAD_SAFE", "0", 0) == -1)
-		SNDERR("Couldn't disable ALSA thread-safe API: %s", strerror(errno));
-#endif
-
-	if ((ret = snd_pcm_ioplug_create(&pcm->io, name, stream, mode)) < 0)
-		goto fail;
-
 	if (codec != NULL && codec[0] != '\0') {
 		if (bluealsa_select_pcm_codec(pcm, codec, &err)) {
 			/* Changing the codec may change the audio format, sampling rate and/or
@@ -1364,11 +1430,27 @@ SND_PCM_PLUGIN_DEFINE_FUNC(bluealsa) {
 		}
 	}
 
-	if ((ret = bluealsa_set_hw_constraint(pcm)) < 0) {
-		snd_pcm_ioplug_delete(&pcm->io);
-		return ret;
+	/* If the BT transport codec is not known (which means that PCM sampling
+	 * rate is also not know), we cannot construct useful constraints. */
+	if (pcm->ba_pcm.sampling == 0) {
+		ret = -EAGAIN;
+		goto fail;
 	}
 
+#if SND_LIB_VERSION >= 0x010102 && SND_LIB_VERSION <= 0x010103
+	/* ALSA library thread-safe API functionality does not play well with ALSA
+	 * IO-plug plug-ins. It causes deadlocks which often make our PCM plug-in
+	 * unusable. As a workaround we are going to disable this functionality. */
+	if (setenv("LIBASOUND_THREAD_SAFE", "0", 0) == -1)
+		SNDERR("Couldn't disable ALSA thread-safe API: %s", strerror(errno));
+#endif
+
+	if ((ret = snd_pcm_ioplug_create(&pcm->io, name, stream, mode)) < 0)
+		goto fail;
+
+	if ((ret = bluealsa_set_hw_constraint(pcm)) < 0)
+		goto fail;
+
 	if (!bluealsa_update_pcm_softvol(pcm, pcm_softvol, &err)) {
 		SNDERR("Couldn't set BlueALSA PCM soft-volume: %s", err.message);
 		dbus_error_free(&err);
@@ -1387,6 +1469,8 @@ fail:
 	dbus_error_free(&err);
 	if (pcm->event_fd != -1)
 		close(pcm->event_fd);
+	pthread_mutex_destroy(&pcm->mutex);
+	pthread_cond_destroy(&pcm->pause_cond);
 	free(pcm);
 	return ret;
 }
