@@ -1,6 +1,6 @@
 /*
  * BlueALSA - ba-rfcomm.h
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -17,6 +17,7 @@
 #endif
 
 #include <pthread.h>
+#include <stdatomic.h>
 #include <stdbool.h>
 #include <stdint.h>
 
@@ -39,6 +40,13 @@ enum ba_rfcomm_signal {
 	BA_RFCOMM_SIGNAL_UPDATE_VOLUME,
 };
 
+struct ba_rfcomm_hfp_codecs {
+	bool cvsd;
+#if ENABLE_MSBC
+	bool msbc;
+#endif
+};
+
 /**
  * Data associated with RFCOMM communication. */
 struct ba_rfcomm {
@@ -75,20 +83,23 @@ struct ba_rfcomm {
 	/* number of failed communication attempts */
 	int retries;
 
-	/* AG/HF supported features bitmask */
-	uint32_t hfp_features;
+	/* AG supported feature mask */
+	uint32_t ag_features;
+	/* HF supported feature mask */
+	uint32_t hf_features;
 
-	/* HF supported HFP codecs */
-	struct {
-		bool cvsd;
-#if ENABLE_MSBC
-		bool msbc;
-#endif
-	} codecs;
+	/* AG supported codecs */
+	struct ba_rfcomm_hfp_codecs ag_codecs;
+	/* HF supported codecs */
+	struct ba_rfcomm_hfp_codecs hf_codecs;
 
-	/* codec selection synchronization */
-	pthread_mutex_t codec_selection_completed_mtx;
-	pthread_cond_t codec_selection_completed;
+	/* HF supported codecs encoded for BAC command and BCS error response. */
+	char hf_bac_bcs_string[4];
+
+	/* Synchronization primitives for codec selection. The condition variable
+	 * shall be used with the codec_id mutex from the associated transport. */
+	pthread_cond_t codec_selection_cond;
+	bool codec_selection_done;
 
 	/* requested codec by the AG */
 	int codec;
@@ -116,7 +127,7 @@ struct ba_rfcomm {
 	 * remove all references, otherwise resources will not be freed. If this
 	 * quirk workaround is enabled, RFCOMM link lost will trigger SCO transport
 	 * destroy rather than a simple unreferencing. */
-	bool link_lost_quirk;
+	atomic_bool link_lost_quirk;
 
 };
 
