@@ -1,6 +1,6 @@
 /*
  * BlueALSA - ba-transport.h
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -20,44 +20,13 @@
 #include <stdbool.h>
 #include <stddef.h>
 #include <stdint.h>
+#include <time.h>
 
 #include "a2dp.h"
 #include "ba-device.h"
-#include "ba-rfcomm.h"
 #include "bluez.h"
 #include "shared/a2dp-codecs.h"
 
-#define BA_TRANSPORT_PROFILE_NONE        (0)
-#define BA_TRANSPORT_PROFILE_A2DP_SOURCE (1 << 0)
-#define BA_TRANSPORT_PROFILE_A2DP_SINK   (2 << 0)
-#define BA_TRANSPORT_PROFILE_HFP_HF      (1 << 2)
-#define BA_TRANSPORT_PROFILE_HFP_AG      (2 << 2)
-#define BA_TRANSPORT_PROFILE_HSP_HS      (1 << 4)
-#define BA_TRANSPORT_PROFILE_HSP_AG      (2 << 4)
-
-#define BA_TRANSPORT_PROFILE_MASK_A2DP \
-	(BA_TRANSPORT_PROFILE_A2DP_SOURCE | BA_TRANSPORT_PROFILE_A2DP_SINK)
-#define BA_TRANSPORT_PROFILE_MASK_HFP \
-	(BA_TRANSPORT_PROFILE_HFP_HF | BA_TRANSPORT_PROFILE_HFP_AG)
-#define BA_TRANSPORT_PROFILE_MASK_HSP \
-	(BA_TRANSPORT_PROFILE_HSP_HS | BA_TRANSPORT_PROFILE_HSP_AG)
-#define BA_TRANSPORT_PROFILE_MASK_SCO \
-	(BA_TRANSPORT_PROFILE_MASK_HFP | BA_TRANSPORT_PROFILE_MASK_HSP)
-#define BA_TRANSPORT_PROFILE_MASK_AG \
-	(BA_TRANSPORT_PROFILE_HSP_AG | BA_TRANSPORT_PROFILE_HFP_AG)
-#define BA_TRANSPORT_PROFILE_MASK_HF \
-	(BA_TRANSPORT_PROFILE_HSP_HS | BA_TRANSPORT_PROFILE_HFP_HF)
-
-/**
- * Selected profile and audio codec.
- *
- * For A2DP vendor codecs the upper byte of the codec field
- * contains the lowest byte of the vendor ID. */
-struct ba_transport_type {
-	uint16_t profile;
-	uint16_t codec;
-};
-
 enum ba_transport_pcm_mode {
 	/* PCM used for capturing audio */
 	BA_TRANSPORT_PCM_MODE_SOURCE,
@@ -91,8 +60,10 @@ struct ba_transport_pcm {
 	/* PCM stream operation mode */
 	enum ba_transport_pcm_mode mode;
 
-	/* PCM access guard */
+	/* guard PCM data updates */
 	pthread_mutex_t mutex;
+	/* updates notification */
+	pthread_cond_t cond;
 
 	/* FIFO file descriptor */
 	int fd;
@@ -111,12 +82,12 @@ struct ba_transport_pcm {
 	 * audio encoding or decoding and data transfer. */
 	unsigned int delay;
 
+	/* indicates whether FIFO buffer was synchronized */
+	bool synced;
+
 	/* internal software volume control */
 	bool soft_volume;
 
-	/* maximal possible Bluetooth volume */
-	unsigned int max_bt_volume;
-
 	/* Volume configuration for channel left [0] and right [1]. In case of
 	 * a monophonic sound, only the left [0] channel shall be used. */
 	struct ba_transport_pcm_volume {
@@ -125,14 +96,13 @@ struct ba_transport_pcm {
 		/* audio signal mute switches */
 		bool soft_mute;
 		bool hard_mute;
-		/* pre-calculated PCM scale factor based on decibel formula
+		/* calculated PCM scale factor based on decibel formula
 		 * pow(10, dB / 20); for muted channel it shall equal 0 */
 		double scale;
 	} volume[2];
 
-	/* data synchronization */
-	pthread_mutex_t synced_mtx;
-	pthread_cond_t synced;
+	/* new PCM client mutex */
+	pthread_mutex_t client_mtx;
 
 	/* exported PCM D-Bus API */
 	char *ba_dbus_path;
@@ -140,6 +110,9 @@ struct ba_transport_pcm {
 
 };
 
+int ba_transport_pcm_volume_level_to_range(int value, int max);
+int ba_transport_pcm_volume_range_to_level(int value, int max);
+
 void ba_transport_pcm_volume_set(
 		struct ba_transport_pcm_volume *volume,
 		const int *level,
@@ -147,10 +120,12 @@ void ba_transport_pcm_volume_set(
 		const bool *hard_mute);
 
 enum ba_transport_thread_state {
-	BA_TRANSPORT_THREAD_STATE_NONE,
+	BA_TRANSPORT_THREAD_STATE_IDLE,
 	BA_TRANSPORT_THREAD_STATE_STARTING,
 	BA_TRANSPORT_THREAD_STATE_RUNNING,
 	BA_TRANSPORT_THREAD_STATE_STOPPING,
+	BA_TRANSPORT_THREAD_STATE_JOINING,
+	BA_TRANSPORT_THREAD_STATE_TERMINATED,
 };
 
 enum ba_transport_thread_signal {
@@ -167,12 +142,14 @@ struct ba_transport_thread {
 
 	/* backward reference to transport */
 	struct ba_transport *t;
+	/* associated PCM */
+	struct ba_transport_pcm *pcm;
 
-	/* guard thread structure */
+	/* guard transport thread data updates */
 	pthread_mutex_t mutex;
+	/* state/id updates notification */
+	pthread_cond_t cond;
 
-	/* guard thread state */
-	pthread_mutex_t state_mtx;
 	/* current state of the thread */
 	enum ba_transport_thread_state state;
 
@@ -185,22 +162,40 @@ struct ba_transport_thread {
 	/* notification PIPE */
 	int pipe[2];
 
-	/* state/id changed notification */
-	pthread_cond_t changed;
-
 };
 
-int ba_transport_thread_set_state(
-		struct ba_transport_thread *th,
-		enum ba_transport_thread_state state,
-		bool force);
+/**
+ * Encoder/decoder transport thread IO function. */
+typedef void *(*ba_transport_thread_func)(struct ba_transport_thread *);
 
-#define ba_transport_thread_set_state_starting(th) \
-	ba_transport_thread_set_state(th, BA_TRANSPORT_THREAD_STATE_STARTING, false)
-#define ba_transport_thread_set_state_running(th) \
-	ba_transport_thread_set_state(th, BA_TRANSPORT_THREAD_STATE_RUNNING, false)
-#define ba_transport_thread_set_state_stopping(th) \
-	ba_transport_thread_set_state(th, BA_TRANSPORT_THREAD_STATE_STOPPING, false)
+int ba_transport_thread_state_set(
+		struct ba_transport_thread *th,
+		enum ba_transport_thread_state state);
+
+#define ba_transport_thread_state_set_idle(th) \
+	ba_transport_thread_state_set(th, BA_TRANSPORT_THREAD_STATE_IDLE)
+#define ba_transport_thread_state_set_running(th) \
+	ba_transport_thread_state_set(th, BA_TRANSPORT_THREAD_STATE_RUNNING)
+#define ba_transport_thread_state_set_stopping(th) \
+	ba_transport_thread_state_set(th, BA_TRANSPORT_THREAD_STATE_STOPPING)
+
+bool ba_transport_thread_state_check(
+		struct ba_transport_thread *th,
+		enum ba_transport_thread_state state);
+
+#define ba_transport_thread_state_check_running(th) \
+	ba_transport_thread_state_check(th, BA_TRANSPORT_THREAD_STATE_RUNNING)
+#define ba_transport_thread_state_check_terminated(th) \
+	ba_transport_thread_state_check(th, BA_TRANSPORT_THREAD_STATE_TERMINATED)
+
+int ba_transport_thread_state_wait(
+		struct ba_transport_thread *th,
+		enum ba_transport_thread_state state);
+
+#define ba_transport_thread_state_wait_running(th) \
+	ba_transport_thread_state_wait(th, BA_TRANSPORT_THREAD_STATE_RUNNING)
+#define ba_transport_thread_state_wait_terminated(th) \
+	ba_transport_thread_state_wait(th, BA_TRANSPORT_THREAD_STATE_TERMINATED)
 
 int ba_transport_thread_bt_acquire(
 		struct ba_transport_thread *th);
@@ -220,18 +215,47 @@ enum ba_transport_thread_manager_command {
 	BA_TRANSPORT_THREAD_MANAGER_CANCEL_IF_NO_CLIENTS,
 };
 
+enum ba_transport_profile {
+	BA_TRANSPORT_PROFILE_NONE        = 0,
+	BA_TRANSPORT_PROFILE_A2DP_SOURCE = (1 << 0),
+	BA_TRANSPORT_PROFILE_A2DP_SINK   = (2 << 0),
+	BA_TRANSPORT_PROFILE_HFP_HF      = (1 << 2),
+	BA_TRANSPORT_PROFILE_HFP_AG      = (2 << 2),
+	BA_TRANSPORT_PROFILE_HSP_HS      = (1 << 4),
+	BA_TRANSPORT_PROFILE_HSP_AG      = (2 << 4),
+};
+
+#define BA_TRANSPORT_PROFILE_MASK_A2DP \
+	(BA_TRANSPORT_PROFILE_A2DP_SOURCE | BA_TRANSPORT_PROFILE_A2DP_SINK)
+#define BA_TRANSPORT_PROFILE_MASK_HFP \
+	(BA_TRANSPORT_PROFILE_HFP_HF | BA_TRANSPORT_PROFILE_HFP_AG)
+#define BA_TRANSPORT_PROFILE_MASK_HSP \
+	(BA_TRANSPORT_PROFILE_HSP_HS | BA_TRANSPORT_PROFILE_HSP_AG)
+#define BA_TRANSPORT_PROFILE_MASK_SCO \
+	(BA_TRANSPORT_PROFILE_MASK_HFP | BA_TRANSPORT_PROFILE_MASK_HSP)
+#define BA_TRANSPORT_PROFILE_MASK_AG \
+	(BA_TRANSPORT_PROFILE_HSP_AG | BA_TRANSPORT_PROFILE_HFP_AG)
+#define BA_TRANSPORT_PROFILE_MASK_HF \
+	(BA_TRANSPORT_PROFILE_HSP_HS | BA_TRANSPORT_PROFILE_HFP_HF)
+
 struct ba_transport {
 
 	/* backward reference to device */
 	struct ba_device *d;
 
-	/* guard modifications of transport type, e.g. codec */
-	pthread_mutex_t type_mtx;
-
 	/* Transport structure covers all transports supported by BlueALSA. However,
 	 * every transport requires specific handling - link acquisition, transport
 	 * specific configuration, freeing resources, etc. */
-	struct ba_transport_type type;
+	enum ba_transport_profile profile;
+
+	/* guard modifications of transport codec */
+	pthread_mutex_t codec_id_mtx;
+	/* For A2DP vendor codecs the upper byte of the codec field
+	 * contains the lowest byte of the vendor ID. */
+	uint16_t codec_id;
+
+	/* synchronization for codec selection */
+	pthread_mutex_t codec_select_client_mtx;
 
 	/* data for D-Bus management */
 	char *bluez_dbus_owner;
@@ -241,6 +265,10 @@ struct ba_transport {
 	 * and the IO threads stopping flag */
 	pthread_mutex_t bt_fd_mtx;
 
+	/* Ensure BT file descriptor acquisition procedure
+	 * is completed atomically. */
+	pthread_mutex_t acquisition_mtx;
+
 	/* This field stores a file descriptor (socket) associated with the BlueZ
 	 * side of the transport. The role of this socket depends on the transport
 	 * type - it can be either A2DP or SCO link. */
@@ -297,12 +325,24 @@ struct ba_transport {
 
 		struct {
 
-			/* associated RFCOMM thread */
+			/* Associated RFCOMM thread for SCO transport handled by local
+			 * HSP/HFP implementation. Otherwise, this field is set to NULL. */
 			struct ba_rfcomm *rfcomm;
 
+#if ENABLE_OFONO
+			/* Associated oFono card and modem paths. In case when SCO transport
+			 * is not oFono-based, these fields are set to NULL. */
+			char *ofono_dbus_path_card;
+			char *ofono_dbus_path_modem;
+#endif
+
 			/* Speaker and microphone signals should to be exposed as
 			 * a separate PCM devices. Hence, there is a requirement
-			 * for separate configurations. */
+			 * for separate configurations.
+			 *
+			 * NOTE: The speaker/microphone notation always refers to the whole
+			 *       AG/HS setup. For AG the speaker is an outgoing audio stream,
+			 *       while for HS the speaker is an incoming audio stream. */
 			struct ba_transport_pcm spk_pcm;
 			struct ba_transport_pcm mic_pcm;
 
@@ -324,18 +364,23 @@ struct ba_transport {
 
 struct ba_transport *ba_transport_new_a2dp(
 		struct ba_device *device,
-		struct ba_transport_type type,
+		enum ba_transport_profile profile,
 		const char *dbus_owner,
 		const char *dbus_path,
 		const struct a2dp_codec *codec,
 		const void *configuration);
 struct ba_transport *ba_transport_new_sco(
 		struct ba_device *device,
-		struct ba_transport_type type,
+		enum ba_transport_profile profile,
 		const char *dbus_owner,
 		const char *dbus_path,
 		int rfcomm_fd);
 
+#if DEBUG
+const char *ba_transport_debug_name(
+		const struct ba_transport *t);
+#endif
+
 struct ba_transport *ba_transport_lookup(
 		struct ba_device *device,
 		const char *dbus_path);
@@ -348,9 +393,6 @@ void ba_transport_unref(struct ba_transport *t);
 struct ba_transport_pcm *ba_transport_pcm_ref(struct ba_transport_pcm *pcm);
 void ba_transport_pcm_unref(struct ba_transport_pcm *pcm);
 
-int ba_transport_pcms_lock(struct ba_transport *t);
-int ba_transport_pcms_unlock(struct ba_transport *t);
-
 int ba_transport_select_codec_a2dp(
 		struct ba_transport *t,
 		const struct a2dp_sep *sep);
@@ -358,6 +400,8 @@ int ba_transport_select_codec_sco(
 		struct ba_transport *t,
 		uint16_t codec_id);
 
+uint16_t ba_transport_get_codec(
+		struct ba_transport *t);
 void ba_transport_set_codec(
 		struct ba_transport *t,
 		uint16_t codec_id);
@@ -379,13 +423,6 @@ bool ba_transport_pcm_is_active(
 int ba_transport_pcm_get_delay(
 		const struct ba_transport_pcm *pcm);
 
-unsigned int ba_transport_pcm_volume_level_to_bt(
-		const struct ba_transport_pcm *pcm,
-		int value);
-int ba_transport_pcm_volume_bt_to_level(
-		const struct ba_transport_pcm *pcm,
-		unsigned int value);
-
 int ba_transport_pcm_volume_update(
 		struct ba_transport_pcm *pcm);
 
@@ -398,13 +435,13 @@ int ba_transport_pcm_release(struct ba_transport_pcm *pcm);
 
 int ba_transport_thread_create(
 		struct ba_transport_thread *th,
-		void *(*routine)(struct ba_transport_thread *),
+		ba_transport_thread_func th_func,
 		const char *name,
 		bool master);
 
 void ba_transport_thread_cleanup(struct ba_transport_thread *th);
 
 #define debug_transport_thread_loop(th, tag) \
-	debug("IO loop: %s: %s: %s", tag, __func__, ba_transport_type_to_string((th)->t->type))
+	debug("IO loop: %s: %s: %s", tag, __func__, ba_transport_debug_name((th)->t))
 
 #endif
