@@ -1,6 +1,6 @@
 /*
  * BlueALSA - bluealsa-dbus.c
- * Copyright (c) 2016-2022 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -9,12 +9,15 @@
  */
 
 #include "bluealsa-dbus.h"
+/* IWYU pragma: no_include "config.h" */
 
 #include <errno.h>
 #include <fcntl.h>
 #include <pthread.h>
 #include <stdbool.h>
 #include <stdint.h>
+#include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <sys/socket.h>
 #include <unistd.h>
@@ -33,6 +36,7 @@
 #include "bluealsa-config.h"
 #include "bluealsa-iface.h"
 #include "bluealsa-skeleton.h"
+#include "bluez.h"
 #include "dbus.h"
 #include "hfp.h"
 #include "utils.h"
@@ -110,7 +114,8 @@ static GVariant *ba_variant_new_bluealsa_codecs(void) {
 	}
 
 	/* Expose A2DP codecs always in the same order. */
-	a2dp_codecs_qsort(a2dp_codecs_tmp, n);
+	qsort(a2dp_codecs_tmp, n, sizeof(*a2dp_codecs_tmp),
+			QSORT_COMPAR(a2dp_codec_ptr_cmp));
 
 	for (size_t i = 0; i < n; i++) {
 		const char *profile = a2dp_codecs_tmp[i]->dir == A2DP_SOURCE ?
@@ -175,24 +180,38 @@ GVariant *ba_variant_new_device_battery(const struct ba_device *d) {
 }
 
 static GVariant *ba_variant_new_transport_type(const struct ba_transport *t) {
-	if (t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE)
+	switch (t->profile) {
+	case BA_TRANSPORT_PROFILE_A2DP_SOURCE:
 		return g_variant_new_string(BLUEALSA_TRANSPORT_TYPE_A2DP_SOURCE);
-	if (t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SINK)
+	case BA_TRANSPORT_PROFILE_A2DP_SINK:
 		return g_variant_new_string(BLUEALSA_TRANSPORT_TYPE_A2DP_SINK);
-	if (t->type.profile & BA_TRANSPORT_PROFILE_HFP_AG)
+	case BA_TRANSPORT_PROFILE_HFP_AG:
 		return g_variant_new_string(BLUEALSA_TRANSPORT_TYPE_HFP_AG);
-	if (t->type.profile & BA_TRANSPORT_PROFILE_HFP_HF)
+	case BA_TRANSPORT_PROFILE_HFP_HF:
 		return g_variant_new_string(BLUEALSA_TRANSPORT_TYPE_HFP_HF);
-	if (t->type.profile & BA_TRANSPORT_PROFILE_HSP_AG)
+	case BA_TRANSPORT_PROFILE_HSP_AG:
 		return g_variant_new_string(BLUEALSA_TRANSPORT_TYPE_HSP_AG);
-	if (t->type.profile & BA_TRANSPORT_PROFILE_HSP_HS)
+	case BA_TRANSPORT_PROFILE_HSP_HS:
 		return g_variant_new_string(BLUEALSA_TRANSPORT_TYPE_HSP_HS);
-	warn("Unsupported transport type: %#x", t->type.profile);
-	return g_variant_new_string("<null>");
+	case BA_TRANSPORT_PROFILE_NONE:
+		break;
+	}
+	error("Unsupported transport type: %#x", t->profile);
+	g_assert_not_reached();
+	return NULL;
 }
 
 static GVariant *ba_variant_new_rfcomm_features(const struct ba_rfcomm *r) {
-	return g_variant_new_uint32(r->hfp_features);
+
+	const char *strv[32];
+	size_t n = 0;
+
+	if (r->sco->profile & BA_TRANSPORT_PROFILE_HFP_AG)
+		n = hfp_hf_features_to_strings(r->hf_features, strv, ARRAYSIZE(strv));
+	if (r->sco->profile & BA_TRANSPORT_PROFILE_HFP_HF)
+		n = hfp_ag_features_to_strings(r->ag_features, strv, ARRAYSIZE(strv));
+
+	return g_variant_new_strv(strv, n);
 }
 
 static GVariant *ba_variant_new_pcm_mode(const struct ba_transport_pcm *pcm) {
@@ -201,6 +220,10 @@ static GVariant *ba_variant_new_pcm_mode(const struct ba_transport_pcm *pcm) {
 	return g_variant_new_string(BLUEALSA_PCM_MODE_SINK);
 }
 
+static GVariant *ba_variant_new_pcm_running(const struct ba_transport_pcm *pcm) {
+	return g_variant_new_boolean(pcm->th->state == BA_TRANSPORT_THREAD_STATE_RUNNING);
+}
+
 static GVariant *ba_variant_new_pcm_format(const struct ba_transport_pcm *pcm) {
 	return g_variant_new_uint16(pcm->format);
 }
@@ -216,13 +239,21 @@ static GVariant *ba_variant_new_pcm_sampling(const struct ba_transport_pcm *pcm)
 static GVariant *ba_variant_new_pcm_codec(const struct ba_transport_pcm *pcm) {
 	const struct ba_transport *t = pcm->t;
 	const char *codec = NULL;
-	if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_A2DP)
-		codec = a2dp_codecs_codec_id_to_string(t->type.codec);
-	if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_SCO)
-		codec = hfp_codec_id_to_string(t->type.codec);
+	if (t->profile & BA_TRANSPORT_PROFILE_MASK_A2DP)
+		codec = a2dp_codecs_codec_id_to_string(t->codec_id);
+	if (t->profile & BA_TRANSPORT_PROFILE_MASK_SCO)
+		codec = hfp_codec_id_to_string(t->codec_id);
 	if (codec != NULL)
 		return g_variant_new_string(codec);
-	return g_variant_new_string("<null>");
+	return NULL;
+}
+
+static GVariant *ba_variant_new_pcm_codec_config(const struct ba_transport_pcm *pcm) {
+	const struct ba_transport *t = pcm->t;
+	if (t->profile & BA_TRANSPORT_PROFILE_MASK_A2DP)
+		return g_variant_new_fixed_array(G_VARIANT_TYPE_BYTE, &t->a2dp.configuration,
+				t->a2dp.codec->capabilities_size, sizeof(uint8_t));
+	return NULL;
 }
 
 static GVariant *ba_variant_new_pcm_delay(const struct ba_transport_pcm *pcm) {
@@ -238,26 +269,36 @@ static uint8_t ba_volume_pack_dbus_volume(bool muted, int value) {
 }
 
 static GVariant *ba_variant_new_pcm_volume(const struct ba_transport_pcm *pcm) {
+	const bool is_sco = pcm->t->profile & BA_TRANSPORT_PROFILE_MASK_SCO;
+	const int max = is_sco ? HFP_VOLUME_GAIN_MAX : BLUEZ_A2DP_VOLUME_MAX;
 	uint8_t ch1 = ba_volume_pack_dbus_volume(pcm->volume[0].scale == 0,
-			ba_transport_pcm_volume_level_to_bt(pcm, pcm->volume[0].level));
+			ba_transport_pcm_volume_level_to_range(pcm->volume[0].level, max));
 	uint8_t ch2 = ba_volume_pack_dbus_volume(pcm->volume[1].scale == 0,
-			ba_transport_pcm_volume_level_to_bt(pcm, pcm->volume[1].level));
+			ba_transport_pcm_volume_level_to_range(pcm->volume[1].level, max));
 	return g_variant_new_uint16((ch1 << 8) | (pcm->channels == 1 ? 0 : ch2));
 }
 
 static void ba_variant_populate_pcm(GVariantBuilder *props, const struct ba_transport_pcm *pcm) {
+
+	GVariant *value;
 	g_variant_builder_init(props, G_VARIANT_TYPE("a{sv}"));
+
 	g_variant_builder_add(props, "{sv}", "Device", ba_variant_new_device_path(pcm->t->d));
 	g_variant_builder_add(props, "{sv}", "Sequence", ba_variant_new_device_sequence(pcm->t->d));
 	g_variant_builder_add(props, "{sv}", "Transport", ba_variant_new_transport_type(pcm->t));
 	g_variant_builder_add(props, "{sv}", "Mode", ba_variant_new_pcm_mode(pcm));
+	g_variant_builder_add(props, "{sv}", "Running", ba_variant_new_pcm_running(pcm));
 	g_variant_builder_add(props, "{sv}", "Format", ba_variant_new_pcm_format(pcm));
 	g_variant_builder_add(props, "{sv}", "Channels", ba_variant_new_pcm_channels(pcm));
 	g_variant_builder_add(props, "{sv}", "Sampling", ba_variant_new_pcm_sampling(pcm));
-	g_variant_builder_add(props, "{sv}", "Codec", ba_variant_new_pcm_codec(pcm));
+	if ((value = ba_variant_new_pcm_codec(pcm)) != NULL)
+		g_variant_builder_add(props, "{sv}", "Codec", value);
+	if ((value = ba_variant_new_pcm_codec_config(pcm)) != NULL)
+		g_variant_builder_add(props, "{sv}", "CodecConfiguration", value);
 	g_variant_builder_add(props, "{sv}", "Delay", ba_variant_new_pcm_delay(pcm));
 	g_variant_builder_add(props, "{sv}", "SoftVolume", ba_variant_new_pcm_soft_volume(pcm));
 	g_variant_builder_add(props, "{sv}", "Volume", ba_variant_new_pcm_volume(pcm));
+
 }
 
 static bool ba_variant_populate_sep(GVariantBuilder *props, const struct a2dp_sep *sep) {
@@ -322,74 +363,6 @@ static bool ba_variant_populate_sep(GVariantBuilder *props, const struct a2dp_se
 	return true;
 }
 
-static void bluealsa_manager_get_pcms(GDBusMethodInvocation *inv, void *userdata) {
-	(void)userdata;
-
-	GVariantBuilder pcms;
-	g_variant_builder_init(&pcms, G_VARIANT_TYPE("a{oa{sv}}"));
-
-	struct ba_adapter *a;
-	size_t i;
-
-	for (i = 0; i < ARRAYSIZE(config.adapters); i++) {
-
-		if ((a = ba_adapter_lookup(i)) == NULL)
-			continue;
-
-		GHashTableIter iter_d, iter_t;
-		GVariantBuilder props;
-		struct ba_device *d;
-		struct ba_transport *t;
-
-		pthread_mutex_lock(&a->devices_mutex);
-		g_hash_table_iter_init(&iter_d, a->devices);
-		while (g_hash_table_iter_next(&iter_d, NULL, (gpointer)&d)) {
-
-			pthread_mutex_lock(&d->transports_mutex);
-			g_hash_table_iter_init(&iter_t, d->transports);
-			while (g_hash_table_iter_next(&iter_t, NULL, (gpointer)&t)) {
-
-				if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_A2DP) {
-
-					if (t->a2dp.pcm.ba_dbus_exported) {
-						ba_variant_populate_pcm(&props, &t->a2dp.pcm);
-						g_variant_builder_add(&pcms, "{oa{sv}}", t->a2dp.pcm.ba_dbus_path, &props);
-						g_variant_builder_clear(&props);
-					}
-
-					if (t->a2dp.pcm_bc.ba_dbus_exported) {
-						ba_variant_populate_pcm(&props, &t->a2dp.pcm_bc);
-						g_variant_builder_add(&pcms, "{oa{sv}}", t->a2dp.pcm_bc.ba_dbus_path, &props);
-						g_variant_builder_clear(&props);
-					}
-
-				}
-				else if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_SCO) {
-
-					ba_variant_populate_pcm(&props, &t->sco.spk_pcm);
-					g_variant_builder_add(&pcms, "{oa{sv}}", t->sco.spk_pcm.ba_dbus_path, &props);
-					g_variant_builder_clear(&props);
-
-					ba_variant_populate_pcm(&props, &t->sco.mic_pcm);
-					g_variant_builder_add(&pcms, "{oa{sv}}", t->sco.mic_pcm.ba_dbus_path, &props);
-					g_variant_builder_clear(&props);
-
-				}
-
-			}
-
-			pthread_mutex_unlock(&d->transports_mutex);
-		}
-
-		pthread_mutex_unlock(&a->devices_mutex);
-		ba_adapter_unref(a);
-
-	}
-
-	g_dbus_method_invocation_return_value(inv, g_variant_new("(a{oa{sv}})", &pcms));
-	g_variant_builder_clear(&pcms);
-}
-
 static GVariant *bluealsa_manager_get_property(const char *property,
 		GError **error, void *userdata) {
 	(void)error;
@@ -412,14 +385,7 @@ static GVariant *bluealsa_manager_get_property(const char *property,
  * Register BlueALSA D-Bus manager interfaces. */
 void bluealsa_dbus_register(void) {
 
-	static const GDBusMethodCallDispatcher dispatchers[] = {
-		{ .method = "GetPCMs",
-			.handler = bluealsa_manager_get_pcms },
-		{ 0 },
-	};
-
 	static const GDBusInterfaceSkeletonVTable vtable = {
-		.dispatchers = dispatchers,
 		.get_property = bluealsa_manager_get_property,
 	};
 
@@ -500,17 +466,21 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv, void *userdata) {
 
 	/* Prevent two (or more) clients trying to
 	 * open the same PCM at the same time. */
-	pthread_mutex_lock(&pcm->mutex);
+	pthread_mutex_lock(&pcm->client_mtx);
 
 	/* preliminary check whether HFP codes is selected */
-	if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_SCO &&
-			t->type.codec == HFP_CODEC_UNDEFINED) {
+	if (t->profile & BA_TRANSPORT_PROFILE_MASK_SCO &&
+			ba_transport_get_codec(t) == HFP_CODEC_UNDEFINED) {
 		g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
 				G_DBUS_ERROR_FAILED, "HFP audio codec not selected");
 		goto fail;
 	}
 
-	if (pcm->fd != -1) {
+	pthread_mutex_lock(&pcm->mutex);
+	const int pcm_fd = pcm->fd;
+	pthread_mutex_unlock(&pcm->mutex);
+
+	if (pcm_fd != -1) {
 		g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
 				G_DBUS_ERROR_FAILED, "%s", strerror(EBUSY));
 		goto fail;
@@ -536,60 +506,54 @@ static void bluealsa_pcm_open(GDBusMethodInvocation *inv, void *userdata) {
 	 * headset will not run voltage converter (power-on its circuit board) until
 	 * the transport is acquired in order to extend battery life. For profiles
 	 * like A2DP Sink and HFP headset, we will wait for incoming connection. */
-	if (t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE ||
-			t->type.profile & BA_TRANSPORT_PROFILE_MASK_AG) {
-
-		enum ba_transport_thread_state state;
+	if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE ||
+			t->profile & BA_TRANSPORT_PROFILE_MASK_AG) {
 
 		if (ba_transport_acquire(t) == -1) {
 			g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
-					G_DBUS_ERROR_FAILED, "Acquire transport: %s", strerror(errno));
+					G_DBUS_ERROR_IO_ERROR, "Acquire transport: %s", strerror(errno));
 			goto fail;
 		}
 
-		/* wait until ready to process audio */
-		pthread_mutex_lock(&th->state_mtx);
-		while ((state = th->state) < BA_TRANSPORT_THREAD_STATE_RUNNING)
-			pthread_cond_wait(&th->changed, &th->state_mtx);
-		pthread_mutex_unlock(&th->state_mtx);
-
-		/* bail if something has gone wrong */
-		if (state != BA_TRANSPORT_THREAD_STATE_RUNNING) {
+		/* Wait until transport thread is ready to process audio. */
+		if (ba_transport_thread_state_wait_running(th) == -1) {
 			g_dbus_method_invocation_return_error(inv, G_DBUS_ERROR,
-					G_DBUS_ERROR_IO_ERROR, "Acquire transport: %s", strerror(EIO));
+					G_DBUS_ERROR_IO_ERROR, "Acquire transport: %s", strerror(errno));
 			goto fail;
 		}
 
 	}
 
+	pthread_mutex_lock(&pcm->mutex);
 	/* get correct PIPE endpoint - PIPE is unidirectional */
 	pcm->fd = pcm_fds[is_sink ? 0 : 1];
 	/* set newly opened PCM as active */
 	pcm->active = true;
+	pthread_mutex_unlock(&pcm->mutex);
 
 	GIOChannel *ch = g_io_channel_unix_new(pcm_fds[2]);
+	g_io_channel_set_close_on_unref(ch, TRUE);
+	g_io_channel_set_encoding(ch, NULL, NULL);
+
 	g_io_add_watch_full(ch, G_PRIORITY_DEFAULT, G_IO_IN,
 			bluealsa_pcm_controller, ba_transport_pcm_ref(pcm),
 			(GDestroyNotify)ba_transport_pcm_unref);
-	g_io_channel_set_close_on_unref(ch, TRUE);
-	g_io_channel_set_encoding(ch, NULL, NULL);
 	g_io_channel_unref(ch);
 
 	/* notify our audio thread that the FIFO is ready */
 	ba_transport_thread_signal_send(th, BA_TRANSPORT_THREAD_SIGNAL_PCM_OPEN);
 
-	pthread_mutex_unlock(&pcm->mutex);
-
 	int fds[2] = { pcm_fds[is_sink ? 1 : 0], pcm_fds[3] };
 	GUnixFDList *fd_list = g_unix_fd_list_new_from_array(fds, 2);
 	g_dbus_method_invocation_return_value_with_unix_fd_list(inv,
 			g_variant_new("(hh)", 0, 1), fd_list);
 	g_object_unref(fd_list);
 
+	pthread_mutex_unlock(&pcm->client_mtx);
 	return;
 
 fail:
-	pthread_mutex_unlock(&pcm->mutex);
+	pthread_mutex_unlock(&pcm->client_mtx);
 	/* clean up created file descriptors */
 	for (i = 0; i < ARRAYSIZE(pcm_fds); i++)
 		if (pcm_fds[i] != -1)
@@ -605,7 +569,7 @@ static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv, void *userdata)
 	GVariantBuilder codecs;
 	g_variant_builder_init(&codecs, G_VARIANT_TYPE("a{sa{sv}}"));
 
-	if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_A2DP) {
+	if (t->profile & BA_TRANSPORT_PROFILE_MASK_A2DP) {
 
 		GArray *codec_ids = g_array_sized_new(FALSE, FALSE, sizeof(uint16_t), 16);
 		size_t i;
@@ -643,15 +607,28 @@ static void bluealsa_pcm_get_codecs(GDBusMethodInvocation *inv, void *userdata)
 		g_array_free(codec_ids, TRUE);
 
 	}
-	else if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_SCO) {
+	else if (t->profile & BA_TRANSPORT_PROFILE_MASK_SCO) {
 
-		if (config.hfp.codecs.cvsd)
+		const struct ba_rfcomm *t_sco_rfcomm = t->sco.rfcomm;
+
+		/* HFP codec is selected by the AG. Because of that, HF is not aware of
+		 * AG supported codecs until the codec is actually selected. Anyway, we
+		 * will try to provide some heuristic here.
+		 * For built-in HFP profiles we will mark given codec as available, if
+		 * both AG and HF can support it. When HFP is provided by an external
+		 * application like oFono, we will mark given codec as available, if it
+		 * is enabled by our global configuration. */
+
+		if ((t_sco_rfcomm != NULL &&
+					t_sco_rfcomm->ag_codecs.cvsd && t_sco_rfcomm->hf_codecs.cvsd) ||
+				(t_sco_rfcomm == NULL && config.hfp.codecs.cvsd))
 			g_variant_builder_add(&codecs, "{sa{sv}}",
 					hfp_codec_id_to_string(HFP_CODEC_CVSD), NULL);
 
 #if ENABLE_MSBC
-		if (config.hfp.codecs.msbc &&
-				t->sco.rfcomm != NULL && t->sco.rfcomm->codecs.msbc)
+		if ((t_sco_rfcomm != NULL &&
+					t_sco_rfcomm->ag_codecs.msbc && t_sco_rfcomm->hf_codecs.msbc) ||
+				(t_sco_rfcomm == NULL && config.hfp.codecs.msbc))
 			g_variant_builder_add(&codecs, "{sa{sv}}",
 					hfp_codec_id_to_string(HFP_CODEC_MSBC), NULL);
 #endif
@@ -674,10 +651,16 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv, void *userdata
 	const char *codec_name;
 	const char *property;
 
+	/* Since transport can provide more than one PCM interface, i.e., source
+	 * and sink for bi-directional transports like HSP/HFP. In such case, both
+	 * PCMs should use the same codec. Given that, we need to lock codec
+	 * selection on the transport level. */
+	pthread_mutex_lock(&t->codec_select_client_mtx);
+
 	a2dp_t a2dp_configuration = {};
 	size_t a2dp_configuration_size = 0;
 
-	g_variant_get(params, "(sa{sv})", &codec_name, &properties);
+	g_variant_get(params, "(&sa{sv})", &codec_name, &properties);
 	while (g_variant_iter_next(properties, "{&sv}", &property, &value)) {
 
 		if (strcmp(property, "Configuration") == 0 &&
@@ -700,7 +683,7 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv, void *userdata
 		value = NULL;
 	}
 
-	if (t->type.profile & BA_TRANSPORT_PROFILE_MASK_A2DP) {
+	if (t->profile & BA_TRANSPORT_PROFILE_MASK_A2DP) {
 
 		/* support for Stream End-Points not enabled in BlueZ */
 		if (t->d->seps == NULL) {
@@ -754,7 +737,12 @@ static void bluealsa_pcm_select_codec(GDBusMethodInvocation *inv, void *userdata
 	}
 	else {
 
-		uint16_t codec_id = hfp_codec_id_from_string(codec_name);
+		uint16_t codec_id;
+		if ((codec_id = hfp_codec_id_from_string(codec_name)) == HFP_CODEC_UNDEFINED) {
+			errmsg = "HFP codec not available";
+			goto fail;
+		}
+
 		if (ba_transport_select_codec_sco(t, codec_id) == -1)
 			goto fail;
 
@@ -771,6 +759,7 @@ fail:
 			G_DBUS_ERROR_FAILED, "%s", errmsg);
 
 final:
+	pthread_mutex_unlock(&t->codec_select_client_mtx);
 	g_variant_iter_free(properties);
 	if (value != NULL)
 		g_variant_unref(value);
@@ -815,10 +804,10 @@ static GVariant *bluealsa_pcm_get_properties(void *userdata) {
 
 static GVariant *bluealsa_pcm_get_property(const char *property,
 		GError **error, void *userdata) {
-	(void)error;
 
 	struct ba_transport_pcm *pcm = (struct ba_transport_pcm *)userdata;
 	struct ba_device *d = pcm->t->d;
+	GVariant *value;
 
 	if (strcmp(property, "Device") == 0)
 		return ba_variant_new_device_path(d);
@@ -828,14 +817,24 @@ static GVariant *bluealsa_pcm_get_property(const char *property,
 		return ba_variant_new_transport_type(pcm->t);
 	if (strcmp(property, "Mode") == 0)
 		return ba_variant_new_pcm_mode(pcm);
+	if (strcmp(property, "Running") == 0)
+		return ba_variant_new_pcm_running(pcm);
 	if (strcmp(property, "Format") == 0)
 		return ba_variant_new_pcm_format(pcm);
 	if (strcmp(property, "Channels") == 0)
 		return ba_variant_new_pcm_channels(pcm);
 	if (strcmp(property, "Sampling") == 0)
 		return ba_variant_new_pcm_sampling(pcm);
-	if (strcmp(property, "Codec") == 0)
-		return ba_variant_new_pcm_codec(pcm);
+	if (strcmp(property, "Codec") == 0) {
+		if ((value = ba_variant_new_pcm_codec(pcm)) == NULL)
+			goto unavailable;
+		return value;
+	}
+	if (strcmp(property, "CodecConfiguration") == 0) {
+		if ((value = ba_variant_new_pcm_codec_config(pcm)) == NULL)
+			goto unavailable;
+		return value;
+	}
 	if (strcmp(property, "Delay") == 0)
 		return ba_variant_new_pcm_delay(pcm);
 	if (strcmp(property, "SoftVolume") == 0)
@@ -845,6 +844,12 @@ static GVariant *bluealsa_pcm_get_property(const char *property,
 
 	g_assert_not_reached();
 	return NULL;
+
+unavailable:
+	if (error != NULL)
+		*error = g_error_new(G_DBUS_ERROR, G_DBUS_ERROR_INVALID_ARGS,
+				"No such property '%s'", property);
+	return NULL;
 }
 
 static bool bluealsa_pcm_set_property(const char *property, GVariant *value,
@@ -861,17 +866,22 @@ static bool bluealsa_pcm_set_property(const char *property, GVariant *value,
 
 	if (strcmp(property, "Volume") == 0) {
 
+		const bool is_sco = pcm->t->profile & BA_TRANSPORT_PROFILE_MASK_SCO;
+		const int max = is_sco ? HFP_VOLUME_GAIN_MAX : BLUEZ_A2DP_VOLUME_MAX;
+
 		uint16_t packed = g_variant_get_uint16(value);
 		uint8_t ch1 = packed >> 8;
 		uint8_t ch2 = packed & 0xFF;
 
-		int ch1_level = ba_transport_pcm_volume_bt_to_level(pcm, ch1 & 0x7F);
+		int ch1_level = ba_transport_pcm_volume_range_to_level(ch1 & 0x7F, max);
 		bool ch1_muted = !!(ch1 & 0x80);
-		int ch2_level = ba_transport_pcm_volume_bt_to_level(pcm, ch2 & 0x7F);
+		int ch2_level = ba_transport_pcm_volume_range_to_level(ch2 & 0x7F, max);
 		bool ch2_muted = !!(ch2 & 0x80);
 
+		pthread_mutex_lock(&pcm->mutex);
 		ba_transport_pcm_volume_set(&pcm->volume[0], &ch1_level, &ch1_muted, NULL);
 		ba_transport_pcm_volume_set(&pcm->volume[1], &ch2_level, &ch2_muted, NULL);
+		pthread_mutex_unlock(&pcm->mutex);
 
 		debug("Setting volume: %u [%.2f dB] %c%c %u [%.2f dB]",
 				ch1 & 0x7F, 0.01 * ch1_level, ch1_muted ? 'x' : '<',
@@ -922,13 +932,6 @@ int bluealsa_dbus_pcm_register(struct ba_transport_pcm *pcm) {
 	g_dbus_object_manager_server_export(bluealsa_dbus_manager, skeleton);
 	pcm->ba_dbus_exported = true;
 
-	GVariantBuilder props;
-	ba_variant_populate_pcm(&props, pcm);
-	g_dbus_connection_emit_signal(config.dbus, NULL,
-			bluealsa_dbus_manager_path, BLUEALSA_IFACE_MANAGER, "PCMAdded",
-			g_variant_new("(oa{sv})", pcm->ba_dbus_path, &props), NULL);
-	g_variant_builder_clear(&props);
-
 fail:
 
 	if (skeleton != NULL)
@@ -944,6 +947,8 @@ void bluealsa_dbus_pcm_update(struct ba_transport_pcm *pcm, unsigned int mask) {
 	GVariantBuilder props;
 	g_variant_builder_init(&props, G_VARIANT_TYPE("a{sv}"));
 
+	if (mask & BA_DBUS_PCM_UPDATE_RUNNING)
+		g_variant_builder_add(&props, "{sv}", "Running", ba_variant_new_pcm_running(pcm));
 	if (mask & BA_DBUS_PCM_UPDATE_FORMAT)
 		g_variant_builder_add(&props, "{sv}", "Format", ba_variant_new_pcm_format(pcm));
 	if (mask & BA_DBUS_PCM_UPDATE_CHANNELS)
@@ -952,6 +957,8 @@ void bluealsa_dbus_pcm_update(struct ba_transport_pcm *pcm, unsigned int mask) {
 		g_variant_builder_add(&props, "{sv}", "Sampling", ba_variant_new_pcm_sampling(pcm));
 	if (mask & BA_DBUS_PCM_UPDATE_CODEC)
 		g_variant_builder_add(&props, "{sv}", "Codec", ba_variant_new_pcm_codec(pcm));
+	if (mask & BA_DBUS_PCM_UPDATE_CODEC_CONFIG)
+		g_variant_builder_add(&props, "{sv}", "CodecConfiguration", ba_variant_new_pcm_codec_config(pcm));
 	if (mask & BA_DBUS_PCM_UPDATE_DELAY)
 		g_variant_builder_add(&props, "{sv}", "Delay", ba_variant_new_pcm_delay(pcm));
 	if (mask & BA_DBUS_PCM_UPDATE_SOFT_VOLUME)
@@ -973,10 +980,6 @@ void bluealsa_dbus_pcm_unregister(struct ba_transport_pcm *pcm) {
 	g_dbus_object_manager_server_unexport(bluealsa_dbus_manager, pcm->ba_dbus_path);
 	pcm->ba_dbus_exported = false;
 
-	g_dbus_connection_emit_signal(config.dbus, NULL,
-			bluealsa_dbus_manager_path, BLUEALSA_IFACE_MANAGER, "PCMRemoved",
-			g_variant_new("(o)", pcm->ba_dbus_path), NULL);
-
 }
 
 static GVariant *bluealsa_rfcomm_get_properties(void *userdata) {
