@@ -1,6 +1,6 @@
 /*
  * BlueALSA - bluealsa-dbus.h
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2022 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -22,13 +22,15 @@
 #include "ba-device.h"
 #include "ba-transport.h"
 
-#define BA_DBUS_PCM_UPDATE_FORMAT      (1 << 0)
-#define BA_DBUS_PCM_UPDATE_CHANNELS    (1 << 1)
-#define BA_DBUS_PCM_UPDATE_SAMPLING    (1 << 2)
-#define BA_DBUS_PCM_UPDATE_CODEC       (1 << 3)
-#define BA_DBUS_PCM_UPDATE_DELAY       (1 << 4)
-#define BA_DBUS_PCM_UPDATE_SOFT_VOLUME (1 << 5)
-#define BA_DBUS_PCM_UPDATE_VOLUME      (1 << 6)
+#define BA_DBUS_PCM_UPDATE_FORMAT       (1 << 0)
+#define BA_DBUS_PCM_UPDATE_CHANNELS     (1 << 1)
+#define BA_DBUS_PCM_UPDATE_SAMPLING     (1 << 2)
+#define BA_DBUS_PCM_UPDATE_CODEC        (1 << 3)
+#define BA_DBUS_PCM_UPDATE_CODEC_CONFIG (1 << 4)
+#define BA_DBUS_PCM_UPDATE_DELAY        (1 << 5)
+#define BA_DBUS_PCM_UPDATE_SOFT_VOLUME  (1 << 6)
+#define BA_DBUS_PCM_UPDATE_VOLUME       (1 << 7)
+#define BA_DBUS_PCM_UPDATE_RUNNING      (1 << 8)
 
 #define BA_DBUS_RFCOMM_UPDATE_FEATURES (1 << 0)
 #define BA_DBUS_RFCOMM_UPDATE_BATTERY  (1 << 1)
