@@ -1,6 +1,6 @@
 /*
  * BlueALSA - bluez-iface.h
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -30,8 +30,8 @@
 #define BLUEZ_TRANSPORT_STATE_PENDING "pending"
 #define BLUEZ_TRANSPORT_STATE_ACTIVE  "active"
 
-extern const GDBusInterfaceInfo bluez_iface_battery_provider;
-extern const GDBusInterfaceInfo bluez_iface_media_endpoint;
-extern const GDBusInterfaceInfo bluez_iface_profile;
+extern const GDBusInterfaceInfo org_bluez_battery_provider1_interface;
+extern const GDBusInterfaceInfo org_bluez_media_endpoint1_interface;
+extern const GDBusInterfaceInfo org_bluez_profile1_interface;
 
 #endif
