@@ -1,6 +1,6 @@
 /*
  * BlueALSA - bluez.c
- * Copyright (c) 2016-2022 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -9,10 +9,10 @@
  */
 
 #include "bluez.h"
+/* IWYU pragma: no_include "config.h" */
 
 #include <errno.h>
 #include <pthread.h>
-#include <stdatomic.h>
 #include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -42,6 +42,7 @@
 #include "sco.h"
 #include "utils.h"
 #include "shared/a2dp-codecs.h"
+#include "shared/bluetooth.h"
 #include "shared/defs.h"
 #include "shared/log.h"
 
@@ -59,16 +60,14 @@ struct bluez_dbus_object_data {
 	GDBusInterfaceSkeleton *ifs;
 	/* associated adapter */
 	int hci_dev_id;
-	/* the type of the transport */
-	struct ba_transport_type ttype;
+	/* registered profile */
+	enum ba_transport_profile profile;
 	/* media endpoint codec */
 	const struct a2dp_codec *codec;
 	/* determine whether object is registered in BlueZ */
 	bool registered;
 	/* determine whether object is used */
 	bool connected;
-	/* memory self-management */
-	atomic_int ref_count;
 };
 
 /**
@@ -86,11 +85,6 @@ static pthread_mutex_t bluez_mutex = PTHREAD_MUTEX_INITIALIZER;
 static GHashTable *dbus_object_data_map = NULL;
 static struct bluez_adapter bluez_adapters[HCI_MAX_DEV] = { 0 };
 
-#define bluez_adapters_device_lookup(hci_dev_id, addr) \
-	g_hash_table_lookup(bluez_adapters[hci_dev_id].device_sep_map, addr)
-#define bluez_adapters_device_get_sep(seps, i) \
-	g_array_index(seps, struct a2dp_sep, i)
-
 static void bluez_register_a2dp_all(struct ba_adapter *);
 static void bluez_register_battery_provider_manager(struct bluez_adapter *);
 
@@ -103,21 +97,36 @@ static struct bluez_adapter *bluez_adapter_new(struct ba_adapter *a) {
 	return &bluez_adapters[a->hci.dev_id];
 }
 
-static void bluez_adapter_free(struct bluez_adapter *adapter) {
-	if (adapter->adapter == NULL)
+static void bluez_adapter_free(struct bluez_adapter *b_adapter) {
+	if (b_adapter->adapter == NULL)
 		return;
-	g_hash_table_destroy(adapter->device_sep_map);
-	adapter->device_sep_map = NULL;
-	g_object_unref(adapter->battery_manager);
-	adapter->battery_manager = NULL;
-	ba_adapter_destroy(adapter->adapter);
-	adapter->adapter = NULL;
+	g_hash_table_destroy(b_adapter->device_sep_map);
+	b_adapter->device_sep_map = NULL;
+	g_object_unref(b_adapter->battery_manager);
+	b_adapter->battery_manager = NULL;
+	ba_adapter_destroy(b_adapter->adapter);
+	b_adapter->adapter = NULL;
 }
 
-static void bluez_dbus_object_data_unref(
+/**
+ * Get Stream End-Points (SEPs) associated with the given device. */
+static GArray *bluez_adapter_get_device_seps(
+		struct bluez_adapter *b_adapter,
+		const bdaddr_t *addr) {
+	GArray *seps;
+	if ((seps = g_hash_table_lookup(b_adapter->device_sep_map, addr)) != NULL)
+		return seps;
+	seps = g_array_new(FALSE, FALSE, sizeof(struct a2dp_sep));
+	g_hash_table_insert(b_adapter->device_sep_map, g_memdup2(addr, sizeof(*addr)), seps);
+	return seps;
+}
+
+static void bluez_dbus_object_data_free(
 		struct bluez_dbus_object_data *obj) {
-	if (atomic_fetch_sub_explicit(&obj->ref_count, 1, memory_order_relaxed) > 1)
-		return;
+	if (obj->ifs != NULL) {
+		g_dbus_interface_skeleton_unexport(obj->ifs);
+		g_object_unref(obj->ifs);
+	}
 	free(obj);
 }
 
@@ -146,6 +155,99 @@ static bool bluez_match_dbus_adapter(
 	return false;
 }
 
+/**
+ * Get BlueZ D-Bus object path for given transport profile. */
+static const char *bluez_transport_profile_to_bluez_object_path(
+		enum ba_transport_profile profile,
+		uint16_t codec_id) {
+	switch (profile) {
+	case BA_TRANSPORT_PROFILE_NONE:
+		return "/";
+	case BA_TRANSPORT_PROFILE_A2DP_SOURCE:
+		switch (codec_id) {
+		case A2DP_CODEC_SBC:
+			return "/A2DP/SBC/source";
+#if ENABLE_MPEG
+		case A2DP_CODEC_MPEG12:
+			return "/A2DP/MPEG/source";
+#endif
+#if ENABLE_AAC
+		case A2DP_CODEC_MPEG24:
+			return "/A2DP/AAC/source";
+#endif
+#if ENABLE_APTX
+		case A2DP_CODEC_VENDOR_APTX:
+			return "/A2DP/aptX/source";
+#endif
+#if ENABLE_APTX_HD
+		case A2DP_CODEC_VENDOR_APTX_HD:
+			return "/A2DP/aptXHD/source";
+#endif
+#if ENABLE_FASTSTREAM
+		case A2DP_CODEC_VENDOR_FASTSTREAM:
+			return "/A2DP/FastStream/source";
+#endif
+#if ENABLE_LC3PLUS
+		case A2DP_CODEC_VENDOR_LC3PLUS:
+			return "/A2DP/LC3plus/source";
+#endif
+#if ENABLE_LDAC
+		case A2DP_CODEC_VENDOR_LDAC:
+			return "/A2DP/LDAC/source";
+#endif
+		default:
+			error("Unsupported A2DP codec: %#x", codec_id);
+			g_assert_not_reached();
+		}
+	case BA_TRANSPORT_PROFILE_A2DP_SINK:
+		switch (codec_id) {
+		case A2DP_CODEC_SBC:
+			return "/A2DP/SBC/sink";
+#if ENABLE_MPEG
+		case A2DP_CODEC_MPEG12:
+			return "/A2DP/MPEG/sink";
+#endif
+#if ENABLE_AAC
+		case A2DP_CODEC_MPEG24:
+			return "/A2DP/AAC/sink";
+#endif
+#if ENABLE_APTX
+		case A2DP_CODEC_VENDOR_APTX:
+			return "/A2DP/aptX/sink";
+#endif
+#if ENABLE_APTX_HD
+		case A2DP_CODEC_VENDOR_APTX_HD:
+			return "/A2DP/aptXHD/sink";
+#endif
+#if ENABLE_FASTSTREAM
+		case A2DP_CODEC_VENDOR_FASTSTREAM:
+			return "/A2DP/FastStream/sink";
+#endif
+#if ENABLE_LC3PLUS
+		case A2DP_CODEC_VENDOR_LC3PLUS:
+			return "/A2DP/LC3plus/sink";
+#endif
+#if ENABLE_LDAC
+		case A2DP_CODEC_VENDOR_LDAC:
+			return "/A2DP/LDAC/sink";
+#endif
+		default:
+			error("Unsupported A2DP codec: %#x", codec_id);
+			g_assert_not_reached();
+		}
+	case BA_TRANSPORT_PROFILE_HFP_HF:
+		return "/HFP/HandsFree";
+	case BA_TRANSPORT_PROFILE_HFP_AG:
+		return "/HFP/AudioGateway";
+	case BA_TRANSPORT_PROFILE_HSP_HS:
+		return "/HSP/Headset";
+	case BA_TRANSPORT_PROFILE_HSP_AG:
+		return "/HSP/AudioGateway";
+	}
+	g_assert_not_reached();
+	return "/";
+}
+
 /**
  * Get transport state from BlueZ state string. */
 static enum bluez_a2dp_transport_state bluez_a2dp_transport_state_from_string(
@@ -284,14 +386,14 @@ static void bluez_endpoint_set_configuration(GDBusMethodInvocation *inv, void *u
 	}
 
 	if (d->seps == NULL)
-		d->seps = bluez_adapters_device_lookup(a->hci.dev_id, &addr);
+		d->seps = bluez_adapter_get_device_seps(&bluez_adapters[a->hci.dev_id], &addr);
 
 	if (ba_transport_lookup(d, transport_path) != NULL) {
 		error("Transport already configured: %s", transport_path);
 		goto fail;
 	}
 
-	if ((t = ba_transport_new_a2dp(d, dbus_obj->ttype,
+	if ((t = ba_transport_new_a2dp(d, dbus_obj->profile,
 					sender, transport_path, codec, &configuration)) == NULL) {
 		error("Couldn't create new transport: %s", strerror(errno));
 		goto fail;
@@ -299,11 +401,16 @@ static void bluez_endpoint_set_configuration(GDBusMethodInvocation *inv, void *u
 
 	/* Skip volume level initialization in case of A2DP Source
 	 * profile and software volume control. */
-	if (!(t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE &&
+	if (!(t->profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE &&
 				t->a2dp.pcm.soft_volume)) {
-		int level = ba_transport_pcm_volume_bt_to_level(&t->a2dp.pcm, volume);
+
+		int level = ba_transport_pcm_volume_range_to_level(volume, BLUEZ_A2DP_VOLUME_MAX);
+
+		pthread_mutex_lock(&t->a2dp.pcm.mutex);
 		ba_transport_pcm_volume_set(&t->a2dp.pcm.volume[0], &level, NULL, NULL);
 		ba_transport_pcm_volume_set(&t->a2dp.pcm.volume[1], &level, NULL, NULL);
+		pthread_mutex_unlock(&t->a2dp.pcm.mutex);
+
 	}
 
 	t->a2dp.bluez_dbus_sep_path = dbus_obj->path;
@@ -311,7 +418,7 @@ static void bluez_endpoint_set_configuration(GDBusMethodInvocation *inv, void *u
 	t->a2dp.volume = volume;
 
 	debug("%s configured for device %s",
-			ba_transport_type_to_string(t->type),
+			ba_transport_debug_name(t),
 			batostr_(&d->addr));
 	hexdump("A2DP selected configuration blob",
 			&configuration, codec->capabilities_size, true);
@@ -462,11 +569,8 @@ static void bluez_register_a2dp(
 		.dispatchers = dispatchers,
 	};
 
-	struct ba_transport_type ttype = {
-		.profile = codec->dir == A2DP_SOURCE ?
-			BA_TRANSPORT_PROFILE_A2DP_SOURCE : BA_TRANSPORT_PROFILE_A2DP_SINK,
-		.codec = codec->codec_id,
-	};
+	enum ba_transport_profile profile = codec->dir == A2DP_SOURCE ?
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE : BA_TRANSPORT_PROFILE_A2DP_SINK;
 
 	int registered = 0;
 	int connected = 0;
@@ -480,7 +584,8 @@ static void bluez_register_a2dp(
 
 		char path[sizeof(dbus_obj->path)];
 		snprintf(path, sizeof(path), "/org/bluez/%s%s/%d", adapter->hci.name,
-				g_dbus_transport_type_to_bluez_object_path(ttype), ++registered);
+				bluez_transport_profile_to_bluez_object_path(profile, codec->codec_id),
+				++registered);
 
 		if ((dbus_obj = g_hash_table_lookup(dbus_object_data_map, path)) == NULL) {
 
@@ -500,12 +605,11 @@ static void bluez_register_a2dp(
 			strncpy(dbus_obj->path, path, sizeof(dbus_obj->path));
 			dbus_obj->hci_dev_id = adapter->hci.dev_id;
 			dbus_obj->codec = codec;
-			dbus_obj->ttype = ttype;
-			dbus_obj->ref_count = 2;
+			dbus_obj->profile = profile;
 
 			bluez_MediaEndpointIfaceSkeleton *ifs_endpoint;
 			if ((ifs_endpoint = bluez_media_endpoint_iface_skeleton_new(&vtable,
-							dbus_obj, (GDestroyNotify)bluez_dbus_object_data_unref)) == NULL) {
+							dbus_obj, NULL)) == NULL) {
 				free(dbus_obj);
 				goto fail;
 			}
@@ -552,11 +656,11 @@ static void bluez_register_a2dp_all(struct ba_adapter *adapter) {
 		switch (c->dir) {
 		case A2DP_SOURCE:
 			if (config.profile.a2dp_source && c->enabled)
-				bluez_register_a2dp(adapter, c, BLUETOOTH_UUID_A2DP_SOURCE);
+				bluez_register_a2dp(adapter, c, BT_UUID_A2DP_SOURCE);
 			break;
 		case A2DP_SINK:
 			if (config.profile.a2dp_sink && c->enabled)
-				bluez_register_a2dp(adapter, c, BLUETOOTH_UUID_A2DP_SINK);
+				bluez_register_a2dp(adapter, c, BT_UUID_A2DP_SINK);
 			break;
 		}
 	}
@@ -727,13 +831,12 @@ static void bluez_profile_new_connection(GDBusMethodInvocation *inv, void *userd
 
 	const char *device_path;
 	GVariantIter *properties;
-	GUnixFDList *fd_list;
 	GError *err = NULL;
 	int fd = -1;
 
-	g_variant_get(params, "(&oha{sv})", &device_path, &fd, &properties);
+	g_variant_get(params, "(&oha{sv})", &device_path, NULL, &properties);
 
-	fd_list = g_dbus_message_get_unix_fd_list(msg);
+	GUnixFDList *fd_list = g_dbus_message_get_unix_fd_list(msg);
 	if ((fd = g_unix_fd_list_get(fd_list, 0, &err)) == -1) {
 		error("Couldn't obtain RFCOMM socket: %s", err->message);
 		goto fail;
@@ -753,7 +856,7 @@ static void bluez_profile_new_connection(GDBusMethodInvocation *inv, void *userd
 		goto fail;
 	}
 
-	if ((t = ba_transport_new_sco(d, dbus_obj->ttype,
+	if ((t = ba_transport_new_sco(d, dbus_obj->profile,
 					sender, device_path, fd)) == NULL) {
 		error("Couldn't create new transport: %s", strerror(errno));
 		goto fail;
@@ -765,7 +868,7 @@ static void bluez_profile_new_connection(GDBusMethodInvocation *inv, void *userd
 	}
 
 	debug("%s configured for device %s",
-			ba_transport_type_to_string(t->type),
+			ba_transport_debug_name(t),
 			batostr_(&d->addr));
 
 	dbus_obj->connected = true;
@@ -892,7 +995,7 @@ final:
  * Register Bluetooth Hands-Free Audio Profile. */
 static void bluez_register_hfp(
 		const char *uuid,
-		uint32_t profile,
+		enum ba_transport_profile profile,
 		uint16_t version,
 		uint16_t features) {
 
@@ -910,10 +1013,6 @@ static void bluez_register_hfp(
 		.dispatchers = dispatchers,
 	};
 
-	struct ba_transport_type ttype = {
-		.profile = profile,
-	};
-
 	pthread_mutex_lock(&bluez_mutex);
 
 	struct bluez_dbus_object_data *dbus_obj;
@@ -921,7 +1020,7 @@ static void bluez_register_hfp(
 
 	char path[sizeof(dbus_obj->path)];
 	snprintf(path, sizeof(path), "/org/bluez%s",
-			g_dbus_transport_type_to_bluez_object_path(ttype));
+			bluez_transport_profile_to_bluez_object_path(profile, -1));
 
 	if ((dbus_obj = g_hash_table_lookup(dbus_object_data_map, path)) == NULL) {
 
@@ -934,12 +1033,11 @@ static void bluez_register_hfp(
 
 		strncpy(dbus_obj->path, path, sizeof(dbus_obj->path));
 		dbus_obj->hci_dev_id = -1;
-		dbus_obj->ttype = ttype;
-		dbus_obj->ref_count = 2;
+		dbus_obj->profile = profile;
 
 		bluez_ProfileIfaceSkeleton *ifs_profile;
 		if ((ifs_profile = bluez_profile_iface_skeleton_new(&vtable,
-						dbus_obj, (GDestroyNotify)bluez_dbus_object_data_unref)) == NULL) {
+						dbus_obj, NULL)) == NULL) {
 			free(dbus_obj);
 			goto fail;
 		}
@@ -980,16 +1078,16 @@ fail:
  * this function will do nothing. */
 static void bluez_register_hfp_all(void) {
 	if (config.profile.hsp_hs)
-		bluez_register_hfp(BLUETOOTH_UUID_HSP_HS, BA_TRANSPORT_PROFILE_HSP_HS,
+		bluez_register_hfp(BT_UUID_HSP_HS, BA_TRANSPORT_PROFILE_HSP_HS,
 				0x0102 /* HSP 1.2 */, 0x1 /* remote audio volume control */);
 	if (config.profile.hsp_ag)
-		bluez_register_hfp(BLUETOOTH_UUID_HSP_AG, BA_TRANSPORT_PROFILE_HSP_AG,
+		bluez_register_hfp(BT_UUID_HSP_AG, BA_TRANSPORT_PROFILE_HSP_AG,
 				0x0102 /* HSP 1.2 */, 0x0);
 	if (config.profile.hfp_hf)
-		bluez_register_hfp(BLUETOOTH_UUID_HFP_HF, BA_TRANSPORT_PROFILE_HFP_HF,
+		bluez_register_hfp(BT_UUID_HFP_HF, BA_TRANSPORT_PROFILE_HFP_HF,
 				0x0107 /* HFP 1.7 */, config.hfp.features_sdp_hf);
 	if (config.profile.hfp_ag)
-		bluez_register_hfp(BLUETOOTH_UUID_HFP_AG, BA_TRANSPORT_PROFILE_HFP_AG,
+		bluez_register_hfp(BT_UUID_HFP_AG, BA_TRANSPORT_PROFILE_HFP_AG,
 				0x0107 /* HFP 1.7 */, config.hfp.features_sdp_ag);
 }
 
@@ -997,6 +1095,26 @@ static void bluez_register_hfp_all(void) {
  * Register to the BlueZ service. */
 static void bluez_register(void) {
 
+	const struct {
+		const char *uuid;
+		enum ba_transport_profile profile;
+		bool enabled;
+		bool global;
+	} uuids[] = {
+		{ BT_UUID_A2DP_SOURCE, BA_TRANSPORT_PROFILE_A2DP_SOURCE,
+			config.profile.a2dp_source, false },
+		{ BT_UUID_A2DP_SINK, BA_TRANSPORT_PROFILE_A2DP_SINK,
+			config.profile.a2dp_sink, false },
+		{ BT_UUID_HSP_HS, BA_TRANSPORT_PROFILE_HSP_HS,
+			config.profile.hsp_hs, true },
+		{ BT_UUID_HSP_AG, BA_TRANSPORT_PROFILE_HSP_AG,
+			config.profile.hsp_ag, true },
+		{ BT_UUID_HFP_HF, BA_TRANSPORT_PROFILE_HFP_HF,
+			config.profile.hfp_hf, true },
+		{ BT_UUID_HFP_AG, BA_TRANSPORT_PROFILE_HFP_AG,
+			config.profile.hfp_ag, true },
+	};
+
 	GError *err = NULL;
 	GVariantIter *objects = NULL;
 	if ((objects = g_dbus_get_managed_objects(config.dbus, BLUEZ_SERVICE, "/", &err)) == NULL) {
@@ -1006,6 +1124,8 @@ static void bluez_register(void) {
 	}
 
 	bool adapters[HCI_MAX_DEV] = { 0 };
+	unsigned int adapters_profiles[HCI_MAX_DEV] = { 0 };
+	unsigned int profiles = 0;
 
 	GVariantIter *interfaces;
 	GVariantIter *properties;
@@ -1016,14 +1136,33 @@ static void bluez_register(void) {
 
 	while (g_variant_iter_next(objects, "{&oa{sa{sv}}}", &object_path, &interfaces)) {
 		while (g_variant_iter_next(interfaces, "{&sa{sv}}", &interface, &properties)) {
-			if (strcmp(interface, BLUEZ_IFACE_ADAPTER) == 0)
+			if (strcmp(interface, BLUEZ_IFACE_ADAPTER) == 0) {
+
+				int hci_dev_id = g_dbus_bluez_object_path_to_hci_dev_id(object_path);
+				unsigned int adapter_profiles = 0;
+				bool valid = false;
+
 				while (g_variant_iter_next(properties, "{&sv}", &property, &value)) {
-					if (strcmp(property, "Address") == 0 &&
-							bluez_match_dbus_adapter(object_path, g_variant_get_string(value, NULL)))
-						/* mark adapter as valid for registration */
-						adapters[g_dbus_bluez_object_path_to_hci_dev_id(object_path)] = true;
+					if (strcmp(property, "Address") == 0)
+						/* check if adapter as valid for registration */
+						valid = bluez_match_dbus_adapter(object_path, g_variant_get_string(value, NULL));
+					else if (strcmp(property, "UUIDs") == 0) {
+						const char **value_uuids = g_variant_get_strv(value, NULL);
+						/* map UUIDs to BlueALSA transport profile mask */
+						for (size_t i = 0; value_uuids[i] != NULL; i++)
+							for (size_t ii = 0; ii < ARRAYSIZE(uuids); ii++)
+								if (strcasecmp(value_uuids[i], uuids[ii].uuid) == 0)
+									adapter_profiles |= uuids[ii].profile;
+						g_free(value_uuids);
+					}
 					g_variant_unref(value);
 				}
+
+				adapters[hci_dev_id] = valid;
+				adapters_profiles[hci_dev_id] = adapter_profiles;
+				profiles |= adapter_profiles;
+
+			}
 			g_variant_iter_free(properties);
 		}
 		g_variant_iter_free(interfaces);
@@ -1035,9 +1174,20 @@ static void bluez_register(void) {
 	for (i = 0; i < ARRAYSIZE(adapters); i++)
 		if (adapters[i] &&
 				(a = ba_adapter_new(i)) != NULL) {
+
+			for (size_t ii = 0; ii < ARRAYSIZE(uuids); ii++)
+				if (uuids[ii].enabled && !uuids[ii].global && adapters_profiles[i] & uuids[ii].profile)
+					warn("UUID already registered in BlueZ [%s]: %s", a->hci.name, uuids[ii].uuid);
+
+			/* register media endpoints */
 			bluez_adapter_new(a);
+
 		}
 
+	for (size_t ii = 0; ii < ARRAYSIZE(uuids); ii++)
+		if (uuids[ii].enabled && uuids[ii].global && profiles & uuids[ii].profile)
+			warn("UUID already registered in BlueZ: %s", uuids[ii].uuid);
+
 	/* HFP has to be registered globally */
 	bluez_register_hfp_all();
 
@@ -1090,7 +1240,7 @@ static void bluez_signal_interfaces_added(GDBusConnection *conn, const char *sen
 			while (g_variant_iter_next(properties, "{&sv}", &property, &value)) {
 				if (strcmp(property, "UUID") == 0) {
 					const char *uuid = g_variant_get_string(value, NULL);
-					if (strcasecmp(uuid, BLUETOOTH_UUID_A2DP_SINK) == 0)
+					if (strcasecmp(uuid, BT_UUID_A2DP_SINK) == 0)
 						sep.dir = A2DP_SINK;
 				}
 				else if (strcmp(property, "Codec") == 0)
@@ -1133,19 +1283,21 @@ static void bluez_signal_interfaces_added(GDBusConnection *conn, const char *sen
 		g_dbus_bluez_object_path_to_bdaddr(object_path, &addr);
 		int dev_id = g_dbus_bluez_object_path_to_hci_dev_id(object_path);
 
-		GArray *seps;
-		if ((seps = bluez_adapters_device_lookup(dev_id, &addr)) == NULL)
-			g_hash_table_insert(bluez_adapters[dev_id].device_sep_map,
-					g_memdup2(&addr, sizeof(addr)), seps = g_array_new(FALSE, FALSE, sizeof(sep)));
-
 		strncpy(sep.bluez_dbus_path, object_path, sizeof(sep.bluez_dbus_path) - 1);
 		if (sep.codec_id == A2DP_CODEC_VENDOR)
 			sep.codec_id = a2dp_get_vendor_codec_id(&sep.capabilities, sep.capabilities_size);
 
-		debug("Adding new Stream End-Point: %s: %s", batostr_(&addr),
+		debug("Adding new Stream End-Point: %s: %s: %s",
+				batostr_(&addr), sep.dir == A2DP_SOURCE ? "SRC" : "SNK",
 				a2dp_codecs_codec_id_to_string(sep.codec_id));
+
+		GArray *seps = bluez_adapter_get_device_seps(&bluez_adapters[dev_id], &addr);
 		g_array_append_val(seps, sep);
 
+		/* Collected SEPs are exposed via BlueALSA D-Bus API. We will sort them
+		 * here, so the D-Bus API will return codecs in the defined order. */
+		g_array_sort(seps, (GCompareFunc)a2dp_sep_cmp);
+
 	}
 
 }
@@ -1171,39 +1323,47 @@ static void bluez_signal_interfaces_removed(GDBusConnection *conn, const char *s
 
 	pthread_mutex_lock(&bluez_mutex);
 
+	/* Check whether this interface belongs to a HCI which exists in our
+	 * local BlueZ adapter cache - HCI that matches our HCI filter. */
+	if (hci_dev_id == -1 || bluez_adapters[hci_dev_id].adapter == NULL)
+		goto final;
+
 	while (g_variant_iter_next(interfaces, "&s", &interface))
 		if (strcmp(interface, BLUEZ_IFACE_ADAPTER) == 0) {
 
 			GHashTableIter iter;
 			struct bluez_dbus_object_data *dbus_obj;
 			g_hash_table_iter_init(&iter, dbus_object_data_map);
-			while (g_hash_table_iter_next(&iter, NULL, (gpointer)&dbus_obj))
-				if (dbus_obj->hci_dev_id == hci_dev_id) {
-					g_dbus_interface_skeleton_unexport(dbus_obj->ifs);
-					g_object_unref(dbus_obj->ifs);
-					g_hash_table_iter_remove(&iter);
-				}
+			while (g_hash_table_iter_next(&iter, NULL, (gpointer)&dbus_obj)) {
+				if (dbus_obj->hci_dev_id != hci_dev_id)
+					continue;
+				g_hash_table_iter_remove(&iter);
+			}
 
 			bluez_adapter_free(&bluez_adapters[hci_dev_id]);
 
 		}
 		else if (strcmp(interface, BLUEZ_IFACE_MEDIA_ENDPOINT) == 0) {
 
-			GArray *seps;
 			bdaddr_t addr;
-			size_t i;
-
 			g_dbus_bluez_object_path_to_bdaddr(object_path, &addr);
-			if ((seps = bluez_adapters_device_lookup(hci_dev_id, &addr)) != NULL)
-				for (i = 0; i < seps->len; i++)
-					if (strcmp(bluez_adapters_device_get_sep(seps, i).bluez_dbus_path, object_path) == 0)
-						g_array_remove_index_fast(seps, i);
+			GArray *seps = bluez_adapter_get_device_seps(&bluez_adapters[hci_dev_id], &addr);
+
+			for (size_t i = 0; i < seps->len; i++) {
+				const struct a2dp_sep *sep = &g_array_index(seps, struct a2dp_sep, i);
+				if (strcmp(sep->bluez_dbus_path, object_path) == 0) {
+					debug("Removing Stream End-Point: %s: %s: %s",
+							batostr_(&addr), sep->dir == A2DP_SOURCE ? "SRC" : "SNK",
+							a2dp_codecs_codec_id_to_string(sep->codec_id));
+					g_array_remove_index_fast(seps, i);
+				}
+			}
 
 		}
 
+final:
 	pthread_mutex_unlock(&bluez_mutex);
 	g_variant_iter_free(interfaces);
-
 }
 
 static void bluez_signal_transport_changed(GDBusConnection *conn, const char *sender,
@@ -1222,26 +1382,26 @@ static void bluez_signal_transport_changed(GDBusConnection *conn, const char *se
 	int hci_dev_id = g_dbus_bluez_object_path_to_hci_dev_id(transport_path);
 	if ((a = ba_adapter_lookup(hci_dev_id)) == NULL) {
 		error("Adapter not available: %s", transport_path);
-		return;
+		goto fail;
 	}
 
-	GVariantIter *properties = NULL;
-	const char *interface;
-	const char *property;
-	GVariant *value;
-
 	bdaddr_t addr;
 	g_dbus_bluez_object_path_to_bdaddr(transport_path, &addr);
 	if ((d = ba_device_lookup(a, &addr)) == NULL) {
 		error("Device not available: %s", transport_path);
-		goto final;
+		goto fail;
 	}
 
 	if ((t = ba_transport_lookup(d, transport_path)) == NULL) {
 		error("Transport not available: %s", transport_path);
-		goto final;
+		goto fail;
 	}
 
+	GVariantIter *properties;
+	const char *interface;
+	const char *property;
+	GVariant *value;
+
 	g_variant_get(params, "(&sa{sv}as)", &interface, &properties, NULL);
 	while (g_variant_iter_next(properties, "{&sv}", &property, &value)) {
 		debug("Signal: %s.%s(): %s: %s", interface_, signal, interface, property);
@@ -1260,15 +1420,21 @@ static void bluez_signal_transport_changed(GDBusConnection *conn, const char *se
 				g_variant_validate_value(value, G_VARIANT_TYPE_UINT16, property)) {
 			/* received volume is in range [0, 127] */
 			uint16_t volume = t->a2dp.volume = g_variant_get_uint16(value);
-			if (t->type.profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE &&
+			if (t->profile & BA_TRANSPORT_PROFILE_A2DP_SOURCE &&
 					t->a2dp.pcm.soft_volume)
 				debug("Skipping A2DP volume update: %u", volume);
 			else {
-				int level = ba_transport_pcm_volume_bt_to_level(&t->a2dp.pcm, volume);
+
+				int level = ba_transport_pcm_volume_range_to_level(volume, BLUEZ_A2DP_VOLUME_MAX);
 				debug("Updating A2DP volume: %u [%.2f dB]", volume, 0.01 * level);
+
+				pthread_mutex_lock(&t->a2dp.pcm.mutex);
 				ba_transport_pcm_volume_set(&t->a2dp.pcm.volume[0], &level, NULL, NULL);
 				ba_transport_pcm_volume_set(&t->a2dp.pcm.volume[1], &level, NULL, NULL);
+				pthread_mutex_unlock(&t->a2dp.pcm.mutex);
+
 				bluealsa_dbus_pcm_update(&t->a2dp.pcm, BA_DBUS_PCM_UPDATE_VOLUME);
+
 			}
 		}
 
@@ -1276,7 +1442,7 @@ static void bluez_signal_transport_changed(GDBusConnection *conn, const char *se
 	}
 	g_variant_iter_free(properties);
 
-final:
+fail:
 	if (a != NULL)
 		ba_adapter_unref(a);
 	if (d != NULL)
@@ -1299,17 +1465,9 @@ static void bluez_disappeared(GDBusConnection *conn, const char *name,
 
 	pthread_mutex_lock(&bluez_mutex);
 
-	GHashTableIter iter;
-	struct bluez_dbus_object_data *dbus_obj;
-	g_hash_table_iter_init(&iter, dbus_object_data_map);
-	while (g_hash_table_iter_next(&iter, NULL, (gpointer)&dbus_obj)) {
-		g_dbus_interface_skeleton_unexport(dbus_obj->ifs);
-		g_object_unref(dbus_obj->ifs);
-		g_hash_table_iter_remove(&iter);
-	}
+	g_hash_table_remove_all(dbus_object_data_map);
 
-	size_t i;
-	for (i = 0; i < ARRAYSIZE(bluez_adapters); i++)
+	for (size_t i = 0; i < ARRAYSIZE(bluez_adapters); i++)
 		bluez_adapter_free(&bluez_adapters[i]);
 
 	pthread_mutex_unlock(&bluez_mutex);
@@ -1345,7 +1503,7 @@ int bluez_init(void) {
 
 	if (dbus_object_data_map == NULL)
 		dbus_object_data_map = g_hash_table_new_full(g_str_hash, g_str_equal,
-				NULL, (GDestroyNotify)bluez_dbus_object_data_unref);
+				NULL, (GDestroyNotify)bluez_dbus_object_data_free);
 
 	bluez_subscribe_signals();
 	bluez_register();
