@@ -43,6 +43,7 @@
 #if ENABLE_OFONO
 # include "ofono.h"
 #endif
+#include "storage.h"
 #if ENABLE_UPOWER
 # include "upower.h"
 #endif
@@ -78,7 +79,8 @@ static char *get_a2dp_codecs(enum a2dp_dir dir) {
 	}
 
 	/* Sort A2DP codecs before displaying them. */
-	a2dp_codecs_qsort(a2dp_codecs_tmp, n);
+	qsort(a2dp_codecs_tmp, n, sizeof(*a2dp_codecs_tmp),
+			QSORT_COMPAR(a2dp_codec_ptr_cmp));
 
 	for (size_t i = 0; i < n; i++)
 		strv[i] = a2dp_codecs_codec_id_to_string(a2dp_codecs_tmp[i]->codec_id);
@@ -143,17 +145,18 @@ static void g_bus_name_lost(GDBusConnection *conn, const char *name, void *userd
 int main(int argc, char **argv) {
 
 	int opt;
-	const char *opts = "hVB:Si:p:c:";
+	const char *opts = "hVSB:i:p:c:";
 	const struct option longopts[] = {
 		{ "help", no_argument, NULL, 'h' },
 		{ "version", no_argument, NULL, 'V' },
-		{ "dbus", required_argument, NULL, 'B' },
 		{ "syslog", no_argument, NULL, 'S' },
+		{ "dbus", required_argument, NULL, 'B' },
 		{ "device", required_argument, NULL, 'i' },
 		{ "profile", required_argument, NULL, 'p' },
 		{ "codec", required_argument, NULL, 'c' },
 		{ "initial-volume", required_argument, NULL, 17 },
 		{ "keep-alive", required_argument, NULL, 8 },
+		{ "disable-realtek-usb-fix", no_argument, NULL, 21 },
 		{ "a2dp-force-mono", no_argument, NULL, 6 },
 		{ "a2dp-force-audio-cd", no_argument, NULL, 7 },
 		{ "a2dp-volume", no_argument, NULL, 9 },
@@ -189,36 +192,20 @@ int main(int argc, char **argv) {
 	opterr = 0;
 	while ((opt = getopt_long(argc, argv, opts, longopts, NULL)) != -1)
 		switch (opt) {
-		case 'S' /* --syslog */ :
-			syslog = true;
-			break;
-		}
-
-	log_open(argv[0], syslog, BLUEALSA_LOGTIME);
-
-	if (bluealsa_config_init() != 0) {
-		error("Couldn't initialize bluealsa config");
-		return EXIT_FAILURE;
-	}
-
-	/* parse options */
-	optind = 0; opterr = 1;
-	while ((opt = getopt_long(argc, argv, opts, longopts, NULL)) != -1)
-		switch (opt) {
-
 		case 'h' /* --help */ :
 			printf("Usage:\n"
 					"  %s -p PROFILE [OPTION]...\n"
 					"\nOptions:\n"
 					"  -h, --help\t\t\tprint this help and exit\n"
 					"  -V, --version\t\t\tprint version and exit\n"
-					"  -B, --dbus=NAME\t\tD-Bus service name suffix\n"
 					"  -S, --syslog\t\t\tsend output to syslog\n"
+					"  -B, --dbus=NAME\t\tD-Bus service name suffix\n"
 					"  -i, --device=hciX\t\tHCI device(s) to use\n"
 					"  -p, --profile=NAME\t\tset enabled BT profiles\n"
 					"  -c, --codec=NAME\t\tset enabled BT audio codecs\n"
 					"  --initial-volume=NUM\t\tinitial volume level [0-100]\n"
 					"  --keep-alive=SEC\t\tkeep Bluetooth transport alive\n"
+					"  --disable-realtek-usb-fix\tdisable fix for mSBC on Realtek USB\n"
 					"  --a2dp-force-mono\t\ttry to force monophonic sound\n"
 					"  --a2dp-force-audio-cd\t\ttry to force 44.1 kHz sampling\n"
 					"  --a2dp-volume\t\t\tnative volume control by default\n"
@@ -271,6 +258,27 @@ int main(int argc, char **argv) {
 			printf("%s\n", PACKAGE_VERSION);
 			return EXIT_SUCCESS;
 
+		case 'S' /* --syslog */ :
+			syslog = true;
+			break;
+		}
+
+	log_open(basename(argv[0]), syslog);
+
+	if (bluealsa_config_init() != 0) {
+		error("Couldn't initialize bluealsa config");
+		return EXIT_FAILURE;
+	}
+
+	/* parse options */
+	optind = 0; opterr = 1;
+	while ((opt = getopt_long(argc, argv, opts, longopts, NULL)) != -1)
+		switch (opt) {
+		case 'h' /* --help */ :
+		case 'V' /* --version */ :
+		case 'S' /* --syslog */ :
+			break;
+
 		case 'B' /* --dbus=NAME */ :
 			snprintf(dbus_service, sizeof(dbus_service), BLUEALSA_SERVICE ".%s", optarg);
 			if (!g_dbus_is_name(dbus_service)) {
@@ -279,9 +287,6 @@ int main(int argc, char **argv) {
 			}
 			break;
 
-		case 'S' /* --syslog */ :
-			break;
-
 		case 'i' /* --device=HCI */ :
 			g_array_append_val(config.hci_filter, optarg);
 			break;
@@ -376,6 +381,10 @@ int main(int argc, char **argv) {
 			config.keep_alive_time = atof(optarg) * 1000;
 			break;
 
+		case 21 /* --disable-realtek-usb-fix */ :
+			config.disable_realtek_usb_fix = true;
+			break;
+
 		case 6 /* --a2dp-force-mono */ :
 			config.a2dp.force_mono = true;
 			break;
@@ -415,13 +424,15 @@ int main(int argc, char **argv) {
 		case 5 /* --aac-bitrate=BPS */ :
 			config.aac_bitrate = atoi(optarg);
 			break;
-		case 15 /* --aac-latm-version=NUM */ :
-			config.aac_latm_version = atoi(optarg);
-			if (config.aac_latm_version > 2) {
-				error("Invalid LATM version [0, 2]: %s", optarg);
+		case 15 /* --aac-latm-version=NUM */ : {
+			char *tmp;
+			config.aac_latm_version = strtoul(optarg, &tmp, 10);
+			if (config.aac_latm_version > 2 || optarg == tmp || *tmp != '\0') {
+				error("Invalid LATM version {0, 1, 2}: %s", optarg);
 				return EXIT_FAILURE;
 			}
 			break;
+		}
 		case 18 /* --aac-true-bps */ :
 			config.aac_true_bps = true;
 			break;
@@ -566,6 +577,14 @@ int main(int argc, char **argv) {
 
 	a2dp_codecs_init();
 
+	const char *storage_base_dir = BLUEALSA_STORAGE_DIR;
+#if ENABLE_SYSTEMD
+	const char *systemd_state_dir;
+	if ((systemd_state_dir = getenv("STATE_DIRECTORY")) != NULL)
+		storage_base_dir = systemd_state_dir;
+#endif
+	storage_init(storage_base_dir);
+
 	/* In order to receive EPIPE while writing to the pipe whose reading end
 	 * is closed, the SIGPIPE signal has to be handled. For more information
 	 * see the io_thread_write_pcm() function. */
@@ -589,5 +608,10 @@ int main(int argc, char **argv) {
 	for (size_t i = 0; i < ARRAYSIZE(config.adapters); i++)
 		ba_adapter_destroy(config.adapters[i]);
 
+	storage_destroy();
+	g_dbus_connection_close_sync(config.dbus, NULL, NULL);
+	g_main_loop_unref(loop);
+	g_free(address);
+
 	return retval;
 }
