@@ -1,7 +1,7 @@
 /*
  * BlueALSA - ofono-iface.h
- * Copyright (c) 2016-2020 Arkadiusz Bokowy
- *               2018 Thierry Bultel
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
+ * Copyright (c) 2018 Thierry Bultel
  *
  * This file is a part of bluez-alsa.
  *
@@ -20,6 +20,7 @@
 #define OFONO_IFACE_HF_AUDIO_AGENT   OFONO_SERVICE ".HandsfreeAudioAgent"
 #define OFONO_IFACE_HF_AUDIO_CARD    OFONO_SERVICE ".HandsfreeAudioCard"
 #define OFONO_IFACE_HF_AUDIO_MANAGER OFONO_SERVICE ".HandsfreeAudioManager"
+#define OFONO_IFACE_CALL_VOLUME      OFONO_SERVICE ".CallVolume"
 
 #define OFONO_AUDIO_CARD_TYPE_AG "gateway"
 #define OFONO_AUDIO_CARD_TYPE_HF "handsfree"
@@ -27,6 +28,10 @@
 #define OFONO_AUDIO_CODEC_CVSD 0x01
 #define OFONO_AUDIO_CODEC_MSBC 0x02
 
-extern const GDBusInterfaceInfo ofono_iface_hf_audio_agent;
+#define OFONO_MODEM_TYPE_HARDWARE "hardware"
+#define OFONO_MODEM_TYPE_HFP      "hfp"
+#define OFONO_MODEM_TYPE_SAP      "sap"
+
+extern const GDBusInterfaceInfo org_ofono_handsfree_audio_agent_interface;
 
 #endif
