@@ -1,6 +1,6 @@
 /*
  * BlueALSA - sco.c
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -9,10 +9,12 @@
  */
 
 #include "sco.h"
+/* IWYU pragma: no_include "config.h" */
 
 #include <errno.h>
 #include <poll.h>
 #include <pthread.h>
+#include <signal.h>
 #include <stdbool.h>
 #include <stdint.h>
 #include <string.h>
@@ -24,6 +26,8 @@
 #include <bluetooth/hci_lib.h>
 #include <bluetooth/sco.h>
 
+#include <glib.h>
+
 #include "ba-device.h"
 #include "bluealsa-config.h"
 #if ENABLE_MSBC
@@ -33,7 +37,7 @@
 #include "hci.h"
 #include "hfp.h"
 #include "io.h"
-#include "utils.h"
+#include "shared/bluetooth.h"
 #include "shared/defs.h"
 #include "shared/ffb.h"
 #include "shared/log.h"
@@ -59,6 +63,12 @@ static void *sco_dispatcher_thread(struct ba_adapter *a) {
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 	pthread_cleanup_push(PTHREAD_CLEANUP(sco_dispatcher_cleanup), &data);
 
+	sigset_t sigset;
+	/* See the ba_transport_thread_create() function for information
+	 * why we have to mask all signals. */
+	sigfillset(&sigset);
+	pthread_sigmask(SIG_SETMASK, &sigset, NULL);
+
 	if ((data.pfd.fd = hci_sco_open(data.a->hci.dev_id)) == -1) {
 		error("Couldn't open SCO socket: %s", strerror(errno));
 		goto fail;
@@ -81,20 +91,21 @@ static void *sco_dispatcher_thread(struct ba_adapter *a) {
 	for (;;) {
 
 		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
+		int poll_rv = poll(&data.pfd, 1, -1);
+		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 
-		if (poll(&data.pfd, 1, -1) == -1) {
+		if (poll_rv == -1) {
 			if (errno == EINTR)
 				continue;
 			error("SCO dispatcher poll error: %s", strerror(errno));
 			goto fail;
 		}
 
-		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
-
 		struct sockaddr_sco addr;
 		socklen_t addrlen = sizeof(addr);
 		struct ba_device *d = NULL;
 		struct ba_transport *t = NULL;
+		char addrstr[18];
 		int fd = -1;
 
 		if ((fd = accept(data.pfd.fd, (struct sockaddr *)&addr, &addrlen)) == -1) {
@@ -102,10 +113,11 @@ static void *sco_dispatcher_thread(struct ba_adapter *a) {
 			goto cleanup;
 		}
 
-		debug("New incoming SCO link: %s: %d", batostr_(&addr.sco_bdaddr), fd);
+		ba2str(&addr.sco_bdaddr, addrstr);
+		debug("New incoming SCO link: %s: %d", addrstr, fd);
 
 		if ((d = ba_device_lookup(data.a, &addr.sco_bdaddr)) == NULL) {
-			error("Couldn't lookup device: %s", batostr_(&addr.sco_bdaddr));
+			error("Couldn't lookup device: %s", addrstr);
 			goto cleanup;
 		}
 
@@ -116,7 +128,7 @@ static void *sco_dispatcher_thread(struct ba_adapter *a) {
 
 #if ENABLE_MSBC
 		struct bt_voice voice = { .setting = BT_VOICE_TRANSPARENT };
-		if (t->type.codec == HFP_CODEC_MSBC &&
+		if (ba_transport_get_codec(t) == HFP_CODEC_MSBC &&
 				setsockopt(fd, SOL_BLUETOOTH, BT_VOICE, &voice, sizeof(voice)) == -1) {
 			error("Couldn't setup transparent voice: %s", strerror(errno));
 			goto cleanup;
@@ -132,11 +144,13 @@ static void *sco_dispatcher_thread(struct ba_adapter *a) {
 		pthread_mutex_lock(&t->bt_fd_mtx);
 
 		t->bt_fd = fd;
-		t->mtu_read = t->mtu_write = hci_sco_get_mtu(fd, a->hci.type);
+		t->mtu_read = t->mtu_write = hci_sco_get_mtu(fd, a);
 		fd = -1;
 
 		pthread_mutex_unlock(&t->bt_fd_mtx);
 
+		ba_transport_thread_state_set_idle(&t->thread_enc);
+		ba_transport_thread_state_set_idle(&t->thread_dec);
 		ba_transport_start(t);
 
 cleanup:
@@ -150,7 +164,6 @@ cleanup:
 	}
 
 fail:
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 	pthread_cleanup_pop(1);
 	return NULL;
 }
@@ -196,7 +209,7 @@ int sco_setup_connection_dispatcher(struct ba_adapter *a) {
 	 * during the whole live-span of the thread, because the thread is canceled
 	 * in the adapter cleanup routine. See the ba_adapter_unref() function. */
 	if ((ret = pthread_create(&a->sco_dispatcher, NULL,
-					PTHREAD_ROUTINE(sco_dispatcher_thread), a)) != 0) {
+					PTHREAD_FUNC(sco_dispatcher_thread), a)) != 0) {
 		error("Couldn't create SCO dispatcher: %s", strerror(ret));
 		a->sco_dispatcher = config.main_thread;
 		return -1;
@@ -214,7 +227,7 @@ static void *sco_cvsd_enc_thread(struct ba_transport_thread *th) {
 	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
 
 	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *pcm = &t->sco.spk_pcm;
+	struct ba_transport_pcm *t_pcm = th->pcm;
 	struct io_poll io = { .timeout = -1 };
 
 	const size_t mtu_samples = t->mtu_write / sizeof(int16_t);
@@ -230,14 +243,19 @@ static void *sco_cvsd_enc_thread(struct ba_transport_thread *th) {
 	}
 
 	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_set_state_running(th);;) {
+	for (ba_transport_thread_state_set_running(th);;) {
 
 		ssize_t samples = ffb_len_in(&buffer);
-		if ((samples = io_poll_and_read_pcm(&io, pcm, buffer.tail, samples)) <= 0) {
-			if (samples == -1)
-				error("PCM poll and read error: %s", strerror(errno));
-			else if (samples == 0)
-				ba_transport_stop_if_no_clients(t);
+		switch (samples = io_poll_and_read_pcm(&io, t_pcm, buffer.tail, samples)) {
+		case -1:
+			if (errno == ESTALE) {
+				ffb_rewind(&buffer);
+				continue;
+			}
+			error("PCM poll and read error: %s", strerror(errno));
+			/* fall-through */
+		case 0:
+			ba_transport_stop_if_no_clients(t);
 			continue;
 		}
 
@@ -262,7 +280,7 @@ static void *sco_cvsd_enc_thread(struct ba_transport_thread *th) {
 			/* keep data transfer at a constant bit rate */
 			asrsync_sync(&io.asrs, mtu_samples);
 			/* update busy delay (encoding overhead) */
-			pcm->delay = asrsync_get_busy_usec(&io.asrs) / 100;
+			t_pcm->delay = asrsync_get_busy_usec(&io.asrs) / 100;
 
 		}
 
@@ -272,8 +290,6 @@ static void *sco_cvsd_enc_thread(struct ba_transport_thread *th) {
 
 exit:
 	debug_transport_thread_loop(th, "EXIT");
-	ba_transport_thread_set_state_stopping(th);
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 fail_init:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -286,7 +302,7 @@ static void *sco_cvsd_dec_thread(struct ba_transport_thread *th) {
 	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
 
 	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *pcm = &t->sco.mic_pcm;
+	struct ba_transport_pcm *t_pcm = th->pcm;
 	struct io_poll io = { .timeout = -1 };
 
 	const size_t mtu_samples = t->mtu_read / sizeof(int16_t);
@@ -301,7 +317,7 @@ static void *sco_cvsd_dec_thread(struct ba_transport_thread *th) {
 	}
 
 	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_set_state_running(th);;) {
+	for (ba_transport_thread_state_set_running(th);;) {
 
 		ssize_t len = ffb_blen_in(&buffer);
 		if ((len = io_poll_and_read_bt(&io, th, buffer.tail, len)) == -1)
@@ -311,27 +327,33 @@ static void *sco_cvsd_dec_thread(struct ba_transport_thread *th) {
 
 		if ((size_t)len == buffer.nmemb * buffer.size) {
 			debug("Resizing CVSD read buffer: %zd -> %zd",
-					buffer.nmemb * buffer.size, buffer.nmemb * buffer.size * 2);
+					buffer.nmemb * buffer.size, buffer.nmemb * 2 * buffer.size);
 			if (ffb_init_int16_t(&buffer, buffer.nmemb * 2) == -1)
 				error("Couldn't resize CVSD read buffer: %s", strerror(errno));
 		}
 
-		if (!ba_transport_pcm_is_active(pcm))
+		if (!ba_transport_pcm_is_active(t_pcm))
 			continue;
 
-		ssize_t samples = len / sizeof(int16_t);
-		io_pcm_scale(pcm, buffer.data, samples);
-		if ((samples = io_pcm_write(pcm, buffer.data, samples)) == -1)
+		if (len > 0)
+			ffb_seek(&buffer, len / buffer.size);
+
+		ssize_t samples;
+		if ((samples = ffb_len_out(&buffer)) <= 0)
+			continue;
+
+		io_pcm_scale(t_pcm, buffer.data, samples);
+		if ((samples = io_pcm_write(t_pcm, buffer.data, samples)) == -1)
 			error("FIFO write error: %s", strerror(errno));
 		else if (samples == 0)
 			ba_transport_stop_if_no_clients(t);
 
+		ffb_shift(&buffer, samples);
+
 	}
 
 exit:
 	debug_transport_thread_loop(th, "EXIT");
-	ba_transport_thread_set_state_stopping(th);
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 fail_ffb:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -345,7 +367,7 @@ static void *sco_msbc_enc_thread(struct ba_transport_thread *th) {
 	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
 
 	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *pcm = &t->sco.spk_pcm;
+	struct ba_transport_pcm *t_pcm = th->pcm;
 	struct io_poll io = { .timeout = -1 };
 	const size_t mtu_write = t->mtu_write;
 
@@ -358,14 +380,20 @@ static void *sco_msbc_enc_thread(struct ba_transport_thread *th) {
 	}
 
 	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_set_state_running(th);;) {
+	for (ba_transport_thread_state_set_running(th);;) {
 
 		ssize_t samples = ffb_len_in(&msbc.pcm);
-		if ((samples = io_poll_and_read_pcm(&io, pcm, msbc.pcm.tail, samples)) <= 0) {
-			if (samples == -1)
-				error("PCM poll and read error: %s", strerror(errno));
-			else if (samples == 0)
-				ba_transport_stop_if_no_clients(t);
+		switch (samples = io_poll_and_read_pcm(&io, t_pcm, msbc.pcm.tail, samples)) {
+		case -1:
+			if (errno == ESTALE) {
+				/* reinitialize mSBC encoder */
+				msbc_init(&msbc);
+				continue;
+			}
+			error("PCM poll and read error: %s", strerror(errno));
+			/* fall-through */
+		case 0:
+			ba_transport_stop_if_no_clients(t);
 			continue;
 		}
 
@@ -399,10 +427,10 @@ static void *sco_msbc_enc_thread(struct ba_transport_thread *th) {
 			/* keep data transfer at a constant bit rate */
 			asrsync_sync(&io.asrs, msbc.frames * MSBC_CODESAMPLES);
 			/* update busy delay (encoding overhead) */
-			pcm->delay = asrsync_get_busy_usec(&io.asrs) / 100;
+			t_pcm->delay = asrsync_get_busy_usec(&io.asrs) / 100;
 
 			/* Move unprocessed data to the front of our linear
-			* buffer and clear the mSBC frame counter. */
+			 * buffer and clear the mSBC frame counter. */
 			ffb_shift(&msbc.data, ffb_blen_out(&msbc.data) - data_len);
 			msbc.frames = 0;
 
@@ -412,8 +440,6 @@ static void *sco_msbc_enc_thread(struct ba_transport_thread *th) {
 
 exit:
 	debug_transport_thread_loop(th, "EXIT");
-	ba_transport_thread_set_state_stopping(th);
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 fail_msbc:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -428,7 +454,7 @@ static void *sco_msbc_dec_thread(struct ba_transport_thread *th) {
 	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
 
 	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *pcm = &t->sco.mic_pcm;
+	struct ba_transport_pcm *t_pcm = th->pcm;
 	struct io_poll io = { .timeout = -1 };
 
 	struct esco_msbc msbc = { .initialized = false };
@@ -440,7 +466,7 @@ static void *sco_msbc_dec_thread(struct ba_transport_thread *th) {
 	}
 
 	debug_transport_thread_loop(th, "START");
-	for (ba_transport_thread_set_state_running(th);;) {
+	for (ba_transport_thread_state_set_running(th);;) {
 
 		ssize_t len = ffb_blen_in(&msbc.data);
 		if ((len = io_poll_and_read_bt(&io, th, msbc.data.tail, len)) == -1)
@@ -448,10 +474,11 @@ static void *sco_msbc_dec_thread(struct ba_transport_thread *th) {
 		else if (len == 0)
 			goto exit;
 
-		if (!ba_transport_pcm_is_active(pcm))
+		if (!ba_transport_pcm_is_active(t_pcm))
 			continue;
 
-		ffb_seek(&msbc.data, len);
+		if (len > 0)
+			ffb_seek(&msbc.data, len);
 
 		int err;
 		if ((err = msbc_decode(&msbc)) < 0) {
@@ -463,8 +490,8 @@ static void *sco_msbc_dec_thread(struct ba_transport_thread *th) {
 		if ((samples = ffb_len_out(&msbc.pcm)) <= 0)
 			continue;
 
-		io_pcm_scale(pcm, msbc.pcm.data, samples);
-		if ((samples = io_pcm_write(pcm, msbc.pcm.data, samples)) == -1)
+		io_pcm_scale(t_pcm, msbc.pcm.data, samples);
+		if ((samples = io_pcm_write(t_pcm, msbc.pcm.data, samples)) == -1)
 			error("FIFO write error: %s", strerror(errno));
 		else if (samples == 0)
 			ba_transport_stop_if_no_clients(t);
@@ -475,8 +502,6 @@ static void *sco_msbc_dec_thread(struct ba_transport_thread *th) {
 
 exit:
 	debug_transport_thread_loop(th, "EXIT");
-	ba_transport_thread_set_state_stopping(th);
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 fail_msbc:
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
@@ -485,7 +510,7 @@ fail_msbc:
 #endif
 
 void *sco_enc_thread(struct ba_transport_thread *th) {
-	switch (th->t->type.codec) {
+	switch (ba_transport_get_codec(th->t)) {
 	case HFP_CODEC_CVSD:
 	default:
 		return sco_cvsd_enc_thread(th);
@@ -496,8 +521,9 @@ void *sco_enc_thread(struct ba_transport_thread *th) {
 	}
 }
 
+__attribute__ ((weak))
 void *sco_dec_thread(struct ba_transport_thread *th) {
-	switch (th->t->type.codec) {
+	switch (ba_transport_get_codec(th->t)) {
 	case HFP_CODEC_CVSD:
 	default:
 		return sco_cvsd_dec_thread(th);
@@ -507,3 +533,23 @@ void *sco_dec_thread(struct ba_transport_thread *th) {
 #endif
 	}
 }
+
+int sco_transport_start(struct ba_transport *t) {
+
+	int rv = 0;
+
+	if (t->profile & BA_TRANSPORT_PROFILE_MASK_AG) {
+		rv |= ba_transport_thread_create(&t->thread_enc, sco_enc_thread, "ba-sco-enc", true);
+		rv |= ba_transport_thread_create(&t->thread_dec, sco_dec_thread, "ba-sco-dec", false);
+		return rv;
+	}
+
+	if (t->profile & BA_TRANSPORT_PROFILE_MASK_HF) {
+		rv |= ba_transport_thread_create(&t->thread_dec, sco_dec_thread, "ba-sco-dec", true);
+		rv |= ba_transport_thread_create(&t->thread_enc, sco_enc_thread, "ba-sco-enc", false);
+		return rv;
+	}
+
+	g_assert_not_reached();
+	return -1;
+}
