@@ -1,6 +1,6 @@
 /*
  * BlueALSA - sco.h
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -20,8 +20,6 @@
 #include "ba-transport.h"
 
 int sco_setup_connection_dispatcher(struct ba_adapter *a);
-
-void *sco_enc_thread(struct ba_transport_thread *th);
-void *sco_dec_thread(struct ba_transport_thread *th);
+int sco_transport_start(struct ba_transport *t);
 
 #endif
