@@ -240,13 +240,69 @@ dbus_bool_t bluealsa_dbus_connection_poll_dispatch(
 	return rv;
 }
 
+static dbus_bool_t bluealsa_dbus_props_get_all(
+		struct ba_dbus_ctx *ctx,
+		const char *path,
+		const char *interface,
+		DBusError *error,
+		dbus_bool_t (*cb)(const char *key, DBusMessageIter *val, void *data, DBusError *err),
+		void *userdata) {
+
+	DBusMessage *msg = NULL, *rep = NULL;
+	dbus_bool_t rv = FALSE;
+
+	if ((msg = dbus_message_new_method_call(ctx->ba_service, path,
+					DBUS_INTERFACE_PROPERTIES, "GetAll")) == NULL) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		goto fail;
+	}
+
+	DBusMessageIter iter;
+	dbus_message_iter_init_append(msg, &iter);
+	if (!dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &interface)) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		goto fail;
+	}
+
+	if ((rep = dbus_connection_send_with_reply_and_block(ctx->conn,
+					msg, DBUS_TIMEOUT_USE_DEFAULT, error)) == NULL)
+		goto fail;
+
+	if (!dbus_message_iter_init(rep, &iter)) {
+		dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE, "Empty response message");
+		goto fail;
+	}
+
+	if (!bluealsa_dbus_message_iter_dict(&iter, error, cb, userdata))
+		goto fail;
+
+	rv = TRUE;
+
+fail:
+	if (rep != NULL)
+		dbus_message_unref(rep);
+	if (msg != NULL)
+		dbus_message_unref(msg);
+	return rv;
+}
+
 /**
- * Callback function for BlueALSA service properties parser. */
-static dbus_bool_t bluealsa_dbus_message_iter_get_props_cb(const char *key,
+ * Callback function for manager object properties parser. */
+static dbus_bool_t bluealsa_dbus_message_iter_get_manager_props_cb(const char *key,
 		DBusMessageIter *value, void *userdata, DBusError *error) {
 	struct ba_service_props *props = (struct ba_service_props *)userdata;
 
-	char type = dbus_message_iter_get_arg_type(value);
+	char type;
+	if ((type = dbus_message_iter_get_arg_type(value)) != DBUS_TYPE_VARIANT) {
+		dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+				"Incorrect property value type: %c != %c", type, DBUS_TYPE_VARIANT);
+		return FALSE;
+	}
+
+	DBusMessageIter variant;
+	dbus_message_iter_recurse(value, &variant);
+	type = dbus_message_iter_get_arg_type(&variant);
+
 	char type_expected;
 
 	if (strcmp(key, "Version") == 0) {
@@ -255,7 +311,7 @@ static dbus_bool_t bluealsa_dbus_message_iter_get_props_cb(const char *key,
 			goto fail;
 
 		const char *tmp;
-		dbus_message_iter_get_basic(value, &tmp);
+		dbus_message_iter_get_basic(&variant, &tmp);
 		strncpy(props->version, tmp, sizeof(props->version) - 1);
 
 	}
@@ -266,7 +322,7 @@ static dbus_bool_t bluealsa_dbus_message_iter_get_props_cb(const char *key,
 
 		const char *tmp[ARRAYSIZE(props->adapters)];
 		size_t length = ARRAYSIZE(tmp);
-		if (!bluealsa_dbus_message_iter_array_get_strings(value, error, tmp, &length))
+		if (!bluealsa_dbus_message_iter_array_get_strings(&variant, error, tmp, &length))
 			return FALSE;
 
 		props->adapters_len = MIN(length, ARRAYSIZE(tmp));
@@ -281,7 +337,7 @@ static dbus_bool_t bluealsa_dbus_message_iter_get_props_cb(const char *key,
 
 		const char *tmp[32];
 		size_t length = ARRAYSIZE(tmp);
-		if (!bluealsa_dbus_message_iter_array_get_strings(value, error, tmp, &length))
+		if (!bluealsa_dbus_message_iter_array_get_strings(&variant, error, tmp, &length))
 			return FALSE;
 
 		props->profiles = malloc(length * sizeof(*props->profiles));
@@ -297,7 +353,7 @@ static dbus_bool_t bluealsa_dbus_message_iter_get_props_cb(const char *key,
 
 		const char *tmp[64];
 		size_t length = ARRAYSIZE(tmp);
-		if (!bluealsa_dbus_message_iter_array_get_strings(value, error, tmp, &length))
+		if (!bluealsa_dbus_message_iter_array_get_strings(&variant, error, tmp, &length))
 			return FALSE;
 
 		props->codecs = malloc(length * sizeof(*props->codecs));
@@ -325,49 +381,14 @@ dbus_bool_t bluealsa_dbus_get_props(
 		struct ba_service_props *props,
 		DBusError *error) {
 
-	static const char *interface = BLUEALSA_INTERFACE_MANAGER;
-	DBusMessage *msg = NULL, *rep = NULL;
-	dbus_bool_t rv = FALSE;
-
 	props->profiles = NULL;
 	props->profiles_len = 0;
 	props->codecs = NULL;
 	props->codecs_len = 0;
 
-	if ((msg = dbus_message_new_method_call(ctx->ba_service, "/org/bluealsa",
-					DBUS_INTERFACE_PROPERTIES, "GetAll")) == NULL) {
-		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
-		goto fail;
-	}
-
-	DBusMessageIter iter;
-	dbus_message_iter_init_append(msg, &iter);
-	if (!dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &interface)) {
-		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
-		goto fail;
-	}
-
-	if ((rep = dbus_connection_send_with_reply_and_block(ctx->conn,
-					msg, DBUS_TIMEOUT_USE_DEFAULT, error)) == NULL)
-		goto fail;
-
-	if (!dbus_message_iter_init(rep, &iter)) {
-		dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE, "Empty response message");
-		goto fail;
-	}
-
-	if (!bluealsa_dbus_message_iter_dict(&iter, error,
-				bluealsa_dbus_message_iter_get_props_cb, props))
-		goto fail;
-
-	rv = TRUE;
-
-fail:
-	if (rep != NULL)
-		dbus_message_unref(rep);
-	if (msg != NULL)
-		dbus_message_unref(msg);
-	return rv;
+	return bluealsa_dbus_props_get_all(ctx,
+			"/org/bluealsa", BLUEALSA_INTERFACE_MANAGER, error,
+			bluealsa_dbus_message_iter_get_manager_props_cb, props);
 }
 
 /**
@@ -388,6 +409,101 @@ void bluealsa_dbus_props_free(
 	}
 }
 
+/**
+ * Callback function for rfcomm object properties parser. */
+static dbus_bool_t bluealsa_dbus_message_iter_get_rfcomm_props_cb(const char *key,
+		DBusMessageIter *value, void *userdata, DBusError *error) {
+	struct ba_rfcomm_props *props = (struct ba_rfcomm_props *)userdata;
+
+	char type;
+	if ((type = dbus_message_iter_get_arg_type(value)) != DBUS_TYPE_VARIANT) {
+		dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+				"Incorrect property value type: %c != %c", type, DBUS_TYPE_VARIANT);
+		return FALSE;
+	}
+
+	DBusMessageIter variant;
+	dbus_message_iter_recurse(value, &variant);
+	type = dbus_message_iter_get_arg_type(&variant);
+
+	char type_expected;
+
+	if (strcmp(key, "Transport") == 0) {
+
+		if (type != (type_expected = DBUS_TYPE_STRING))
+			goto fail;
+
+		const char *tmp;
+		dbus_message_iter_get_basic(&variant, &tmp);
+		strncpy(props->transport, tmp, sizeof(props->transport) - 1);
+
+	}
+	else if (strcmp(key, "Features") == 0) {
+
+		if (type != (type_expected = DBUS_TYPE_ARRAY))
+			goto fail;
+
+		const char *tmp[32];
+		size_t length = ARRAYSIZE(tmp);
+		if (!bluealsa_dbus_message_iter_array_get_strings(&variant, error, tmp, &length))
+			return FALSE;
+
+		props->features = malloc(length * sizeof(*props->features));
+		props->features_len = MIN(length, ARRAYSIZE(tmp));
+		for (size_t i = 0; i < length; i++)
+			props->features[i] = strdup(tmp[i]);
+
+	}
+	else if (strcmp(key, "Battery") == 0) {
+
+		if (type != (type_expected = DBUS_TYPE_BYTE))
+			goto fail;
+
+		signed char level;
+		dbus_message_iter_get_basic(&variant, &level);
+		props->battery = level;
+
+	}
+
+	return TRUE;
+
+fail:
+	dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+			"Incorrect variant for '%s': %c != %c", key, type, type_expected);
+	return FALSE;
+}
+
+/**
+ * Get properties of BlueALSA RFCOMM object.
+ *
+ * This function allocates resources within the properties structure, which
+ * shall be freed with the bluealsa_dbus_rfcomm_props_free() function. */
+dbus_bool_t bluealsa_dbus_get_rfcomm_props(
+		struct ba_dbus_ctx *ctx,
+		const char *rfcomm_path,
+		struct ba_rfcomm_props *props,
+		DBusError *error) {
+
+	props->features = NULL;
+	props->features_len = 0;
+
+	return bluealsa_dbus_props_get_all(ctx,
+			rfcomm_path, BLUEALSA_INTERFACE_RFCOMM, error,
+			bluealsa_dbus_message_iter_get_rfcomm_props_cb, props);
+}
+
+/**
+ * Free BlueALSA RFCOMM properties structure. */
+void bluealsa_dbus_rfcomm_props_free(
+		struct ba_rfcomm_props *props) {
+	if (props->features != NULL) {
+		for (size_t i = 0; i < props->features_len; i++)
+			free(props->features[i]);
+		free(props->features);
+		props->features = NULL;
+	}
+}
+
 dbus_bool_t bluealsa_dbus_get_pcms(
 		struct ba_dbus_ctx *ctx,
 		struct ba_pcm **pcms,
@@ -565,7 +681,133 @@ const char *bluealsa_dbus_pcm_get_codec_canonical_name(
 }
 
 /**
- * Select BlueALSA PCM Bluetooth codec. */
+ * Callback function for BlueALSA PCM codec props parser. */
+static dbus_bool_t bluealsa_dbus_message_iter_pcm_get_codec_props_cb(const char *key,
+		DBusMessageIter *value, void *userdata, DBusError *error) {
+	struct ba_pcm_codec *codec = (struct ba_pcm_codec *)userdata;
+
+	char type;
+	if ((type = dbus_message_iter_get_arg_type(value)) != DBUS_TYPE_VARIANT) {
+		dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+				"Incorrect property value type: %c != %c", type, DBUS_TYPE_VARIANT);
+		return FALSE;
+	}
+
+	DBusMessageIter variant;
+	dbus_message_iter_recurse(value, &variant);
+	type = dbus_message_iter_get_arg_type(&variant);
+
+	char type_expected;
+
+	if (strcmp(key, "Capabilities") == 0) {
+		if (type != (type_expected = DBUS_TYPE_ARRAY))
+			goto fail;
+
+		DBusMessageIter iter;
+		uint8_t *data;
+		int len;
+
+		dbus_message_iter_recurse(&variant, &iter);
+		dbus_message_iter_get_fixed_array(&iter, &data, &len);
+
+		codec->data_len = MIN(len, sizeof(codec->data));
+		memcpy(codec->data, data, codec->data_len);
+
+	}
+
+	return TRUE;
+
+fail:
+	dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+			"Incorrect variant for '%s': %c != %c", key, type, type_expected);
+	return FALSE;
+}
+
+/**
+ * Callback function for BlueALSA PCM codec list parser. */
+static dbus_bool_t bluealsa_dbus_message_iter_pcm_get_codecs_cb(const char *key,
+		DBusMessageIter *value, void *userdata, DBusError *error) {
+
+	struct ba_pcm_codecs *codecs = (struct ba_pcm_codecs *)userdata;
+	const size_t len = codecs->codecs_len;
+
+	struct ba_pcm_codec *tmp = codecs->codecs;
+	if ((tmp = realloc(tmp, (len + 1) * sizeof(*tmp))) == NULL) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		return FALSE;
+	}
+
+	struct ba_pcm_codec *codec = &tmp[len];
+	codecs->codecs = tmp;
+
+	memset(codec, 0, sizeof(*codec));
+	strncpy(codec->name, key, sizeof(codec->name));
+	codec->name[sizeof(codec->name) - 1] = '\0';
+
+	if (!bluealsa_dbus_message_iter_dict(value, error,
+				bluealsa_dbus_message_iter_pcm_get_codec_props_cb, codec))
+		return FALSE;
+
+	codecs->codecs_len = len + 1;
+	return TRUE;
+}
+
+/**
+ * Get BlueALSA PCM Bluetooth audio codecs. */
+dbus_bool_t bluealsa_dbus_pcm_get_codecs(
+		struct ba_dbus_ctx *ctx,
+		const char *pcm_path,
+		struct ba_pcm_codecs *codecs,
+		DBusError *error) {
+
+	DBusMessage *msg = NULL, *rep = NULL;
+	dbus_bool_t rv = FALSE;
+
+	if ((msg = dbus_message_new_method_call(ctx->ba_service, pcm_path,
+					BLUEALSA_INTERFACE_PCM, "GetCodecs")) == NULL) {
+		dbus_set_error(error, DBUS_ERROR_NO_MEMORY, NULL);
+		goto fail;
+	}
+
+	if ((rep = dbus_connection_send_with_reply_and_block(ctx->conn,
+					msg, DBUS_TIMEOUT_USE_DEFAULT, error)) == NULL)
+		goto fail;
+
+	DBusMessageIter iter;
+	if (!dbus_message_iter_init(rep, &iter)) {
+		dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE, "Empty response message");
+		goto fail;
+	}
+
+	codecs->codecs = NULL;
+	codecs->codecs_len = 0;
+
+	if (!bluealsa_dbus_message_iter_dict(&iter, error,
+				bluealsa_dbus_message_iter_pcm_get_codecs_cb, codecs)) {
+		free(codecs->codecs);
+		goto fail;
+	}
+
+	rv = TRUE;
+
+fail:
+	if (msg != NULL)
+		dbus_message_unref(msg);
+	if (rep != NULL)
+		dbus_message_unref(rep);
+	return rv;
+}
+
+/**
+ * Free BlueALSA PCM codecs structure. */
+void bluealsa_dbus_pcm_codecs_free(
+		struct ba_pcm_codecs *codecs) {
+	free(codecs->codecs);
+	codecs->codecs = NULL;
+}
+
+/**
+ * Select BlueALSA PCM Bluetooth audio codec. */
 dbus_bool_t bluealsa_dbus_pcm_select_codec(
 		struct ba_dbus_ctx *ctx,
 		const char *pcm_path,
@@ -809,7 +1051,6 @@ dbus_bool_t bluealsa_dbus_message_iter_dict(
 			dbus_message_iter_next(&iter_dict)) {
 
 		DBusMessageIter iter_entry;
-		DBusMessageIter iter_entry_val;
 		const char *key;
 
 		if (dbus_message_iter_get_arg_type(&iter_dict) != DBUS_TYPE_DICT_ENTRY)
@@ -818,12 +1059,10 @@ dbus_bool_t bluealsa_dbus_message_iter_dict(
 		if (dbus_message_iter_get_arg_type(&iter_entry) != DBUS_TYPE_STRING)
 			goto fail;
 		dbus_message_iter_get_basic(&iter_entry, &key);
-		if (!dbus_message_iter_next(&iter_entry) ||
-				dbus_message_iter_get_arg_type(&iter_entry) != DBUS_TYPE_VARIANT)
+		if (!dbus_message_iter_next(&iter_entry))
 			goto fail;
-		dbus_message_iter_recurse(&iter_entry, &iter_entry_val);
 
-		if (!cb(key, &iter_entry_val, userdata, error))
+		if (!cb(key, &iter_entry, userdata, error))
 			return FALSE;
 
 	}
@@ -833,7 +1072,7 @@ dbus_bool_t bluealsa_dbus_message_iter_dict(
 fail:
 	signature = dbus_message_iter_get_signature(iter);
 	dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
-			"Incorrect signature: %s != a{sv}", signature);
+			"Incorrect signature: %s != a{s#}", signature);
 	dbus_free(signature);
 	return FALSE;
 }
@@ -905,29 +1144,39 @@ fail:
 /**
  * Callback function for BlueALSA PCM properties parser. */
 static dbus_bool_t bluealsa_dbus_message_iter_get_pcm_props_cb(const char *key,
-		DBusMessageIter *variant, void *userdata, DBusError *error) {
+		DBusMessageIter *value, void *userdata, DBusError *error) {
 	struct ba_pcm *pcm = (struct ba_pcm *)userdata;
 
-	char type = dbus_message_iter_get_arg_type(variant);
+	char type;
+	if ((type = dbus_message_iter_get_arg_type(value)) != DBUS_TYPE_VARIANT) {
+		dbus_set_error(error, DBUS_ERROR_INVALID_SIGNATURE,
+				"Incorrect property value type: %c != %c", type, DBUS_TYPE_VARIANT);
+		return FALSE;
+	}
+
+	DBusMessageIter variant;
+	dbus_message_iter_recurse(value, &variant);
+	type = dbus_message_iter_get_arg_type(&variant);
+
 	char type_expected;
 	const char *tmp;
 
 	if (strcmp(key, "Device") == 0) {
 		if (type != (type_expected = DBUS_TYPE_OBJECT_PATH))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &tmp);
+		dbus_message_iter_get_basic(&variant, &tmp);
 		strncpy(pcm->device_path, tmp, sizeof(pcm->device_path) - 1);
 		path2ba(tmp, &pcm->addr);
 	}
 	else if (strcmp(key, "Sequence") == 0) {
 		if (type != (type_expected = DBUS_TYPE_UINT32))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &pcm->sequence);
+		dbus_message_iter_get_basic(&variant, &pcm->sequence);
 	}
 	else if (strcmp(key, "Transport") == 0) {
 		if (type != (type_expected = DBUS_TYPE_STRING))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &tmp);
+		dbus_message_iter_get_basic(&variant, &tmp);
 		if (strstr(tmp, "A2DP-source") != NULL)
 			pcm->transport = BA_PCM_TRANSPORT_A2DP_SOURCE;
 		else if (strstr(tmp, "A2DP-sink") != NULL)
@@ -944,47 +1193,67 @@ static dbus_bool_t bluealsa_dbus_message_iter_get_pcm_props_cb(const char *key,
 	else if (strcmp(key, "Mode") == 0) {
 		if (type != (type_expected = DBUS_TYPE_STRING))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &tmp);
+		dbus_message_iter_get_basic(&variant, &tmp);
 		if (strcmp(tmp, "source") == 0)
 			pcm->mode = BA_PCM_MODE_SOURCE;
 		else if (strcmp(tmp, "sink") == 0)
 			pcm->mode = BA_PCM_MODE_SINK;
 	}
+	else if (strcmp(key, "Running") == 0) {
+		if (type != (type_expected = DBUS_TYPE_BOOLEAN))
+			goto fail;
+		dbus_message_iter_get_basic(&variant, &pcm->running);
+	}
 	else if (strcmp(key, "Format") == 0) {
 		if (type != (type_expected = DBUS_TYPE_UINT16))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &pcm->format);
+		dbus_message_iter_get_basic(&variant, &pcm->format);
 	}
 	else if (strcmp(key, "Channels") == 0) {
 		if (type != (type_expected = DBUS_TYPE_BYTE))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &pcm->channels);
+		dbus_message_iter_get_basic(&variant, &pcm->channels);
 	}
 	else if (strcmp(key, "Sampling") == 0) {
 		if (type != (type_expected = DBUS_TYPE_UINT32))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &pcm->sampling);
+		dbus_message_iter_get_basic(&variant, &pcm->sampling);
 	}
 	else if (strcmp(key, "Codec") == 0) {
 		if (type != (type_expected = DBUS_TYPE_STRING))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &tmp);
-		strncpy(pcm->codec, tmp, sizeof(pcm->codec) - 1);
+		dbus_message_iter_get_basic(&variant, &tmp);
+		strncpy(pcm->codec.name, tmp, sizeof(pcm->codec.name) - 1);
+	}
+	else if (strcmp(key, "CodecConfiguration") == 0) {
+		if (type != (type_expected = DBUS_TYPE_ARRAY))
+			goto fail;
+
+		DBusMessageIter iter;
+		uint8_t *data;
+		int len;
+
+		dbus_message_iter_recurse(&variant, &iter);
+		dbus_message_iter_get_fixed_array(&iter, &data, &len);
+
+		pcm->codec.data_len = MIN(len, sizeof(pcm->codec.data));
+		memcpy(pcm->codec.data, data, pcm->codec.data_len);
+
 	}
 	else if (strcmp(key, "Delay") == 0) {
 		if (type != (type_expected = DBUS_TYPE_UINT16))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &pcm->delay);
+		dbus_message_iter_get_basic(&variant, &pcm->delay);
 	}
 	else if (strcmp(key, "SoftVolume") == 0) {
 		if (type != (type_expected = DBUS_TYPE_BOOLEAN))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &pcm->soft_volume);
+		dbus_message_iter_get_basic(&variant, &pcm->soft_volume);
 	}
 	else if (strcmp(key, "Volume") == 0) {
 		if (type != (type_expected = DBUS_TYPE_UINT16))
 			goto fail;
-		dbus_message_iter_get_basic(variant, &pcm->volume.raw);
+		dbus_message_iter_get_basic(&variant, &pcm->volume.raw);
 	}
 
 	return TRUE;
