@@ -1,6 +1,6 @@
 /*
  * BlueALSA - log.c
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -9,6 +9,7 @@
  */
 
 #include "shared/log.h"
+/* IWYU pragma: no_include "config.h" */
 
 #include <pthread.h>
 #include <stdarg.h>
@@ -16,8 +17,10 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/time.h>
 #include <syslog.h>
 #include <time.h>
+#include <unistd.h>
 
 #if WITH_LIBUNWIND
 # define UNW_LOCAL_ONLY
@@ -33,33 +36,41 @@
 static char *_ident = NULL;
 /* if true, system logging is enabled */
 static bool _syslog = false;
-/* if true, print logging time */
-static bool _time = BLUEALSA_LOGTIME;
 
-void log_open(const char *ident, bool syslog, bool time) {
+#if DEBUG_TIME
 
-	free(_ident);
-	_ident = strdup(ident);
+/* point "zero" for relative time */
+static struct timespec _ts0;
+
+__attribute__ ((constructor))
+static void init_ts0(void) {
+	gettimestamp(&_ts0);
+}
+
+#endif
+
+void log_open(const char *ident, bool syslog) {
+
+	if (ident != NULL)
+		_ident = strdup(ident);
 
 	if ((_syslog = syslog) == true)
 		openlog(ident, 0, LOG_USER);
 
-	_time = time;
-
 }
 
-static void vlog(int priority, const char *format, va_list ap) {
+static const char *priority2str[] = {
+	[LOG_EMERG] = "X",
+	[LOG_ALERT] = "A",
+	[LOG_CRIT] = "C",
+	[LOG_ERR] = "E",
+	[LOG_WARNING] = "W",
+	[LOG_NOTICE] = "N",
+	[LOG_INFO] = "I",
+	[LOG_DEBUG] = "D",
+};
 
-	static const char *priority2str[] = {
-		[LOG_EMERG] = "X",
-		[LOG_ALERT] = "A",
-		[LOG_CRIT] = "C",
-		[LOG_ERR] = "E",
-		[LOG_WARNING] = "W",
-		[LOG_NOTICE] = "N",
-		[LOG_INFO] = "I",
-		[LOG_DEBUG] = "D",
-	};
+static void vlog(int priority, const char *format, va_list ap) {
 
 	int oldstate;
 
@@ -70,26 +81,41 @@ static void vlog(int priority, const char *format, va_list ap) {
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &oldstate);
 
 	if (_syslog) {
+
 		va_list ap_syslog;
 		va_copy(ap_syslog, ap);
 		vsyslog(priority, format, ap_syslog);
 		va_end(ap_syslog);
+
 	}
+	else {
 
-	flockfile(stderr);
-
-	if (_ident != NULL)
-		fprintf(stderr, "%s: ", _ident);
-	if (_time) {
+#if DEBUG_TIME
 		struct timespec ts;
 		gettimestamp(&ts);
-		fprintf(stderr, "%lu.%.9lu: ", (long int)ts.tv_sec, ts.tv_nsec);
-	}
-	fprintf(stderr, "%s: ", priority2str[priority]);
-	vfprintf(stderr, format, ap);
-	fputs("\n", stderr);
+		timespecsub(&ts, &_ts0, &ts);
+#endif
 
-	funlockfile(stderr);
+		flockfile(stderr);
+
+		if (_ident != NULL)
+			fprintf(stderr, "%s: ", _ident);
+
+#if DEBUG_TIME
+		fprintf(stderr, "%lu.%.9lu: ", (long int)ts.tv_sec, ts.tv_nsec);
+#endif
+
+#if DEBUG && HAVE_GETTID
+		fprintf(stderr, "[%d] ", gettid());
+#endif
+
+		fprintf(stderr, "%s: ", priority2str[priority]);
+		vfprintf(stderr, format, ap);
+		fputs("\n", stderr);
+
+		funlockfile(stderr);
+
+	}
 
 	pthread_setcancelstate(oldstate, NULL);
 
