@@ -17,11 +17,10 @@
 #endif
 
 #include <stdint.h>
-#include <sys/time.h>
 #include <time.h>
 
 #if HAVE_LIBBSD
-# include <bsd/sys/time.h>
+# include <bsd/sys/time.h> /* IWYU pragma: keep */
 #else
 # define timespecadd(ts_a, ts_b, dest) do { \
 		(dest)->tv_sec = (ts_a)->tv_sec + (ts_b)->tv_sec; \
@@ -56,7 +55,7 @@ struct asrsync {
 
 	/* time-stamp from the previous sync */
 	struct timespec ts;
-	/* transfered frames since ts0 */
+	/* transferred frames since ts0 */
 	uint32_t frames;
 
 	/* time spent outside of the sync function */
