@@ -1,6 +1,6 @@
 /*
  * test-alsa-pcm.c
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -17,13 +17,11 @@
 #include <getopt.h>
 #include <libgen.h>
 #include <poll.h>
-#include <signal.h>
 #include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <sys/wait.h>
 #include <time.h>
 #include <unistd.h>
 
@@ -34,9 +32,11 @@
 #include "shared/log.h"
 #include "shared/rt.h"
 
+#include "inc/check.inc"
+#include "inc/mock.inc"
 #include "inc/preload.inc"
-#include "inc/server.inc"
 #include "inc/sine.inc"
+#include "inc/spawn.inc"
 
 #define dumprv(fn) fprintf(stderr, #fn " = %d\n", (int)fn)
 
@@ -45,45 +45,7 @@ static unsigned int pcm_channels = 2;
 static unsigned int pcm_sampling = 44100;
 static snd_pcm_format_t pcm_format = SND_PCM_FORMAT_S16_LE;
 /* big enough buffer to keep one period of data */
-static int16_t buffer[1024 * 8];
-
-static int snd_pcm_open_bluealsa(
-		snd_pcm_t **pcmp,
-		const char *service,
-		const char *extra_config,
-		snd_pcm_stream_t stream,
-		int mode) {
-
-	char buffer[256];
-	snd_config_t *conf = NULL;
-	snd_input_t *input = NULL;
-	int err;
-
-	sprintf(buffer,
-			"pcm.bluealsa {\n"
-			"  type bluealsa\n"
-			"  service \"org.bluealsa.%s\"\n"
-			"  device \"12:34:56:78:9A:BC\"\n"
-			"  profile \"a2dp\"\n"
-			"  delay 0\n"
-			"  %s\n"
-			"}\n", service, extra_config);
-
-	if ((err = snd_config_top(&conf)) < 0)
-		goto fail;
-	if ((err = snd_input_buffer_open(&input, buffer, strlen(buffer))) != 0)
-		goto fail;
-	if ((err = snd_config_load(conf, input)) != 0)
-		goto fail;
-	err = snd_pcm_open_lconf(pcmp, "bluealsa", stream, mode, conf);
-
-fail:
-	if (conf != NULL)
-		snd_config_delete(conf);
-	if (input != NULL)
-		snd_input_close(input);
-	return err;
-}
+static int16_t pcm_buffer[1024 * 8];
 
 static int set_hw_params(snd_pcm_t *pcm, snd_pcm_format_t format, int channels,
 		int rate, unsigned int *buffer_time, unsigned int *period_time) {
@@ -159,36 +121,43 @@ static int set_sw_params(snd_pcm_t *pcm, snd_pcm_uframes_t buffer_size, snd_pcm_
 	return 0;
 }
 
-static int test_pcm_open(pid_t *pid, snd_pcm_t **pcm, snd_pcm_stream_t stream) {
+static int test_pcm_open(struct spawn_process *sp_ba_mock, snd_pcm_t **pcm,
+		snd_pcm_stream_t stream) {
 
 	if (pcm_device != NULL)
 		return snd_pcm_open(pcm, pcm_device, stream, 0);
 
-	const char *service = "test";
-	if ((*pid = spawn_bluealsa_server(service, 1, true, false,
-					stream == SND_PCM_STREAM_PLAYBACK,
-					stream == SND_PCM_STREAM_CAPTURE,
-					false, false)) == -1)
+	const char *profile = NULL;
+	if (stream == SND_PCM_STREAM_PLAYBACK)
+		profile = "--profile=a2dp-source";
+	if (stream == SND_PCM_STREAM_CAPTURE)
+		profile = "--profile=a2dp-sink";
+
+	if (spawn_bluealsa_mock(sp_ba_mock, NULL, true,
+				"--timeout=1000",
+				profile,
+				NULL) == -1)
 		return -1;
-	return snd_pcm_open_bluealsa(pcm, service, "", stream, 0);
+
+	return snd_pcm_open(pcm, "bluealsa:DEV=12:34:56:78:9A:BC", stream, 0);
 }
 
-static int test_pcm_close(pid_t pid, snd_pcm_t *pcm) {
+static int test_pcm_close(struct spawn_process *sp_ba_mock, snd_pcm_t *pcm) {
 	int rv = 0;
 	if (pcm != NULL)
 		rv = snd_pcm_close(pcm);
-	if (pid != -1) {
-		kill(pid, SIGTERM);
-		waitpid(pid, NULL, 0);
+	if (sp_ba_mock != NULL) {
+		spawn_terminate(sp_ba_mock, 0);
+		spawn_close(sp_ba_mock, NULL);
 	}
 	return rv;
 }
 
 static int16_t *test_sine_s16le(snd_pcm_uframes_t size) {
 	static size_t x = 0;
-	assert(ARRAYSIZE(buffer) >= size * pcm_channels);
-	x = snd_pcm_sine_s16_2le(buffer, size, pcm_channels, x, 441.0 / pcm_sampling);
-	return buffer;
+	assert(ARRAYSIZE(pcm_buffer) >= size * pcm_channels);
+	x = snd_pcm_sine_s16_2le(pcm_buffer, size, pcm_channels, x, 441.0 / pcm_sampling);
+	return pcm_buffer;
 }
 
 static snd_pcm_state_t snd_pcm_state_runtime(snd_pcm_t *pcm) {
@@ -200,35 +169,34 @@ static snd_pcm_state_t snd_pcm_state_runtime(snd_pcm_t *pcm) {
 	return snd_pcm_status_get_state(status);
 }
 
-START_TEST(dump_capture) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(dump_capture) {
 
 	snd_output_t *output;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 
 	ck_assert_int_eq(snd_output_stdio_attach(&output, stdout, 0), 0);
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_CAPTURE), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_CAPTURE), 0);
 
 	ck_assert_int_eq(snd_pcm_dump(pcm, output), 0);
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
+	ck_assert_int_eq(snd_output_close(output), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_capture_start) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(test_capture_start) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
 	snd_pcm_uframes_t buffer_size;
 	snd_pcm_uframes_t period_size;
 	snd_pcm_sframes_t delay;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 	size_t i;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_CAPTURE), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_CAPTURE), 0);
 	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
 				&buffer_time, &period_time), 0);
 	ck_assert_int_eq(snd_pcm_get_params(pcm, &buffer_size, &period_size), 0);
@@ -256,7 +224,7 @@ START_TEST(test_capture_start) {
 
 	/* read few periods from capture PCM */
 	for (i = 0; i < buffer_size / period_size; i++)
-		ck_assert_int_eq(snd_pcm_readi(pcm, buffer, period_size), period_size);
+		ck_assert_int_eq(snd_pcm_readi(pcm, pcm_buffer, period_size), period_size);
 
 	/* after reading there should be no more than one period of data in buffer */
 	snd_pcm_sframes_t avail;
@@ -265,21 +233,20 @@ START_TEST(test_capture_start) {
 	ck_assert_int_eq(snd_pcm_delay(pcm, &delay), 0);
 	ck_assert_int_ge(delay, avail);
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_capture_pause) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(test_capture_pause) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
 	snd_pcm_uframes_t buffer_size;
 	snd_pcm_uframes_t period_size;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_CAPTURE), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_CAPTURE), 0);
 	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
 				&buffer_time, &period_time), 0);
 	ck_assert_int_eq(snd_pcm_get_params(pcm, &buffer_size, &period_size), 0);
@@ -329,22 +296,21 @@ START_TEST(test_capture_pause) {
 
 	}
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_capture_overrun) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(test_capture_overrun) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
 	snd_pcm_uframes_t buffer_size;
 	snd_pcm_uframes_t period_size;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 	size_t i;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_CAPTURE), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_CAPTURE), 0);
 	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
 				&buffer_time, &period_time), 0);
 	ck_assert_int_eq(snd_pcm_get_params(pcm, &buffer_size, &period_size), 0);
@@ -353,7 +319,7 @@ START_TEST(test_capture_overrun) {
 
 	/* check that PCM is running and we can read from it */
 	ck_assert_int_eq(snd_pcm_state_runtime(pcm), SND_PCM_STATE_RUNNING);
-	ck_assert_int_eq(snd_pcm_readi(pcm, buffer, period_size), period_size);
+	ck_assert_int_eq(snd_pcm_readi(pcm, pcm_buffer, period_size), period_size);
 
 	/* allow overrun to occur */
 	usleep(buffer_time + period_time);
@@ -372,21 +338,20 @@ START_TEST(test_capture_overrun) {
 
 	/* make sure that PCM is indeed readable */
 	for (i = 0; i < buffer_size / period_size; i++)
-		ck_assert_int_eq(snd_pcm_readi(pcm, buffer, period_size), period_size);
+		ck_assert_int_eq(snd_pcm_readi(pcm, pcm_buffer, period_size), period_size);
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_capture_poll) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(test_capture_poll) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_CAPTURE), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_CAPTURE), 0);
 	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
 				&buffer_time, &period_time), 0);
 
@@ -394,16 +359,19 @@ START_TEST(test_capture_poll) {
 	unsigned short revents;
 	int count = snd_pcm_poll_descriptors_count(pcm);
 	ck_assert_int_eq(snd_pcm_poll_descriptors(pcm, pfds, ARRAYSIZE(pfds)), count);
+	int rv;
 
 	ck_assert_int_eq(snd_pcm_prepare(pcm), 0);
-	/* for a capture PCM just after prepare, the poll() call shall block
-	 * forever or at least the dispatched event shall be set to 0 */
-	ck_assert_int_ne(poll(pfds, count, 250), -1);
-	snd_pcm_poll_descriptors_revents(pcm, pfds, count, &revents);
-	ck_assert_int_eq(revents, 0);
-
-	/* make sure that further calls to poll() will actually block */
-	ck_assert_int_eq(poll(pfds, count, 250), 0);
+	/* For a capture PCM just after prepare, the poll() call shall block
+	 * forever or at least the dispatched event shall be set to 0. */
+	for (;;) {
+		ck_assert_int_ne(rv = poll(pfds, count, 750), -1);
+		/* make sure that at some point poll() will actually block */
+		if (rv == 0)
+			break;
+		snd_pcm_poll_descriptors_revents(pcm, pfds, count, &revents);
+		ck_assert_int_eq(revents, 0);
+	}
 
 	ck_assert_int_eq(snd_pcm_start(pcm), 0);
 	do { /* started capture PCM shall not block forever */
@@ -413,21 +381,20 @@ START_TEST(test_capture_poll) {
 	/* we should get read event flag set */
 	ck_assert_int_eq(revents & POLLIN, POLLIN);
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(dump_playback) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(dump_playback) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
 	snd_output_t *output;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 
 	ck_assert_int_eq(snd_output_stdio_attach(&output, stdout, 0), 0);
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
 
 	ck_assert_int_eq(snd_pcm_dump(pcm, output), 0);
 
@@ -445,27 +412,49 @@ START_TEST(dump_playback) {
 	dumprv(snd_pcm_hw_params_can_resume(params));
 	dumprv(snd_pcm_hw_params_can_sync_start(params));
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
+	ck_assert_int_eq(snd_output_close(output), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(ba_test_playback_hw_constraints) {
+CK_START_TEST(ba_test_playback_hw_constraints) {
 
 	if (pcm_device != NULL)
 		return;
 
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
-
 	/* hard-coded values used in the bluealsa-mock */
 	const unsigned int server_channels = 2;
 	const unsigned int server_rate = 44100;
 
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	snd_pcm_hw_params_t *params;
-	pid_t pid = -1;
-	int d;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
+	ck_assert_int_ne(spawn_bluealsa_mock(&sp_ba_mock, NULL, true,
+				"--timeout=1000",
+				"--profile=a2dp-source",
+				NULL), -1);
+
+	snd_config_t *top;
+	ck_assert_int_ge(snd_config_top(&top), 0);
+
+	const char *config =
+		"pcm.ba-direct {\n"
+		"  type bluealsa\n"
+		"  device \"12:34:56:78:9A:BC\"\n"
+		"  profile \"a2dp\"\n"
+		"}\n";
+	snd_input_t *input;
+	ck_assert_int_eq(snd_input_buffer_open(&input, config, strlen(config)), 0);
+	ck_assert_int_eq(snd_config_load(top, input), 0);
+
+	ck_assert_int_eq(snd_pcm_open_lconf(&pcm,
+				"ba-direct", SND_PCM_STREAM_PLAYBACK, 0, top), 0);
+
+	snd_config_delete(top);
+	snd_input_close(input);
+
+	snd_pcm_hw_params_t *params;
+	int d;
 
 	snd_pcm_hw_params_alloca(&params);
 	snd_pcm_hw_params_any(pcm, params);
@@ -520,46 +509,114 @@ START_TEST(ba_test_playback_hw_constraints) {
 	ck_assert_int_eq(time, 11888616);
 	ck_assert_int_eq(d, 1);
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(ba_test_playback_extra_setup) {
+CK_START_TEST(ba_test_playback_no_codec_selected) {
 
 	if (pcm_device != NULL)
 		return;
 
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
-
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 
-	const char *service = "test";
-	ck_assert_int_ne(pid = spawn_bluealsa_server(service, 1, true, false, true, false, false, false), -1);
+	ck_assert_int_ne(spawn_bluealsa_mock(&sp_ba_mock, NULL, true,
+				"--timeout=1000",
+				"--profile=hfp-ag",
+				NULL), -1);
 
-	ck_assert_int_eq(snd_pcm_open_bluealsa(&pcm, service, "codec \"sbc\"", SND_PCM_STREAM_PLAYBACK, 0), 0);
-	ck_assert_int_eq(test_pcm_close(-1, pcm), 0);
+	int rv = 0;
+#if ENABLE_MSBC
+	rv = -EAGAIN;
+#endif
 
-	ck_assert_int_eq(snd_pcm_open_bluealsa(&pcm, service, "volume \"50+\"", SND_PCM_STREAM_PLAYBACK, 0), 0);
-	ck_assert_int_eq(test_pcm_close(-1, pcm), 0);
+	ck_assert_int_eq(snd_pcm_open(&pcm,
+				"bluealsa:DEV=12:34:56:78:9A:BC,PROFILE=sco",
+				SND_PCM_STREAM_PLAYBACK, 0), rv);
 
-	ck_assert_int_eq(snd_pcm_open_bluealsa(&pcm, service, "softvol true", SND_PCM_STREAM_PLAYBACK, 0), 0);
-	ck_assert_int_eq(test_pcm_close(-1, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-	ck_assert_int_eq(test_pcm_close(pid, NULL), 0);
+} CK_END_TEST
 
-} END_TEST
+CK_START_TEST(ba_test_playback_no_such_device) {
 
-START_TEST(test_playback_hw_set_free) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+	if (pcm_device != NULL)
+		return;
+
+	struct spawn_process sp_ba_mock;
+	snd_pcm_t *pcm = NULL;
+
+	ck_assert_int_ne(spawn_bluealsa_mock(&sp_ba_mock, "test", true,
+				"--timeout=1000",
+				NULL), -1);
+
+	ck_assert_int_eq(snd_pcm_open(&pcm,
+				"bluealsa:DEV=DE:AD:DE:AD:DE:AD,SRV=org.bluealsa.test",
+				SND_PCM_STREAM_PLAYBACK, 0), -ENODEV);
+
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
+
+} CK_END_TEST
+
+CK_START_TEST(ba_test_playback_extra_setup) {
+
+	if (pcm_device != NULL)
+		return;
+
+	struct spawn_process sp_ba_mock;
+	snd_pcm_t *pcm = NULL;
+
+	ck_assert_int_ne(spawn_bluealsa_mock(&sp_ba_mock, NULL, true,
+				"--timeout=1000",
+				"--profile=a2dp-source",
+				"--profile=hfp-ag",
+				NULL), -1);
+
+	ck_assert_int_eq(snd_pcm_open(&pcm,
+				"bluealsa:DEV=12:34:56:78:9A:BC,CODEC=SBC",
+				SND_PCM_STREAM_PLAYBACK, 0), 0);
+	ck_assert_int_eq(test_pcm_close(NULL, pcm), 0);
+
+	ck_assert_int_eq(snd_pcm_open(&pcm,
+				"bluealsa:DEV=12:34:56:78:9A:BC,CODEC=SBC:ffff0822",
+				SND_PCM_STREAM_PLAYBACK, 0), 0);
+	ck_assert_int_eq(test_pcm_close(NULL, pcm), 0);
+
+	ck_assert_int_eq(snd_pcm_open(&pcm,
+				"bluealsa:DEV=12:34:56:78:9A:BC,PROFILE=sco,CODEC=CVSD",
+				SND_PCM_STREAM_PLAYBACK, 0), 0);
+	ck_assert_int_eq(test_pcm_close(NULL, pcm), 0);
+
+	ck_assert_int_eq(snd_pcm_open(&pcm,
+				"bluealsa:DEV=12:34:56:78:9A:BC,DELAY=10",
+				SND_PCM_STREAM_PLAYBACK, 0), 0);
+	ck_assert_int_eq(test_pcm_close(NULL, pcm), 0);
+
+	ck_assert_int_eq(snd_pcm_open(&pcm,
+				"bluealsa:DEV=12:34:56:78:9A:BC,VOL=50+",
+				SND_PCM_STREAM_PLAYBACK, 0), 0);
+	ck_assert_int_eq(test_pcm_close(NULL, pcm), 0);
+
+	ck_assert_int_eq(snd_pcm_open(&pcm,
+				"bluealsa:DEV=12:34:56:78:9A:BC,SOFTVOL=true",
+				SND_PCM_STREAM_PLAYBACK, 0), 0);
+	ck_assert_int_eq(test_pcm_close(NULL, pcm), 0);
+
+	spawn_terminate(&sp_ba_mock, 0);
+	spawn_close(&sp_ba_mock, NULL);
+
+} CK_END_TEST
+
+CK_START_TEST(test_playback_hw_set_free) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 	size_t i;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
 
 	for (i = 0; i < 5; i++) {
 		int set_hw_param_ret;
@@ -576,23 +633,22 @@ START_TEST(test_playback_hw_set_free) {
 		ck_assert_int_eq(snd_pcm_hw_free(pcm), 0);
 	}
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_playback_start) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(test_playback_start) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
 	snd_pcm_uframes_t buffer_size;
 	snd_pcm_uframes_t period_size;
 	snd_pcm_sframes_t delay;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 	size_t i;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
 	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
 				&buffer_time, &period_time), 0);
 	ck_assert_int_eq(snd_pcm_get_params(pcm, &buffer_size, &period_size), 0);
@@ -618,23 +674,22 @@ START_TEST(test_playback_start) {
 	ck_assert_int_eq(snd_pcm_writei(pcm, test_sine_s16le(period_size), period_size), period_size);
 	ck_assert_int_eq(snd_pcm_state_runtime(pcm), SND_PCM_STATE_RUNNING);
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_playback_drain) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(test_playback_drain) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
 	snd_pcm_uframes_t buffer_size;
 	snd_pcm_uframes_t period_size;
 	struct timespec t0, t, diff;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 	size_t i;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
 	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
 				&buffer_time, &period_time), 0);
 	ck_assert_int_eq(snd_pcm_get_params(pcm, &buffer_size, &period_size), 0);
@@ -656,22 +711,21 @@ START_TEST(test_playback_drain) {
 	/* verify whether elapsed time is at least PCM buffer time length */
 	ck_assert_uint_gt(diff.tv_sec * 1000000 + diff.tv_nsec / 1000, buffer_time);
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_playback_pause) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(test_playback_pause) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
 	snd_pcm_uframes_t buffer_size;
 	snd_pcm_uframes_t period_size;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 	size_t i;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
 	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
 				&buffer_time, &period_time), 0);
 	ck_assert_int_eq(snd_pcm_get_params(pcm, &buffer_size, &period_size), 0);
@@ -721,12 +775,11 @@ START_TEST(test_playback_pause) {
 
 	}
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_playback_reset) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(test_playback_reset) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
@@ -734,11 +787,11 @@ START_TEST(test_playback_reset) {
 	snd_pcm_uframes_t period_size;
 	snd_pcm_sframes_t frames;
 	snd_pcm_sframes_t delay;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 	size_t i;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
 	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
 				&buffer_time, &period_time), 0);
 	ck_assert_int_eq(snd_pcm_get_params(pcm, &buffer_size, &period_size), 0);
@@ -778,22 +831,21 @@ retry:
 	 * a period of delay, so this test is not as strict as it should be :) */
 	ck_assert_int_le(delay, 3 * period_size / 2);
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_playback_underrun) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(test_playback_underrun) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
 	snd_pcm_uframes_t buffer_size;
 	snd_pcm_uframes_t period_size;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 	size_t i;
 
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
 	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
 				&buffer_time, &period_time), 0);
 	ck_assert_int_eq(snd_pcm_get_params(pcm, &buffer_size, &period_size), 0);
@@ -823,9 +875,9 @@ START_TEST(test_playback_underrun) {
 		ck_assert_int_eq(snd_pcm_writei(pcm, test_sine_s16le(period_size), period_size), period_size);
 	ck_assert_int_eq(snd_pcm_state_runtime(pcm), SND_PCM_STATE_RUNNING);
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
 /**
  * Make reference test for device unplug.
@@ -845,12 +897,11 @@ START_TEST(test_playback_underrun) {
  * - snd_pcm_resume(pcm) = -38
  * - snd_pcm_avail(pcm) = -19
  * - snd_pcm_avail_update(pcm) = 15081
- * - snd_pcm_writei(pcm, buffer, frames) = -19
+ * - snd_pcm_writei(pcm, pcm_buffer, frames) = -19
  * - snd_pcm_wait(pcm, 10) = -19
  * - snd_pcm_close(pcm) = 0
  */
-START_TEST(reference_playback_device_unplug) {
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
+CK_START_TEST(reference_playback_device_unplug) {
 
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
@@ -890,23 +941,21 @@ START_TEST(reference_playback_device_unplug) {
 	dumprv(snd_pcm_wait(pcm, 10));
 	dumprv(snd_pcm_close(pcm));
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(ba_test_playback_device_unplug) {
+CK_START_TEST(ba_test_playback_device_unplug) {
 
 	if (pcm_device != NULL)
 		return;
 
-	fprintf(stderr, "\nSTART TEST: %s (%s:%d)\n", __func__, __FILE__, __LINE__);
-
 	unsigned int buffer_time = 200000;
 	unsigned int period_time = 25000;
 	snd_pcm_sframes_t frames = 0;
+	struct spawn_process sp_ba_mock;
 	snd_pcm_t *pcm = NULL;
-	pid_t pid = -1;
 
 	ck_assert_ptr_eq(pcm_device, NULL);
-	ck_assert_int_eq(test_pcm_open(&pid, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
+	ck_assert_int_eq(test_pcm_open(&sp_ba_mock, &pcm, SND_PCM_STREAM_PLAYBACK), 0);
 	ck_assert_int_eq(set_hw_params(pcm, pcm_format, pcm_channels, pcm_sampling,
 				&buffer_time, &period_time), 0);
 	ck_assert_int_eq(snd_pcm_prepare(pcm), 0);
@@ -940,13 +989,12 @@ START_TEST(ba_test_playback_device_unplug) {
 	ck_assert_int_eq(snd_pcm_close(pcm), 0);
 #endif
 
-	ck_assert_int_eq(test_pcm_close(pid, pcm), 0);
+	ck_assert_int_eq(test_pcm_close(&sp_ba_mock, pcm), 0);
 
-} END_TEST
+} CK_END_TEST
 
-int main(int argc, char *argv[]) {
-
-	preload(argc, argv, ".libs/aloader.so");
+int main(int argc, char *argv[], char *envp[]) {
+	preload(argc, argv, envp, ".libs/aloader.so");
 
 	int opt;
 	const char *opts = "hD:c:f:r:";
@@ -959,6 +1007,10 @@ int main(int argc, char *argv[]) {
 		{ 0, 0, 0, 0 },
 	};
 
+	bool run_capture = false;
+	bool run_playback = false;
+	bool run_unplug = false;
+
 	while ((opt = getopt_long(argc, argv, opts, longopts, NULL)) != -1)
 		switch (opt) {
 		case 'h' /* --help */ :
@@ -982,51 +1034,61 @@ int main(int argc, char *argv[]) {
 			return 1;
 		}
 
-	/* test-alsa-pcm and bluealsa-mock shall
-	 * be placed in the same directory */
 	char *argv_0 = strdup(argv[0]);
-	bluealsa_mock_path = dirname(argv_0);
-
-	Suite *s = suite_create(__FILE__);
-	SRunner *sr = srunner_create(s);
-
-	TCase *tc_capture = tcase_create("capture");
-	tcase_add_test(tc_capture, dump_capture);
-	tcase_add_test(tc_capture, test_capture_start);
-	tcase_add_test(tc_capture, test_capture_pause);
-	tcase_add_test(tc_capture, test_capture_overrun);
-	tcase_add_test(tc_capture, test_capture_poll);
-
-	TCase *tc_playback = tcase_create("playback");
-	tcase_add_test(tc_playback, dump_playback);
-	tcase_add_test(tc_playback, ba_test_playback_hw_constraints);
-	tcase_add_test(tc_playback, ba_test_playback_extra_setup);
-	tcase_add_test(tc_playback, test_playback_hw_set_free);
-	tcase_add_test(tc_playback, test_playback_start);
-	tcase_add_test(tc_playback, test_playback_drain);
-	tcase_add_test(tc_playback, test_playback_pause);
-	tcase_add_test(tc_playback, test_playback_reset);
-	tcase_add_test(tc_playback, test_playback_underrun);
-	tcase_add_test(tc_playback, ba_test_playback_device_unplug);
-
-	TCase *tc_unplug = tcase_create("unplug");
-	tcase_add_test(tc_unplug, reference_playback_device_unplug);
+	snprintf(bluealsa_mock_path, sizeof(bluealsa_mock_path),
+			"%s/mock/bluealsa-mock", dirname(argv_0));
 
 	if (argc == optind) {
-		suite_add_tcase(s, tc_capture);
-		suite_add_tcase(s, tc_playback);
+		run_capture = true;
+		run_playback = true;
 	}
 	else {
 		for (; optind < argc; optind++) {
 			if (strcmp(argv[optind], "capture") == 0)
-				suite_add_tcase(s, tc_capture);
+				run_capture = true;
 			else if (strcmp(argv[optind], "playback") == 0)
-				suite_add_tcase(s, tc_playback);
+				run_playback = true;
 			else if (strcmp(argv[optind], "unplug") == 0)
-				suite_add_tcase(s, tc_unplug);
+				run_unplug = true;
 		}
 	}
 
+	Suite *s = suite_create(__FILE__);
+	SRunner *sr = srunner_create(s);
+
+	if (run_capture) {
+		TCase *tc = tcase_create("capture");
+		tcase_add_test(tc, dump_capture);
+		tcase_add_test(tc, test_capture_start);
+		tcase_add_test(tc, test_capture_pause);
+		tcase_add_test(tc, test_capture_overrun);
+		tcase_add_test(tc, test_capture_poll);
+		suite_add_tcase(s, tc);
+	}
+
+	if (run_playback) {
+		TCase *tc = tcase_create("playback");
+		tcase_add_test(tc, dump_playback);
+		tcase_add_test(tc, ba_test_playback_hw_constraints);
+		tcase_add_test(tc, ba_test_playback_no_codec_selected);
+		tcase_add_test(tc, ba_test_playback_no_such_device);
+		tcase_add_test(tc, ba_test_playback_extra_setup);
+		tcase_add_test(tc, test_playback_hw_set_free);
+		tcase_add_test(tc, test_playback_start);
+		tcase_add_test(tc, test_playback_drain);
+		tcase_add_test(tc, test_playback_pause);
+		tcase_add_test(tc, test_playback_reset);
+		tcase_add_test(tc, test_playback_underrun);
+		tcase_add_test(tc, ba_test_playback_device_unplug);
+		suite_add_tcase(s, tc);
+	}
+
+	if (run_unplug) {
+		TCase *tc = tcase_create("unplug");
+		tcase_add_test(tc, reference_playback_device_unplug);
+		suite_add_tcase(s, tc);
+	}
+
 	srunner_run_all(sr, CK_ENV);
 	int nf = srunner_ntests_failed(sr);
 
