@@ -1,6 +1,6 @@
 /*
  * test-io.c
- * Copyright (c) 2016-2022 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -13,6 +13,7 @@
 #endif
 
 #include <errno.h>
+#include <fcntl.h>
 #include <getopt.h>
 #include <poll.h>
 #include <pthread.h>
@@ -37,6 +38,28 @@
 #endif
 
 #include "a2dp.h"
+#if ENABLE_AAC
+# include "a2dp-aac.h"
+#endif
+#if ENABLE_APTX
+# include "a2dp-aptx.h"
+#endif
+#if ENABLE_APTX_HD
+# include "a2dp-aptx-hd.h"
+#endif
+#if ENABLE_FASTSTREAM
+# include "a2dp-faststream.h"
+#endif
+#if ENABLE_LC3PLUS
+# include "a2dp-lc3plus.h"
+#endif
+#if ENABLE_LDAC
+# include "a2dp-ldac.h"
+#endif
+#if ENABLE_MPEG
+# include "a2dp-mpeg.h"
+#endif
+#include "a2dp-sbc.h"
 #include "ba-adapter.h"
 #include "ba-device.h"
 #include "ba-rfcomm.h"
@@ -46,23 +69,21 @@
 #include "bluez.h"
 #include "hfp.h"
 #include "io.h"
-#include "rtp.h"
-#include "sco.h"
+#if ENABLE_OFONO
+# include "ofono.h"
+#endif
+#if ENABLE_LC3PLUS || ENABLE_LDAC
+# include "rtp.h"
+#endif
+#include "storage.h"
 #include "shared/a2dp-codecs.h"
 #include "shared/defs.h"
 #include "shared/log.h"
 
 #include "../src/a2dp.c"
-#include "../src/a2dp-aac.c"
-#include "../src/a2dp-aptx-hd.c"
-#include "../src/a2dp-aptx.c"
-#include "../src/a2dp-faststream.c"
-#include "../src/a2dp-lc3plus.c"
-#include "../src/a2dp-ldac.c"
-#include "../src/a2dp-mpeg.c"
-#include "../src/a2dp-sbc.c"
 #include "../src/ba-transport.c"
 #include "inc/btd.inc"
+#include "inc/check.inc"
 #include "inc/sine.inc"
 
 #define CHECK_VERSION ( \
@@ -70,22 +91,47 @@
 		(CHECK_MINOR_VERSION << 8 & 0x00ff00) | \
 		(CHECK_MICRO_VERSION << 0 & 0x0000ff))
 
+void *a2dp_aac_dec_thread(struct ba_transport_thread *th);
+void *a2dp_aac_enc_thread(struct ba_transport_thread *th);
+void *a2dp_aptx_dec_thread(struct ba_transport_thread *th);
+void *a2dp_aptx_enc_thread(struct ba_transport_thread *th);
+void *a2dp_aptx_hd_dec_thread(struct ba_transport_thread *th);
+void *a2dp_aptx_hd_enc_thread(struct ba_transport_thread *th);
+void *a2dp_faststream_dec_thread(struct ba_transport_thread *th);
+void *a2dp_faststream_enc_thread(struct ba_transport_thread *th);
+void *a2dp_lc3plus_dec_thread(struct ba_transport_thread *th);
+void *a2dp_lc3plus_enc_thread(struct ba_transport_thread *th);
+void *a2dp_ldac_dec_thread(struct ba_transport_thread *th);
+void *a2dp_ldac_enc_thread(struct ba_transport_thread *th);
+void *a2dp_mp3_enc_thread(struct ba_transport_thread *th);
+void *a2dp_mpeg_dec_thread(struct ba_transport_thread *th);
+void *a2dp_sbc_dec_thread(struct ba_transport_thread *th);
+void *a2dp_sbc_enc_thread(struct ba_transport_thread *th);
+void *sco_dec_thread(struct ba_transport_thread *th);
+void *sco_enc_thread(struct ba_transport_thread *th);
+
 int bluealsa_dbus_pcm_register(struct ba_transport_pcm *pcm) {
-	debug("%s: %p", __func__, (void *)pcm); return 0; }
+	debug("%s: %p", __func__, (void *)pcm); (void)pcm; return 0; }
 void bluealsa_dbus_pcm_update(struct ba_transport_pcm *pcm, unsigned int mask) {
-	debug("%s: %p %#x", __func__, (void *)pcm, mask); }
+	debug("%s: %p %#x", __func__, (void *)pcm, mask); (void)pcm; (void)mask; }
 void bluealsa_dbus_pcm_unregister(struct ba_transport_pcm *pcm) {
-	debug("%s: %p", __func__, (void *)pcm); }
+	debug("%s: %p", __func__, (void *)pcm); (void)pcm; }
 struct ba_rfcomm *ba_rfcomm_new(struct ba_transport *sco, int fd) {
-	debug("%s: %p", __func__, (void *)sco); (void)fd; return NULL; }
+	debug("%s: %p", __func__, (void *)sco); (void)sco; (void)fd; return NULL; }
 void ba_rfcomm_destroy(struct ba_rfcomm *r) {
-	debug("%s: %p", __func__, (void *)r); }
+	debug("%s: %p", __func__, (void *)r); (void)r; }
 int ba_rfcomm_send_signal(struct ba_rfcomm *r, enum ba_rfcomm_signal sig) {
-	debug("%s: %p: %#x", __func__, (void *)r, sig); return 0; }
+	debug("%s: %p: %#x", __func__, (void *)r, sig); (void)r; (void)sig; return 0; }
 bool bluez_a2dp_set_configuration(const char *current_dbus_sep_path,
 		const struct a2dp_sep *sep, GError **error) {
-	debug("%s: %s", __func__, current_dbus_sep_path); (void)sep;
-	(void)error; return false; }
+	debug("%s: %s: %p", __func__, current_dbus_sep_path, sep);
+	(void)current_dbus_sep_path; (void)sep; (void)error; return false; }
+int ofono_call_volume_update(struct ba_transport *t) {
+	debug("%s: %p", __func__, t); (void)t; return 0; }
+int storage_device_load(const struct ba_device *d) { (void)d; return 0; }
+int storage_device_save(const struct ba_device *d) { (void)d; return 0; }
+int storage_pcm_data_sync(struct ba_transport_pcm *pcm) { (void)pcm; return 0; }
+int storage_pcm_data_update(const struct ba_transport_pcm *pcm) { (void)pcm; return 0; }
 
 static const a2dp_sbc_t config_sbc_44100_stereo = {
 	.frequency = SBC_SAMPLING_FREQ_44100,
@@ -160,7 +206,8 @@ static bool enable_vbr_mode = false;
 static bool dump_data = false;
 static bool packet_loss = false;
 
-static struct bt_dump *btd = NULL;
+/* input BT dump file */
+static struct bt_dump *btdin = NULL;
 
 #if HAVE_SNDFILE
 static void *pcm_write_frames_sndfile_async(void *userdata) {
@@ -324,7 +371,7 @@ static void bt_data_write(struct ba_transport *t) {
 
 	if (input_bt_file != NULL) {
 
-		while ((len = bt_dump_read(btd, buffer, sizeof(buffer))) != -1) {
+		while ((len = bt_dump_read(btdin, buffer, sizeof(buffer))) != -1) {
 			if (packet_loss && random() < INT32_MAX / 3 && !first_packet) {
 				debug("Simulating packet loss: Dropping BT packet!");
 				continue;
@@ -379,12 +426,12 @@ static void *test_io_thread_dump_bt(struct ba_transport_thread *th) {
 
 	if (dump_data) {
 		char fname[64];
-		sprintf(fname, "encoded-%s.btd", transport_type_to_fname(t->type));
+		sprintf(fname, "encoded-%s.btd", transport_to_fname(t));
 		ck_assert_ptr_ne(btd = bt_dump_create(fname, t), NULL);
 	}
 
 	debug_transport_thread_loop(th, "START");
-	ba_transport_thread_set_state_running(th);
+	ba_transport_thread_state_set_running(th);
 	while (poll(pfds, ARRAYSIZE(pfds), 500) > 0) {
 
 		if ((len = io_bt_read(th, buffer, sizeof(buffer))) <= 0) {
@@ -416,32 +463,8 @@ static void *test_io_thread_dump_pcm(struct ba_transport_thread *th) {
 
 	pthread_cleanup_push(PTHREAD_CLEANUP(ba_transport_thread_cleanup), th);
 
-	struct ba_transport *t = th->t;
-	struct ba_transport_pcm *t_pcm = NULL;
-
-	switch (t->type.profile) {
-	case BA_TRANSPORT_PROFILE_A2DP_SOURCE:
-		t_pcm = &t->a2dp.pcm;
-		break;
-	case BA_TRANSPORT_PROFILE_A2DP_SINK:
-		t_pcm = &t->a2dp.pcm_bc;
-		break;
-	case BA_TRANSPORT_PROFILE_HFP_AG:
-	case BA_TRANSPORT_PROFILE_HSP_AG:
-		t_pcm = &t->sco.spk_pcm;
-		break;
-	case BA_TRANSPORT_PROFILE_HFP_HF:
-	case BA_TRANSPORT_PROFILE_HSP_HS:
-		t_pcm = &t->sco.mic_pcm;
-		break;
-	default:
-		g_assert_not_reached();
-	}
-
-	struct pollfd pfds[] = {{ t_pcm->fd, POLLIN, 0 }};
+	struct ba_transport_pcm *t_pcm = th->pcm;
 	size_t decoded_samples_total = 0;
-	uint8_t buffer[2048];
-	ssize_t len;
 
 #if HAVE_SNDFILE
 	SNDFILE *sf = NULL;
@@ -471,7 +494,7 @@ static void *test_io_thread_dump_pcm(struct ba_transport_thread *th) {
 	if (dump_data) {
 #if HAVE_SNDFILE
 		char fname[64];
-		sprintf(fname, "decoded-%s.wav", transport_type_to_fname(t->type));
+		sprintf(fname, "decoded-%s.wav", transport_to_fname(th->t));
 		ck_assert_ptr_ne(sf = sf_open(fname, SFM_WRITE, &sf_info), NULL);
 #else
 		error("Dumping audio files requires sndfile library!");
@@ -479,8 +502,18 @@ static void *test_io_thread_dump_pcm(struct ba_transport_thread *th) {
 	}
 
 	debug_transport_thread_loop(th, "START");
-	ba_transport_thread_set_state_running(th);
-	while (poll(pfds, ARRAYSIZE(pfds), 500) > 0) {
+	for (ba_transport_thread_state_set_running(th);;) {
+
+		struct pollfd pfds[] = {{ -1, POLLIN, 0 }};
+		uint8_t buffer[2048];
+		ssize_t len;
+
+		pthread_mutex_lock(&t_pcm->mutex);
+		pfds[0].fd = t_pcm->fd;
+		pthread_mutex_unlock(&t_pcm->mutex);
+
+		if (poll(pfds, ARRAYSIZE(pfds), 500) <= 0)
+			break;
 
 		if ((len = read(pfds[0].fd, buffer, sizeof(buffer))) == -1) {
 			debug("PCM read error: %s", strerror(errno));
@@ -543,46 +576,8 @@ static void test_io(struct ba_transport *t_src, struct ba_transport *t_snk,
 	if (dec == test_io_thread_dump_bt && input_bt_file != NULL)
 		return;
 
-	struct ba_transport_pcm *t_src_pcm = NULL;
-	struct ba_transport_pcm *t_snk_pcm = NULL;
-
-	switch (t_src->type.profile) {
-	case BA_TRANSPORT_PROFILE_A2DP_SOURCE:
-		t_src_pcm = &t_src->a2dp.pcm;
-		break;
-	case BA_TRANSPORT_PROFILE_A2DP_SINK:
-		t_src_pcm = &t_src->a2dp.pcm_bc;
-		break;
-	case BA_TRANSPORT_PROFILE_HFP_AG:
-	case BA_TRANSPORT_PROFILE_HSP_AG:
-		t_src_pcm = &t_src->sco.spk_pcm;
-		break;
-	case BA_TRANSPORT_PROFILE_HFP_HF:
-	case BA_TRANSPORT_PROFILE_HSP_HS:
-		t_src_pcm = &t_src->sco.mic_pcm;
-		break;
-	default:
-		g_assert_not_reached();
-	}
-
-	switch (t_snk->type.profile) {
-	case BA_TRANSPORT_PROFILE_A2DP_SOURCE:
-		t_snk_pcm = &t_snk->a2dp.pcm_bc;
-		break;
-	case BA_TRANSPORT_PROFILE_A2DP_SINK:
-		t_snk_pcm = &t_snk->a2dp.pcm;
-		break;
-	case BA_TRANSPORT_PROFILE_HFP_AG:
-	case BA_TRANSPORT_PROFILE_HSP_AG:
-		t_snk_pcm = &t_snk->sco.mic_pcm;
-		break;
-	case BA_TRANSPORT_PROFILE_HFP_HF:
-	case BA_TRANSPORT_PROFILE_HSP_HS:
-		t_snk_pcm = &t_snk->sco.spk_pcm;
-		break;
-	default:
-		g_assert_not_reached();
-	}
+	struct ba_transport_pcm *t_src_pcm = t_src->thread_enc.pcm;
+	struct ba_transport_pcm *t_snk_pcm = t_snk->thread_dec.pcm;
 
 	int bt_fds[2];
 	ck_assert_int_eq(socketpair(AF_UNIX, SOCK_SEQPACKET | SOCK_NONBLOCK, 0, bt_fds), 0);
@@ -621,20 +616,18 @@ static void test_io(struct ba_transport *t_src, struct ba_transport *t_snk,
 	ba_transport_pcm_release(t_src_pcm);
 	pthread_mutex_unlock(&t_src_pcm->mutex);
 
-	transport_thread_cancel_prepare(&t_src->thread_enc);
-	transport_thread_cancel(&t_src->thread_enc);
+	ba_transport_stop(t_src);
 
 	pthread_mutex_lock(&t_snk_pcm->mutex);
 	ba_transport_pcm_release(t_snk_pcm);
 	pthread_mutex_unlock(&t_snk_pcm->mutex);
 
-	transport_thread_cancel_prepare(&t_snk->thread_dec);
-	transport_thread_cancel(&t_snk->thread_dec);
+	ba_transport_stop(t_snk);
 
 }
 
 static int test_transport_acquire(struct ba_transport *t) {
-	debug("Acquire transport: %d", t->bt_fd);
+	debug("Acquire transport: %d", t->bt_fd); (void)t;
 	return 0;
 }
 
@@ -645,15 +638,15 @@ static int test_transport_release_bt_a2dp(struct ba_transport *t) {
 
 static struct ba_transport *test_transport_new_a2dp(
 		struct ba_device *device,
-		struct ba_transport_type type,
+		enum ba_transport_profile profile,
 		const char *dbus_path,
 		const struct a2dp_codec *codec,
 		const void *configuration) {
 #if DEBUG
 	if (input_bt_file != NULL)
-		configuration = &btd->a2dp_configuration;
+		configuration = &btdin->a2dp_configuration;
 #endif
-	struct ba_transport *t = ba_transport_new_a2dp(device, type, ":test",
+	struct ba_transport *t = ba_transport_new_a2dp(device, profile, ":test",
 			dbus_path, codec, configuration);
 	t->acquire = test_transport_acquire;
 	t->release = test_transport_release_bt_a2dp;
@@ -662,31 +655,28 @@ static struct ba_transport *test_transport_new_a2dp(
 
 static struct ba_transport *test_transport_new_sco(
 		struct ba_device *device,
-		struct ba_transport_type type,
+		enum ba_transport_profile profile,
 		const char *dbus_path) {
-	struct ba_transport *t = ba_transport_new_sco(device, type, ":test",
+	struct ba_transport *t = ba_transport_new_sco(device, profile, ":test",
 			dbus_path, -1);
 	t->acquire = test_transport_acquire;
 	return t;
 }
 
-START_TEST(test_a2dp_sbc) {
+CK_START_TEST(test_a2dp_sbc) {
 
-	struct ba_transport_type ttype = {
-		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
-		.codec = A2DP_CODEC_SBC };
-	struct ba_transport *t1 = test_transport_new_a2dp(device1, ttype, "/path/sbc",
-			&a2dp_sbc_source, &config_sbc_44100_stereo);
-	ttype.profile = BA_TRANSPORT_PROFILE_A2DP_SINK;
-	struct ba_transport *t2 = test_transport_new_a2dp(device2, ttype, "/path/sbc",
-			&a2dp_sbc_sink, &config_sbc_44100_stereo);
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/sbc", &a2dp_sbc_source,
+			&config_sbc_44100_stereo);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/sbc", &a2dp_sbc_sink,
+			&config_sbc_44100_stereo);
 
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 153 * 3;
 		test_io(t1, t2, a2dp_sbc_enc_thread, a2dp_sbc_dec_thread, 4 * 1024);
 	}
 	else {
-		debug("\n\n*** A2DP codec: SBC ***");
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 153 * 3;
 		test_io(t1, t2, a2dp_sbc_enc_thread, test_io_thread_dump_bt, 2 * 1024);
 		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_sbc_dec_thread, 2 * 1024);
@@ -695,28 +685,159 @@ START_TEST(test_a2dp_sbc) {
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
 
-} END_TEST
+} CK_END_TEST
+
+CK_START_TEST(test_a2dp_sbc_invalid_config) {
+
+	const a2dp_sbc_t config_sbc_invalid = { 0 };
+	struct ba_transport *t = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/sbc", &a2dp_sbc_source,
+			&config_sbc_invalid);
+
+	int bt_fds[2];
+	ck_assert_int_eq(socketpair(AF_UNIX, SOCK_SEQPACKET | SOCK_NONBLOCK, 0, bt_fds), 0);
+	debug("Created BT socket pair: %d, %d", bt_fds[0], bt_fds[1]);
+	t->mtu_read = t->mtu_write = 153 * 3;
+	t->bt_fd = bt_fds[1];
+
+	struct ba_transport_thread *th = &t->thread_enc;
+	ck_assert_int_eq(ba_transport_thread_create(th, a2dp_sbc_enc_thread, "sbc", true), 0);
+	ck_assert_int_eq(ba_transport_thread_state_wait_running(th), -1);
+
+	ba_transport_destroy(t);
+	close(bt_fds[0]);
+
+} CK_END_TEST
+
+CK_START_TEST(test_a2dp_sbc_pcm_drop) {
+
+	int16_t pcm_zero[90] = { 0 };
+	int16_t pcm_rand[ARRAYSIZE(pcm_zero)];
+	for (size_t i = 0; i < ARRAYSIZE(pcm_rand); i++)
+		pcm_rand[i] = rand();
+
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/sbc", &a2dp_sbc_source,
+			&config_sbc_44100_stereo);
+
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/sbc", &a2dp_sbc_sink,
+			&config_sbc_44100_stereo);
+
+	struct ba_transport_thread *th1 = &t1->thread_enc;
+	struct ba_transport_thread *th2 = &t2->thread_dec;
+
+	int bt_fds[2];
+	ck_assert_int_eq(socketpair(AF_UNIX, SOCK_SEQPACKET | SOCK_NONBLOCK, 0, bt_fds), 0);
+	debug("Created BT socket pair: %d, %d", bt_fds[0], bt_fds[1]);
+	t1->mtu_read = t1->mtu_write = 256;
+	t2->mtu_read = t2->mtu_write = 256;
+	t1->bt_fd = bt_fds[1];
+	t2->bt_fd = bt_fds[0];
+
+	int pcm_snk_fds[2];
+	ck_assert_int_eq(pipe2(pcm_snk_fds, O_NONBLOCK), 0);
+	debug("Created PCM pipe pair: %d, %d", pcm_snk_fds[0], pcm_snk_fds[1]);
+	th1->pcm->fd = pcm_snk_fds[0];
+	th1->pcm->active = true;
+
+	int pcm_src_fds[2];
+	ck_assert_int_eq(pipe2(pcm_src_fds, O_NONBLOCK), 0);
+	debug("Created PCM pipe pair: %d, %d", pcm_src_fds[0], pcm_src_fds[1]);
+	th2->pcm->fd = pcm_src_fds[1];
+	th2->pcm->active = true;
+
+	/* sink PCM */
+	struct ba_transport_pcm *pcm = th1->pcm;
+
+	/* write zero samples to PCM FIFO until it is full */
+	while (write(pcm_snk_fds[1], pcm_zero, sizeof(pcm_zero)) > 0)
+		continue;
+
+	/* drop PCM samples before IO thread was started */
+	ck_assert_int_eq(ba_transport_pcm_drop(pcm), 0);
+
+	/* start IO thread and make sure it is running */
+	ck_assert_int_eq(ba_transport_thread_create(th1, a2dp_sbc_enc_thread, "sbc", true), 0);
+	ck_assert_int_eq(ba_transport_thread_state_wait_running(th1), 0);
+
+	/* wait for 50 ms - let the thread to run for a while */
+	usleep(50000);
+
+	uint8_t bt_buffer[1024];
+	/* try to read data from BT socket - it should be empty because
+	 * the PCM has been dropped before the thread was started */
+	ck_assert_int_eq(read(bt_fds[0], bt_buffer, sizeof(bt_buffer)), -1);
+	ck_assert_int_eq(errno, EAGAIN);
+
+	/* write non-zero samples to PCM FIFO and process some of it */
+	while (write(pcm_snk_fds[1], pcm_rand, sizeof(pcm_rand)) > 0)
+		continue;
+	usleep(50000);
+
+	/* drop PCM samples */
+	ck_assert_int_eq(ba_transport_pcm_drop(pcm), 0);
+
+	/* write a little bit of non-zero samples and wait for processing; the
+	 * number of samples shall be small enough to not produce a single codec
+	 * frame, but enough to fill-in internal buffers */
+	ck_assert_int_gt(write(pcm_snk_fds[1], pcm_rand, sizeof(pcm_rand)), 0);
+	usleep(10000);
+
+	/* again drop PCM samples */
+	ck_assert_int_eq(ba_transport_pcm_drop(pcm), 0);
+
+	/* flush already processed data */
+	while (read(bt_fds[0], bt_buffer, sizeof(bt_buffer)) > 0)
+		continue;
+
+	/* After PCM has been dropped, IO thread should not process any more
+	 * non-zero samples. We will check this by writing zero samples and
+	 * checking if decoded data is all silence. */
+
+	ck_assert_int_eq(ba_transport_thread_create(th2, a2dp_sbc_dec_thread, "sbc", true), 0);
+	ck_assert_int_eq(ba_transport_thread_state_wait_running(th2), 0);
+
+	/* write some zero samples to PCM FIFO and process them */
+	for (size_t i = 0; i < 100; i++)
+		if (write(pcm_snk_fds[1], pcm_zero, sizeof(pcm_zero)) <= 0)
+			break;
+	usleep(250000);
+
+	ssize_t rv;
+	int16_t pcm_buffer[1024];
+	/* read decoded data and check if it is all silence */
+	while ((rv = read(pcm_src_fds[0], pcm_buffer, sizeof(pcm_buffer))) > 0) {
+		const size_t samples = rv / sizeof(pcm_buffer[0]);
+		for (size_t i = 0; i < samples; i++)
+			ck_assert_int_eq(pcm_buffer[i], 0);
+	}
+
+	ba_transport_destroy(t1);
+	ba_transport_destroy(t2);
+	close(pcm_snk_fds[0]);
+	close(pcm_src_fds[1]);
+	close(bt_fds[0]);
+
+} CK_END_TEST
 
 #if ENABLE_MP3LAME
-START_TEST(test_a2dp_mp3) {
+CK_START_TEST(test_a2dp_mp3) {
 
 	config_mp3_44100_stereo.vbr = enable_vbr_mode ? 1 : 0;
 
-	struct ba_transport_type ttype = {
-		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
-		.codec = A2DP_CODEC_MPEG12 };
-	struct ba_transport *t1 = test_transport_new_a2dp(device1, ttype, "/path/mp3",
-			&a2dp_mpeg_source, &config_mp3_44100_stereo);
-	ttype.profile = BA_TRANSPORT_PROFILE_A2DP_SINK;
-	struct ba_transport *t2 = test_transport_new_a2dp(device2, ttype, "/path/mp3",
-			&a2dp_mpeg_sink, &config_mp3_44100_stereo);
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/mp3", &a2dp_mpeg_source,
+			&config_mp3_44100_stereo);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/mp3", &a2dp_mpeg_sink,
+			&config_mp3_44100_stereo);
 
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 1024;
 		test_io(t1, t2, a2dp_mp3_enc_thread, a2dp_mpeg_dec_thread, 4 * 1024);
 	}
 	else {
-		debug("\n\n*** A2DP codec: MP3 ***");
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 512;
 		test_io(t1, t2, a2dp_mp3_enc_thread, test_io_thread_dump_bt, 3 * 1024);
 		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_mpeg_dec_thread, 3 * 1024);
@@ -725,31 +846,28 @@ START_TEST(test_a2dp_mp3) {
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
 
-} END_TEST
+} CK_END_TEST
 #endif
 
 #if ENABLE_AAC
-START_TEST(test_a2dp_aac) {
+CK_START_TEST(test_a2dp_aac) {
 
 	config.aac_afterburner = true;
 	config.aac_prefer_vbr = enable_vbr_mode;
 	config_aac_44100_stereo.vbr = enable_vbr_mode ? 1 : 0;
 
-	struct ba_transport_type ttype = {
-		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
-		.codec = A2DP_CODEC_MPEG24 };
-	struct ba_transport *t1 = test_transport_new_a2dp(device1, ttype, "/path/aac",
-			&a2dp_aac_source, &config_aac_44100_stereo);
-	ttype.profile = BA_TRANSPORT_PROFILE_A2DP_SINK;
-	struct ba_transport *t2 = test_transport_new_a2dp(device2, ttype, "/path/aac",
-			&a2dp_aac_sink, &config_aac_44100_stereo);
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/aac", &a2dp_aac_source,
+			&config_aac_44100_stereo);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/aac", &a2dp_aac_sink,
+			&config_aac_44100_stereo);
 
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 450;
 		test_io(t1, t2, a2dp_aac_enc_thread, a2dp_aac_dec_thread, 4 * 1024);
 	}
 	else {
-		debug("\n\n*** A2DP codec: AAC ***");
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 190;
 		test_io(t1, t2, a2dp_aac_enc_thread, test_io_thread_dump_bt, 2 * 1024);
 		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_aac_dec_thread, 2 * 1024);
@@ -758,20 +876,18 @@ START_TEST(test_a2dp_aac) {
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
 
-} END_TEST
+} CK_END_TEST
 #endif
 
 #if ENABLE_APTX
-START_TEST(test_a2dp_aptx) {
+CK_START_TEST(test_a2dp_aptx) {
 
-	struct ba_transport_type ttype = {
-		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
-		.codec = A2DP_CODEC_VENDOR_APTX };
-	struct ba_transport *t1 = test_transport_new_a2dp(device1, ttype, "/path/aptx",
-			&a2dp_aptx_source, &config_aptx_44100_stereo);
-	ttype.profile = BA_TRANSPORT_PROFILE_A2DP_SINK;
-	struct ba_transport *t2 = test_transport_new_a2dp(device2, ttype, "/path/aptx",
-			&a2dp_aptx_sink, &config_aptx_44100_stereo);
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/aptx", &a2dp_aptx_source,
+			&config_aptx_44100_stereo);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/aptx", &a2dp_aptx_sink,
+			&config_aptx_44100_stereo);
 
 	if (aging_duration) {
 #if HAVE_APTX_DECODE
@@ -780,7 +896,6 @@ START_TEST(test_a2dp_aptx) {
 #endif
 	}
 	else {
-		debug("\n\n*** A2DP codec: apt-X ***");
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 40;
 		test_io(t1, t2, a2dp_aptx_enc_thread, test_io_thread_dump_bt, 2 * 1024);
 #if HAVE_APTX_DECODE
@@ -791,20 +906,18 @@ START_TEST(test_a2dp_aptx) {
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
 
-} END_TEST
+} CK_END_TEST
 #endif
 
 #if ENABLE_APTX_HD
-START_TEST(test_a2dp_aptx_hd) {
+CK_START_TEST(test_a2dp_aptx_hd) {
 
-	struct ba_transport_type ttype = {
-		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
-		.codec = A2DP_CODEC_VENDOR_APTX_HD };
-	struct ba_transport *t1 = test_transport_new_a2dp(device1, ttype, "/path/aptxhd",
-			&a2dp_aptx_hd_source, &config_aptx_hd_44100_stereo);
-	ttype.profile = BA_TRANSPORT_PROFILE_A2DP_SINK;
-	struct ba_transport *t2 = test_transport_new_a2dp(device2, ttype, "/path/aptxhd",
-			&a2dp_aptx_hd_sink, &config_aptx_hd_44100_stereo);
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/aptxhd", &a2dp_aptx_hd_source,
+			&config_aptx_hd_44100_stereo);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/aptxhd", &a2dp_aptx_hd_sink,
+			&config_aptx_hd_44100_stereo);
 
 	if (aging_duration) {
 #if HAVE_APTX_HD_DECODE
@@ -813,7 +926,6 @@ START_TEST(test_a2dp_aptx_hd) {
 #endif
 	}
 	else {
-		debug("\n\n*** A2DP codec: apt-X HD ***");
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 60;
 		test_io(t1, t2, a2dp_aptx_hd_enc_thread, test_io_thread_dump_bt, 2 * 1024);
 #if HAVE_APTX_HD_DECODE
@@ -824,20 +936,18 @@ START_TEST(test_a2dp_aptx_hd) {
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
 
-} END_TEST
+} CK_END_TEST
 #endif
 
 #if ENABLE_FASTSTREAM
-START_TEST(test_a2dp_faststream) {
+CK_START_TEST(test_a2dp_faststream_music) {
 
-	struct ba_transport_type ttype = {
-		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
-		.codec = A2DP_CODEC_VENDOR_FASTSTREAM };
-	struct ba_transport *t1 = test_transport_new_a2dp(device1, ttype, "/path/faststream",
-			&a2dp_faststream_source, &config_faststream_44100_16000);
-	ttype.profile = BA_TRANSPORT_PROFILE_A2DP_SINK;
-	struct ba_transport *t2 = test_transport_new_a2dp(device2, ttype, "/path/faststream",
-			&a2dp_faststream_sink, &config_faststream_44100_16000);
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/faststream", &a2dp_faststream_source,
+			&config_faststream_44100_16000);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/faststream", &a2dp_faststream_sink,
+			&config_faststream_44100_16000);
 
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 72 * 3;
@@ -845,10 +955,32 @@ START_TEST(test_a2dp_faststream) {
 	}
 	else {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 72 * 3;
-		debug("\n\n*** A2DP codec: FastStream MUSIC ***");
 		test_io(t1, t2, a2dp_faststream_enc_thread, test_io_thread_dump_bt, 2 * 1024);
 		test_io(t1, t2, test_io_thread_dump_pcm, a2dp_faststream_dec_thread, 2 * 1024);
-		debug("\n\n*** A2DP codec: FastStream VOICE ***");
+	}
+
+	ba_transport_destroy(t1);
+	ba_transport_destroy(t2);
+
+} CK_END_TEST
+#endif
+
+#if ENABLE_FASTSTREAM
+CK_START_TEST(test_a2dp_faststream_voice) {
+
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/faststream", &a2dp_faststream_source,
+			&config_faststream_44100_16000);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/faststream", &a2dp_faststream_sink,
+			&config_faststream_44100_16000);
+
+	if (aging_duration) {
+		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 72 * 3;
+		test_io(t2, t1, a2dp_faststream_enc_thread, a2dp_faststream_dec_thread, 4 * 1024);
+	}
+	else {
+		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 72 * 3;
 		test_io(t2, t1, a2dp_faststream_enc_thread, test_io_thread_dump_bt, 2 * 1024);
 		test_io(t2, t1, test_io_thread_dump_pcm, a2dp_faststream_dec_thread, 2 * 1024);
 	}
@@ -856,20 +988,18 @@ START_TEST(test_a2dp_faststream) {
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
 
-} END_TEST
+} CK_END_TEST
 #endif
 
 #if ENABLE_LC3PLUS
-START_TEST(test_a2dp_lc3plus) {
+CK_START_TEST(test_a2dp_lc3plus) {
 
-	struct ba_transport_type ttype = {
-		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
-		.codec = A2DP_CODEC_VENDOR_LC3PLUS };
-	struct ba_transport *t1 = test_transport_new_a2dp(device1, ttype, "/path/lc3plus",
-			&a2dp_lc3plus_source, &config_lc3plus_48000_stereo);
-	ttype.profile = BA_TRANSPORT_PROFILE_A2DP_SINK;
-	struct ba_transport *t2 = test_transport_new_a2dp(device2, ttype, "/path/lc3plus",
-			&a2dp_lc3plus_sink, &config_lc3plus_48000_stereo);
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/lc3plus", &a2dp_lc3plus_source,
+			&config_lc3plus_48000_stereo);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/lc3plus", &a2dp_lc3plus_sink,
+			&config_lc3plus_48000_stereo);
 
 	if (aging_duration) {
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write =
@@ -877,7 +1007,6 @@ START_TEST(test_a2dp_lc3plus) {
 		test_io(t1, t2, a2dp_lc3plus_enc_thread, a2dp_lc3plus_dec_thread, 4 * 1024);
 	}
 	else {
-		debug("\n\n*** A2DP codec: LC3plus ***");
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write =
 			RTP_HEADER_LEN + sizeof(rtp_media_header_t) + 300;
 		test_io(t1, t2, a2dp_lc3plus_enc_thread, test_io_thread_dump_bt, 2 * 1024);
@@ -887,23 +1016,21 @@ START_TEST(test_a2dp_lc3plus) {
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
 
-} END_TEST
+} CK_END_TEST
 #endif
 
 #if ENABLE_LDAC
-START_TEST(test_a2dp_ldac) {
+CK_START_TEST(test_a2dp_ldac) {
 
 	config.ldac_abr = true;
 	config.ldac_eqmid = LDACBT_EQMID_HQ;
 
-	struct ba_transport_type ttype = {
-		.profile = BA_TRANSPORT_PROFILE_A2DP_SOURCE,
-		.codec = A2DP_CODEC_VENDOR_LDAC };
-	struct ba_transport *t1 = test_transport_new_a2dp(device1, ttype, "/path/ldac",
-			&a2dp_ldac_source, &config_ldac_44100_stereo);
-	ttype.profile = BA_TRANSPORT_PROFILE_A2DP_SINK;
-	struct ba_transport *t2 = test_transport_new_a2dp(device2, ttype, "/path/ldac",
-			&a2dp_ldac_sink, &config_ldac_44100_stereo);
+	struct ba_transport *t1 = test_transport_new_a2dp(device1,
+			BA_TRANSPORT_PROFILE_A2DP_SOURCE, "/path/ldac", &a2dp_ldac_source,
+			&config_ldac_44100_stereo);
+	struct ba_transport *t2 = test_transport_new_a2dp(device2,
+			BA_TRANSPORT_PROFILE_A2DP_SINK, "/path/ldac", &a2dp_ldac_sink,
+			&config_ldac_44100_stereo);
 
 	if (aging_duration) {
 #if HAVE_LDAC_DECODE
@@ -913,7 +1040,6 @@ START_TEST(test_a2dp_ldac) {
 #endif
 	}
 	else {
-		debug("\n\n*** A2DP codec: LDAC ***");
 		t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write =
 			RTP_HEADER_LEN + sizeof(rtp_media_header_t) + 660 + 6;
 		test_io(t1, t2, a2dp_ldac_enc_thread, test_io_thread_dump_bt, 2 * 1024);
@@ -925,17 +1051,16 @@ START_TEST(test_a2dp_ldac) {
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
 
-} END_TEST
+} CK_END_TEST
 #endif
 
-START_TEST(test_sco_cvsd) {
+CK_START_TEST(test_sco_cvsd) {
 
-	struct ba_transport_type ttype = {
-		.profile = BA_TRANSPORT_PROFILE_HSP_AG };
-	struct ba_transport *t1 = test_transport_new_sco(device1, ttype, "/path/sco/cvsd");
-	struct ba_transport *t2 = test_transport_new_sco(device2, ttype, "/path/sco/cvsd");
+	struct ba_transport *t1 = test_transport_new_sco(device1,
+			BA_TRANSPORT_PROFILE_HSP_AG, "/path/sco/cvsd");
+	struct ba_transport *t2 = test_transport_new_sco(device2,
+			BA_TRANSPORT_PROFILE_HSP_AG, "/path/sco/cvsd");
 
-	debug("\n\n*** SCO codec: CVSD ***");
 	t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 48;
 	test_io(t1, t2, sco_enc_thread, test_io_thread_dump_bt, 600);
 	test_io(t1, t2, test_io_thread_dump_pcm, sco_dec_thread, 600);
@@ -943,21 +1068,21 @@ START_TEST(test_sco_cvsd) {
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
 
-} END_TEST
+} CK_END_TEST
 
 #if ENABLE_MSBC
-START_TEST(test_sco_msbc) {
+CK_START_TEST(test_sco_msbc) {
 
 	adapter->hci.features[2] = LMP_TRSP_SCO;
 	adapter->hci.features[3] = LMP_ESCO;
 
-	struct ba_transport_type ttype = {
-		.profile = BA_TRANSPORT_PROFILE_HFP_AG,
-		.codec = HFP_CODEC_MSBC };
-	struct ba_transport *t1 = test_transport_new_sco(device1, ttype, "/path/sco/msbc");
-	struct ba_transport *t2 = test_transport_new_sco(device2, ttype, "/path/sco/msbc");
+	struct ba_transport *t1 = test_transport_new_sco(device1,
+			BA_TRANSPORT_PROFILE_HFP_AG, "/path/sco/msbc");
+	ba_transport_set_codec(t1, HFP_CODEC_MSBC);
+	struct ba_transport *t2 = test_transport_new_sco(device2,
+			BA_TRANSPORT_PROFILE_HFP_AG, "/path/sco/msbc");
+	ba_transport_set_codec(t2, HFP_CODEC_MSBC);
 
-	debug("\n\n*** SCO codec: mSBC ***");
 	t1->mtu_read = t1->mtu_write = t2->mtu_read = t2->mtu_write = 24;
 	test_io(t1, t2, sco_enc_thread, test_io_thread_dump_bt, 600);
 	test_io(t1, t2, test_io_thread_dump_pcm, sco_dec_thread, 600);
@@ -965,7 +1090,7 @@ START_TEST(test_sco_msbc) {
 	ba_transport_destroy(t1);
 	ba_transport_destroy(t2);
 
-} END_TEST
+} CK_END_TEST
 #endif
 
 int main(int argc, char *argv[]) {
@@ -979,6 +1104,8 @@ int main(int argc, char *argv[]) {
 #endif
 	} codecs[] = {
 		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_SBC), test_a2dp_sbc },
+		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_SBC), test_a2dp_sbc_invalid_config },
+		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_SBC), test_a2dp_sbc_pcm_drop },
 #if ENABLE_MP3LAME
 		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_MPEG12), test_a2dp_mp3 },
 #endif
@@ -992,7 +1119,8 @@ int main(int argc, char *argv[]) {
 		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_VENDOR_APTX_HD), test_a2dp_aptx_hd },
 #endif
 #if ENABLE_FASTSTREAM
-		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_VENDOR_FASTSTREAM), test_a2dp_faststream },
+		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_VENDOR_FASTSTREAM), test_a2dp_faststream_music },
+		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_VENDOR_FASTSTREAM), test_a2dp_faststream_voice },
 #endif
 #if ENABLE_LC3PLUS
 		{ a2dp_codecs_codec_id_to_string(A2DP_CODEC_VENDOR_LC3PLUS), test_a2dp_lc3plus },
@@ -1069,33 +1197,35 @@ int main(int argc, char *argv[]) {
 
 	if (input_bt_file != NULL) {
 
-		if ((btd = bt_dump_open(input_bt_file)) == NULL) {
+		if ((btdin = bt_dump_open(input_bt_file)) == NULL) {
 			error("Couldn't open input BT dump file: %s", strerror(errno));
 			return EXIT_FAILURE;
 		}
 
 		const char *codec = "";
-		switch (btd->mode) {
+		switch (btdin->mode) {
 		case BT_DUMP_MODE_A2DP_SOURCE:
 		case BT_DUMP_MODE_A2DP_SINK:
-			codec = a2dp_codecs_codec_id_to_string(btd->transport_codec_id);
-			debug("BT dump A2DP codec: %s (%#x)", codec, btd->transport_codec_id);
+			codec = a2dp_codecs_codec_id_to_string(btdin->transport_codec_id);
+			debug("BT dump A2DP codec: %s (%#x)", codec, btdin->transport_codec_id);
 			hexdump("BT dump A2DP configuration",
-					&btd->a2dp_configuration, btd->a2dp_configuration_size, true);
+					&btdin->a2dp_configuration, btdin->a2dp_configuration_size, true);
 			break;
 		case BT_DUMP_MODE_SCO:
-			codec = hfp_codec_id_to_string(btd->transport_codec_id);
-			debug("BT dump HFP codec: %s (%#x)", codec, btd->transport_codec_id);
+			codec = hfp_codec_id_to_string(btdin->transport_codec_id);
+			debug("BT dump HFP codec: %s (%#x)", codec, btdin->transport_codec_id);
 			break;
 		}
 
 		enabled_codecs = 0;
 		for (size_t i = 0; i < ARRAYSIZE(codecs); i++)
 			if (strcmp(codec, codecs[i].name) == 0)
-				enabled_codecs = 1 << i;
+				enabled_codecs |= 1 << i;
 
 	}
 
+	bluealsa_config_init();
+
 	bdaddr_t addr1 = {{ 1, 2, 3, 4, 5, 6 }};
 	bdaddr_t addr2 = {{ 1, 2, 3, 7, 8, 9 }};
 	adapter = ba_adapter_new(0);
@@ -1108,7 +1238,7 @@ int main(int argc, char *argv[]) {
 
 	suite_add_tcase(s, tc);
 
-	tcase_set_timeout(tc, aging_duration + 5);
+	tcase_set_timeout(tc, aging_duration + 10);
 	if (input_bt_file != NULL || input_pcm_file != NULL)
 		tcase_set_timeout(tc, aging_duration + 3600);
 
