@@ -1,6 +1,6 @@
 /*
  * test-utils.c
- * Copyright (c) 2016-2022 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -14,32 +14,31 @@
 
 #include <errno.h>
 #include <stdint.h>
+#include <stdlib.h>
 #include <string.h>
 #include <time.h>
 
 #include <bluetooth/bluetooth.h>
 #include <check.h>
 
-#include "ba-transport.h"
 #include "hci.h"
-#include "hfp.h"
 #include "utils.h"
-#include "shared/a2dp-codecs.h"
-#include "shared/defs.h"
 #include "shared/ffb.h"
 #include "shared/hex.h"
 #include "shared/nv.h"
 #include "shared/rt.h"
 
-START_TEST(test_g_dbus_bluez_object_path_to_hci_dev_id) {
+#include "inc/check.inc"
+
+CK_START_TEST(test_g_dbus_bluez_object_path_to_hci_dev_id) {
 
 	ck_assert_int_eq(g_dbus_bluez_object_path_to_hci_dev_id("/org/bluez"), -1);
 	ck_assert_int_eq(g_dbus_bluez_object_path_to_hci_dev_id("/org/bluez/hci0"), 0);
 	ck_assert_int_eq(g_dbus_bluez_object_path_to_hci_dev_id("/org/bluez/hci5"), 5);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_g_dbus_bluez_object_path_to_bdaddr) {
+CK_START_TEST(test_g_dbus_bluez_object_path_to_bdaddr) {
 
 	bdaddr_t addr_ok = {{ 0xBC, 0x9A, 0x78, 0x56, 0x34, 0x12 }};
 	bdaddr_t addr;
@@ -55,58 +54,9 @@ START_TEST(test_g_dbus_bluez_object_path_to_bdaddr) {
 	ck_assert_ptr_eq(g_dbus_bluez_object_path_to_bdaddr(
 				"/org/bluez/dev_12_34_56_78_9A_XX", &addr), NULL);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_dbus_profile_object_path) {
-
-	static const struct {
-		struct ba_transport_type ttype;
-		const char *path;
-	} profiles[] = {
-		/* test null/invalid path */
-		{ { 0, -1 }, "/" },
-		{ { 0, -1 }, "/Invalid" },
-		/* test A2DP profiles */
-		{ { BA_TRANSPORT_PROFILE_A2DP_SOURCE, A2DP_CODEC_SBC }, "/A2DP/SBC/source" },
-		{ { BA_TRANSPORT_PROFILE_A2DP_SOURCE, A2DP_CODEC_SBC }, "/A2DP/SBC/source/1" },
-		{ { BA_TRANSPORT_PROFILE_A2DP_SOURCE, A2DP_CODEC_SBC }, "/A2DP/SBC/source/2" },
-		{ { BA_TRANSPORT_PROFILE_A2DP_SINK, A2DP_CODEC_SBC }, "/A2DP/SBC/sink" },
-#if ENABLE_MPEG
-		{ { BA_TRANSPORT_PROFILE_A2DP_SOURCE, A2DP_CODEC_MPEG12 }, "/A2DP/MPEG/source" },
-		{ { BA_TRANSPORT_PROFILE_A2DP_SINK, A2DP_CODEC_MPEG12 }, "/A2DP/MPEG/sink" },
-#endif
-#if ENABLE_AAC
-		{ { BA_TRANSPORT_PROFILE_A2DP_SOURCE, A2DP_CODEC_MPEG24 }, "/A2DP/AAC/source" },
-		{ { BA_TRANSPORT_PROFILE_A2DP_SINK, A2DP_CODEC_MPEG24 }, "/A2DP/AAC/sink" },
-#endif
-#if ENABLE_APTX
-		{ { BA_TRANSPORT_PROFILE_A2DP_SOURCE, A2DP_CODEC_VENDOR_APTX }, "/A2DP/aptX/source" },
-		{ { BA_TRANSPORT_PROFILE_A2DP_SINK, A2DP_CODEC_VENDOR_APTX }, "/A2DP/aptX/sink" },
-#endif
-#if ENABLE_APTX_HD
-		{ { BA_TRANSPORT_PROFILE_A2DP_SOURCE, A2DP_CODEC_VENDOR_APTX_HD }, "/A2DP/aptXHD/source" },
-		{ { BA_TRANSPORT_PROFILE_A2DP_SINK, A2DP_CODEC_VENDOR_APTX_HD }, "/A2DP/aptXHD/sink" },
-#endif
-#if ENABLE_LDAC
-		{ { BA_TRANSPORT_PROFILE_A2DP_SOURCE, A2DP_CODEC_VENDOR_LDAC }, "/A2DP/LDAC/source" },
-		{ { BA_TRANSPORT_PROFILE_A2DP_SINK, A2DP_CODEC_VENDOR_LDAC }, "/A2DP/LDAC/sink" },
-#endif
-		/* test HSP/HFP profiles */
-		{ { BA_TRANSPORT_PROFILE_HSP_HS, HFP_CODEC_CVSD }, "/HSP/Headset" },
-		{ { BA_TRANSPORT_PROFILE_HSP_AG, HFP_CODEC_CVSD }, "/HSP/AudioGateway" },
-		{ { BA_TRANSPORT_PROFILE_HFP_HF, HFP_CODEC_UNDEFINED }, "/HFP/HandsFree" },
-		{ { BA_TRANSPORT_PROFILE_HFP_AG, HFP_CODEC_UNDEFINED }, "/HFP/AudioGateway" },
-	};
-
-	size_t i;
-	for (i = 0; i < ARRAYSIZE(profiles); i++) {
-		const char *path = g_dbus_transport_type_to_bluez_object_path(profiles[i].ttype);
-		ck_assert_str_eq(strstr(profiles[i].path, path), profiles[i].path);
-	}
-
-} END_TEST
-
-START_TEST(test_g_variant_sanitize_object_path) {
+CK_START_TEST(test_g_variant_sanitize_object_path) {
 
 	char path1[] = "/some/valid_path/123";
 	char path2[] = "/a#$*/invalid-path";
@@ -114,9 +64,10 @@ START_TEST(test_g_variant_sanitize_object_path) {
 	ck_assert_str_eq(g_variant_sanitize_object_path(path1), "/some/valid_path/123");
 	ck_assert_str_eq(g_variant_sanitize_object_path(path2), "/a___/invalid_path");
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_batostr_) {
+#if DEBUG
+CK_START_TEST(test_batostr_) {
 
 	const bdaddr_t ba = {{ 1, 2, 3, 4, 5, 6 }};
 	char tmp[18];
@@ -124,9 +75,10 @@ START_TEST(test_batostr_) {
 	ba2str(&ba, tmp);
 	ck_assert_str_eq(batostr_(&ba), tmp);
 
-} END_TEST
+} CK_END_TEST
+#endif
 
-START_TEST(test_nv_find) {
+CK_START_TEST(test_nv_find) {
 
 	const nv_entry_t entries[] = {
 		{ "name1", .v.i = 1 },
@@ -137,9 +89,9 @@ START_TEST(test_nv_find) {
 	ck_assert_ptr_eq(nv_find(entries, "invalid"), NULL);
 	ck_assert_ptr_eq(nv_find(entries, "name2"), &entries[1]);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_nv_join_names) {
+CK_START_TEST(test_nv_join_names) {
 
 	const nv_entry_t entries_zero[] = {{ 0 }};
 	const nv_entry_t entries[] = {
@@ -156,9 +108,9 @@ START_TEST(test_nv_join_names) {
 	ck_assert_str_eq(tmp = nv_join_names(entries), "name1, name2");
 	free(tmp);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_difftimespec) {
+CK_START_TEST(test_difftimespec) {
 
 	struct timespec ts1, ts2, ts;
 
@@ -216,9 +168,9 @@ START_TEST(test_difftimespec) {
 	ck_assert_int_eq(ts.tv_sec, 2);
 	ck_assert_int_eq(ts.tv_nsec, 0);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_ffb) {
+CK_START_TEST(test_ffb) {
 
 	ffb_t ffb_u8 = { 0 };
 	ffb_t ffb_16 = { 0 };
@@ -274,9 +226,9 @@ START_TEST(test_ffb) {
 	ffb_free(&ffb_16);
 	ck_assert_ptr_eq(ffb_16.data, NULL);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_ffb_resize) {
+CK_START_TEST(test_ffb_resize) {
 
 	const char *data = "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZ";
 	const size_t data_len = strlen(data);
@@ -299,9 +251,9 @@ START_TEST(test_ffb_resize) {
 
 	ffb_free(&ffb);
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_bin2hex) {
+CK_START_TEST(test_bin2hex) {
 
 	const uint8_t bin[] = { 0xDE, 0xAD, 0xBE, 0xEF };
 	char hex[sizeof(bin) * 2 + 1];
@@ -309,9 +261,9 @@ START_TEST(test_bin2hex) {
 	ck_assert_int_eq(bin2hex(bin, hex, sizeof(bin)), 8);
 	ck_assert_str_eq(hex, "deadbeef");
 
-} END_TEST
+} CK_END_TEST
 
-START_TEST(test_hex2bin) {
+CK_START_TEST(test_hex2bin) {
 
 	const uint8_t bin_ok[] = { 0xDE, 0xAD, 0xBE, 0xEF };
 	const char *hex = "DEADbeef";
@@ -323,7 +275,7 @@ START_TEST(test_hex2bin) {
 	ck_assert_int_eq(hex2bin(hex, bin, 3), -1);
 	ck_assert_int_eq(errno, EINVAL);
 
-} END_TEST
+} CK_END_TEST
 
 int main(void) {
 
@@ -350,9 +302,10 @@ int main(void) {
 
 	tcase_add_test(tc, test_g_dbus_bluez_object_path_to_hci_dev_id);
 	tcase_add_test(tc, test_g_dbus_bluez_object_path_to_bdaddr);
-	tcase_add_test(tc, test_dbus_profile_object_path);
 	tcase_add_test(tc, test_g_variant_sanitize_object_path);
+#if DEBUG
 	tcase_add_test(tc, test_batostr_);
+#endif
 
 	srunner_run_all(sr, CK_ENV);
 	int nf = srunner_ntests_failed(sr);
