@@ -1,6 +1,6 @@
 /*
  * BlueALSA - aplay.c
- * Copyright (c) 2016-2021 Arkadiusz Bokowy
+ * Copyright (c) 2016-2023 Arkadiusz Bokowy
  *
  * This file is a part of bluez-alsa.
  *
@@ -19,10 +19,12 @@
 #include <poll.h>
 #include <pthread.h>
 #include <signal.h>
+#include <stdatomic.h>
 #include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/param.h>
 #include <unistd.h>
 
 #include <alsa/asoundlib.h>
@@ -37,7 +39,7 @@
 #include "alsa-pcm.h"
 #include "dbus.h"
 
-struct pcm_worker {
+struct io_worker {
 	pthread_t thread;
 	/* used BlueALSA PCM device */
 	struct ba_pcm ba_pcm;
@@ -46,12 +48,13 @@ struct pcm_worker {
 	/* file descriptor of PCM control */
 	int ba_pcm_ctrl_fd;
 	/* opened playback PCM device */
-	snd_pcm_t *pcm;
+	snd_pcm_t *snd_pcm;
 	/* mixer for volume control */
-	snd_mixer_t *mixer;
-	snd_mixer_elem_t *mixer_elem;
+	snd_mixer_t *snd_mixer;
+	snd_mixer_elem_t *snd_mixer_elem;
+	bool mixer_has_mute_switch;
 	/* if true, playback is active */
-	bool active;
+	atomic_bool active;
 	/* human-readable BT address */
 	char addr[18];
 };
@@ -69,7 +72,9 @@ static bdaddr_t *ba_addrs = NULL;
 static size_t ba_addrs_count = 0;
 static unsigned int pcm_buffer_time = 500000;
 static unsigned int pcm_period_time = 100000;
-static bool pcm_mixer = true;
+
+/* local PCM muted state for software mute */
+static bool pcm_muted = false;
 
 static struct ba_dbus_ctx dbus_ctx;
 static char dbus_ba_service[32] = BLUEALSA_SERVICE;
@@ -77,12 +82,15 @@ static char dbus_ba_service[32] = BLUEALSA_SERVICE;
 static struct ba_pcm *ba_pcms = NULL;
 static size_t ba_pcms_count = 0;
 
+static pthread_mutex_t single_playback_mutex = PTHREAD_MUTEX_INITIALIZER;
+static bool force_single_playback = false;
+
 static pthread_rwlock_t workers_lock = PTHREAD_RWLOCK_INITIALIZER;
-static struct pcm_worker *workers = NULL;
+static struct io_worker *workers = NULL;
 static size_t workers_count = 0;
 static size_t workers_size = 0;
 
-static bool main_loop_on = true;
+static atomic_bool main_loop_on = true;
 static void main_loop_stop(int sig) {
 	/* Call to this handler restores the default action, so on the
 	 * second call the program will be forcefully terminated. */
@@ -180,17 +188,17 @@ static void print_bt_device_list(void) {
 				ba2str(&dev.bt_addr, bt_addr);
 
 				printf("%s: %s [%s], %s%s\n",
-					dev.hci_name, bt_addr, dev.name,
-					dev.trusted ? "trusted ": "", dev.icon);
+						dev.hci_name, bt_addr, dev.name,
+						dev.trusted ? "trusted " : "", dev.icon);
 
 			}
 
 			printf("  %s (%s): %s %d channel%s %d Hz\n",
-				bluealsa_get_profile(pcm),
-				pcm->codec,
-				snd_pcm_format_name(bluealsa_get_snd_pcm_format(pcm)),
-				pcm->channels, pcm->channels != 1 ? "s" : "",
-				pcm->sampling);
+					bluealsa_get_profile(pcm),
+					pcm->codec.name,
+					snd_pcm_format_name(bluealsa_get_snd_pcm_format(pcm)),
+					pcm->channels, pcm->channels != 1 ? "s" : "",
+					pcm->sampling);
 
 		}
 	}
@@ -219,39 +227,54 @@ static void print_bt_pcm_list(void) {
 		ba2str(&dev.bt_addr, bt_addr);
 
 		printf(
-				"bluealsa:SRV=%s,DEV=%s,PROFILE=%s\n"
+				"bluealsa:DEV=%s,PROFILE=%s,SRV=%s\n"
 				"    %s, %s%s, %s\n"
 				"    %s (%s): %s %d channel%s %d Hz\n",
-			dbus_ba_service,
-			bt_addr,
-			pcm->transport & BA_PCM_TRANSPORT_MASK_A2DP ? "a2dp" : "sco",
-			dev.name,
-			dev.trusted ? "trusted ": "", dev.icon,
-			pcm->mode == BA_PCM_MODE_SINK ? "playback" : "capture",
-			bluealsa_get_profile(pcm),
-			pcm->codec,
-			snd_pcm_format_name(bluealsa_get_snd_pcm_format(pcm)),
-			pcm->channels, pcm->channels != 1 ? "s" : "",
-			pcm->sampling);
+				bt_addr,
+				pcm->transport & BA_PCM_TRANSPORT_MASK_A2DP ? "a2dp" : "sco",
+				dbus_ba_service,
+				dev.name,
+				dev.trusted ? "trusted " : "", dev.icon,
+				pcm->mode == BA_PCM_MODE_SINK ? "playback" : "capture",
+				bluealsa_get_profile(pcm),
+				pcm->codec.name,
+				snd_pcm_format_name(bluealsa_get_snd_pcm_format(pcm)),
+				pcm->channels, pcm->channels != 1 ? "s" : "",
+				pcm->sampling);
 
 	}
 
 }
 
-static struct ba_pcm *get_ba_pcm(const char *path) {
+static struct ba_pcm *ba_pcm_add(const struct ba_pcm *pcm) {
+	struct ba_pcm *tmp;
+	if ((tmp = realloc(ba_pcms, (ba_pcms_count + 1) * sizeof(*ba_pcms))) == NULL)
+		return NULL;
+	ba_pcms = tmp;
+	memcpy(&ba_pcms[ba_pcms_count], pcm, sizeof(*ba_pcms));
+	return &ba_pcms[ba_pcms_count++];
+}
 
-	size_t i;
-
-	for (i = 0; i < ba_pcms_count; i++)
+static struct ba_pcm *ba_pcm_get(const char *path) {
+	for (size_t i = 0; i < ba_pcms_count; i++)
 		if (strcmp(ba_pcms[i].pcm_path, path) == 0)
 			return &ba_pcms[i];
-
 	return NULL;
 }
 
-static struct pcm_worker *get_active_worker(void) {
+static void ba_pcm_remove(const char *path) {
+	for (size_t i = 0; i < ba_pcms_count; i++)
+		if (strcmp(ba_pcms[i].pcm_path, path) == 0) {
+			memmove(&ba_pcms[i], &ba_pcms[i + 1],
+					(ba_pcms_count - i - 1) * sizeof(*ba_pcms));
+			ba_pcms_count--;
+			break;
+		}
+}
 
-	struct pcm_worker *w = NULL;
+static struct io_worker *get_active_io_worker(void) {
+
+	struct io_worker *w = NULL;
 	size_t i;
 
 	pthread_rwlock_rdlock(&workers_lock);
@@ -299,19 +322,12 @@ final:
 }
 
 /**
- * Synchronize BlueALSA PCM volume with ALSA mixer element. */
-static int pcm_worker_mixer_volume_sync(
-		struct pcm_worker *worker,
+ * Update BlueALSA PCM volume according to ALSA mixer element. */
+static int io_worker_mixer_volume_sync_ba_pcm(
+		struct io_worker *worker,
 		struct ba_pcm *ba_pcm) {
 
-	/* skip sync in case of software volume */
-	if (ba_pcm->soft_volume)
-		return 0;
-
-	snd_mixer_elem_t *elem = worker->mixer_elem;
-	if (elem == NULL)
-		return 0;
-
+	snd_mixer_elem_t *elem = worker->snd_mixer_elem;
 	const int vmax = BA_PCM_VOLUME_MAX(ba_pcm);
 	long long volume_db_sum = 0;
 	bool muted = true;
@@ -320,15 +336,22 @@ static int pcm_worker_mixer_volume_sync(
 	for (ch = 0; snd_mixer_selem_has_playback_channel(elem, ch) == 1; ch++) {
 
 		long ch_volume_db;
-		int ch_switch;
+		int ch_switch = 1;
 
 		int err;
-		if ((err = snd_mixer_selem_get_playback_dB(elem, 0, &ch_volume_db)) != 0 ||
-				(err = snd_mixer_selem_get_playback_switch(elem, 0, &ch_switch)) != 0) {
-			error("Couldn't get playback volume: %s", snd_strerror(err));
+		if ((err = snd_mixer_selem_get_playback_dB(elem, 0, &ch_volume_db)) != 0) {
+			error("Couldn't get ALSA mixer playback dB level: %s", snd_strerror(err));
 			return -1;
 		}
 
+		/* mute switch is an optional feature for a mixer element */
+		if ((worker->mixer_has_mute_switch = snd_mixer_selem_has_playback_switch(elem))) {
+			if ((err = snd_mixer_selem_get_playback_switch(elem, 0, &ch_switch)) != 0) {
+				error("Couldn't get ALSA mixer playback switch: %s", snd_strerror(err));
+				return -1;
+			}
+		}
+
 		volume_db_sum += ch_volume_db;
 		if (ch_switch == 1)
 			muted = false;
@@ -343,6 +366,11 @@ static int pcm_worker_mixer_volume_sync(
 	 * round to the nearest integer. */
 	int volume = lround(pow(2, (0.01 * volume_db_sum / ch) / 10) * vmax);
 
+	/* If mixer element does not support playback switch,
+	 * use our global muted state. */
+	if (!worker->mixer_has_mute_switch)
+		muted = pcm_muted;
+
 	ba_pcm->volume.ch1_muted = muted;
 	ba_pcm->volume.ch1_volume = volume;
 	ba_pcm->volume.ch2_muted = muted;
@@ -350,7 +378,7 @@ static int pcm_worker_mixer_volume_sync(
 
 	DBusError err = DBUS_ERROR_INIT;
 	if (!bluealsa_dbus_pcm_update(&dbus_ctx, ba_pcm, BLUEALSA_PCM_VOLUME, &err)) {
-		error("Couldn't update PCM: %s", err.message);
+		error("Couldn't update BlueALSA source PCM: %s", err.message);
 		dbus_error_free(&err);
 		return -1;
 	}
@@ -360,15 +388,15 @@ static int pcm_worker_mixer_volume_sync(
 
 /**
  * Update ALSA mixer element according to BlueALSA PCM volume. */
-static int pcm_worker_mixer_volume_update(
-		struct pcm_worker *worker,
+static int io_worker_mixer_volume_sync_snd_mixer_elem(
+		struct io_worker *worker,
 		struct ba_pcm *ba_pcm) {
 
 	/* skip update in case of software volume */
 	if (ba_pcm->soft_volume)
 		return 0;
 
-	snd_mixer_elem_t *elem = worker->mixer_elem;
+	snd_mixer_elem_t *elem = worker->snd_mixer_elem;
 	if (elem == NULL)
 		return 0;
 
@@ -388,20 +416,60 @@ static int pcm_worker_mixer_volume_update(
 		muted = ba_pcm->volume.ch1_muted || ba_pcm->volume.ch2_muted;
 	}
 
+	/* keep local muted state up to date */
+	pcm_muted = muted;
+
 	/* convert loudness to dB using decibel formula */
 	long db = 10 * log2(1.0 * volume / vmax) * 100;
 
 	int err;
-	if ((err = snd_mixer_selem_set_playback_dB_all(elem, db, 0)) != 0 ||
+	if ((err = snd_mixer_selem_set_playback_dB_all(elem, db, 0)) != 0) {
+		error("Couldn't set ALSA mixer playback dB level: %s", snd_strerror(err));
+		return -1;
+	}
+
+	/* mute switch is an optional feature for a mixer element */
+	if (worker->mixer_has_mute_switch &&
 			(err = snd_mixer_selem_set_playback_switch_all(elem, !muted)) != 0) {
-		error("Couldn't set playback volume: %s", snd_strerror(err));
+		error("Couldn't set ALSA mixer playback mute switch: %s", snd_strerror(err));
 		return -1;
 	}
 
 	return 0;
 }
 
-static void pcm_worker_routine_exit(struct pcm_worker *worker) {
+int io_worker_mixer_elem_callback(snd_mixer_elem_t *elem, unsigned int mask) {
+	struct io_worker *worker = snd_mixer_elem_get_callback_private(elem);
+	if (mask & SND_CTL_EVENT_MASK_VALUE)
+		io_worker_mixer_volume_sync_ba_pcm(worker, &worker->ba_pcm);
+	return 0;
+}
+
+/**
+ * Setup volume synchronization between ALSA mixer and BlueALSA PCM. */
+static int io_worker_mixer_volume_sync_setup(
+		struct io_worker *worker) {
+
+	/* skip setup in case of software volume */
+	if (worker->ba_pcm.soft_volume)
+		return 0;
+
+	snd_mixer_elem_t *elem = worker->snd_mixer_elem;
+	if (elem == NULL)
+		return 0;
+
+	debug("Setting up ALSA mixer volume synchronization");
+
+	snd_mixer_elem_set_callback(elem, io_worker_mixer_elem_callback);
+	snd_mixer_elem_set_callback_private(elem, worker);
+
+	/* initial synchronization */
+	io_worker_mixer_volume_sync_ba_pcm(worker, &worker->ba_pcm);
+
+	return 0;
+}
+
+static void io_worker_routine_exit(struct io_worker *worker) {
 	if (worker->ba_pcm_fd != -1) {
 		close(worker->ba_pcm_fd);
 		worker->ba_pcm_fd = -1;
@@ -410,19 +478,19 @@ static void pcm_worker_routine_exit(struct pcm_worker *worker) {
 		close(worker->ba_pcm_ctrl_fd);
 		worker->ba_pcm_ctrl_fd = -1;
 	}
-	if (worker->pcm != NULL) {
-		snd_pcm_close(worker->pcm);
-		worker->pcm = NULL;
+	if (worker->snd_pcm != NULL) {
+		snd_pcm_close(worker->snd_pcm);
+		worker->snd_pcm = NULL;
 	}
-	if (worker->mixer != NULL) {
-		snd_mixer_close(worker->mixer);
-		worker->mixer_elem = NULL;
-		worker->mixer = NULL;
+	if (worker->snd_mixer != NULL) {
+		snd_mixer_close(worker->snd_mixer);
+		worker->snd_mixer_elem = NULL;
+		worker->snd_mixer = NULL;
 	}
-	debug("Exiting PCM worker %s", worker->addr);
+	debug("Exiting IO worker %s", worker->addr);
 }
 
-static void *pcm_worker_routine(struct pcm_worker *w) {
+static void *io_worker_routine(struct io_worker *w) {
 
 	snd_pcm_format_t pcm_format = bluealsa_get_snd_pcm_format(&w->ba_pcm);
 	ssize_t pcm_format_size = snd_pcm_format_size(pcm_format, 1);
@@ -433,7 +501,7 @@ static void *pcm_worker_routine(struct pcm_worker *w) {
 	 * in order to prevent memory leaks and resources not being released. */
 	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 
-	pthread_cleanup_push(PTHREAD_CLEANUP(pcm_worker_routine_exit), w);
+	pthread_cleanup_push(PTHREAD_CLEANUP(io_worker_routine_exit), w);
 	pthread_cleanup_push(PTHREAD_CLEANUP(ffb_free), &buffer);
 
 	/* create buffer big enough to hold 100 ms of PCM data */
@@ -443,9 +511,10 @@ static void *pcm_worker_routine(struct pcm_worker *w) {
 	}
 
 	DBusError err = DBUS_ERROR_INIT;
+	debug("Opening BlueALSA source PCM: %s", w->ba_pcm.pcm_path);
 	if (!bluealsa_dbus_pcm_open(&dbus_ctx, w->ba_pcm.pcm_path,
 				&w->ba_pcm_fd, &w->ba_pcm_ctrl_fd, &err)) {
-		error("Couldn't open PCM: %s", err.message);
+		error("Couldn't open BlueALSA source PCM: %s", err.message);
 		dbus_error_free(&err);
 		goto fail;
 	}
@@ -454,163 +523,247 @@ static void *pcm_worker_routine(struct pcm_worker *w) {
 	 * will be opened, this value will be adjusted to one period size. */
 	size_t pcm_max_read_len_init = pcm_1s_samples / 100 * pcm_format_size;
 	size_t pcm_max_read_len = pcm_max_read_len_init;
+
+	/* Track the lock state of the single playback mutex within this thread. */
+	bool single_playback_mutex_locked = false;
+
+	/* Intervals in seconds between consecutive PCM open retry attempts. */
+	const unsigned int pcm_open_retry_intervals[] = { 1, 1, 2, 3, 5 };
+	size_t pcm_open_retry_pcm_samples = 0;
 	size_t pcm_open_retries = 0;
 
-	/* These variables determine how and when the pause command will be send
-	 * to the device player. In order not to flood BT connection with AVRCP
-	 * packets, we are going to send pause command every 0.5 second. */
-	size_t pause_threshold = pcm_1s_samples / 2 * pcm_format_size;
-	size_t pause_counter = 0;
-	size_t pause_bytes = 0;
+	size_t pause_retry_pcm_samples = pcm_1s_samples;
+	size_t pause_retries = 0;
 
-	struct pollfd pfds[] = {{ w->ba_pcm_fd, POLLIN, 0 }};
 	int timeout = -1;
 
-	debug("Starting PCM loop");
+	debug("Starting IO loop");
 	while (main_loop_on) {
-		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
 
-		ssize_t ret;
+		if (single_playback_mutex_locked) {
+			pthread_mutex_unlock(&single_playback_mutex);
+			single_playback_mutex_locked = false;
+		}
+
+		struct pollfd fds[16] = {{ w->ba_pcm_fd, POLLIN, 0 }};
+		nfds_t nfds = 1;
+
+		if (w->snd_mixer != NULL)
+			nfds += snd_mixer_poll_descriptors_count(w->snd_mixer);
+
+		if (nfds > ARRAYSIZE(fds)) {
+			error("Poll FD array size exceeded: %zu > %zu", nfds, ARRAYSIZE(fds));
+			goto fail;
+		}
+
+		if (w->snd_mixer != NULL)
+			snd_mixer_poll_descriptors(w->snd_mixer, fds + 1, nfds - 1);
 
 		/* Reading from the FIFO won't block unless there is an open connection
 		 * on the writing side. However, the server does not open PCM FIFO until
 		 * a transport is created. With the A2DP, the transport is created when
 		 * some clients (BT device) requests audio transfer. */
-		switch (poll(pfds, ARRAYSIZE(pfds), timeout)) {
+
+		pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
+		int poll_rv = poll(fds, nfds, timeout);
+		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
+
+		switch (poll_rv) {
 		case -1:
 			if (errno == EINTR)
 				continue;
-			error("PCM FIFO poll error: %s", strerror(errno));
+			error("IO loop poll error: %s", strerror(errno));
 			goto fail;
 		case 0:
-			debug("Device marked as inactive: %s", w->addr);
-			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
-			pcm_max_read_len = pcm_max_read_len_init;
-			pause_counter = pause_bytes = 0;
-			ffb_rewind(&buffer);
-			if (w->pcm != NULL) {
-				snd_pcm_close(w->pcm);
-				w->pcm = NULL;
-			}
+			debug("BT device marked as inactive: %s", w->addr);
+			pause_retry_pcm_samples = pcm_1s_samples;
+			pause_retries = 0;
 			w->active = false;
 			timeout = -1;
-			continue;
+			goto close_alsa;
 		}
 
-		/* FIFO has been terminated on the writing side */
-		if (pfds[0].revents & POLLHUP)
+		if (w->snd_mixer != NULL)
+			snd_mixer_handle_events(w->snd_mixer);
+
+		size_t read_samples = 0;
+		if (fds[0].revents & POLLIN) {
+
+			ssize_t ret;
+			size_t _in = MIN(pcm_max_read_len, ffb_blen_in(&buffer));
+			if ((ret = read(w->ba_pcm_fd, buffer.tail, _in)) == -1) {
+				if (errno == EINTR)
+					continue;
+				error("BlueALSA source PCM read error: %s", strerror(errno));
+				goto fail;
+			}
+
+			read_samples = ret / pcm_format_size;
+			if (ret % pcm_format_size != 0)
+				warn("Invalid read from BlueALSA source PCM: %zd %% %zd != 0", ret, pcm_format_size);
+
+		}
+		else if (fds[0].revents & POLLHUP) {
+			/* source PCM FIFO has been terminated on the writing side */
+			debug("BlueALSA source PCM disconnected: %s", w->ba_pcm.pcm_path);
 			break;
-
-		#define MIN(a,b) a < b ? a : b
-		size_t _in = MIN(pcm_max_read_len, ffb_blen_in(&buffer));
-		if ((ret = read(w->ba_pcm_fd, buffer.tail, _in)) == -1) {
-			if (errno == EINTR)
-				continue;
-			error("PCM FIFO read error: %s", strerror(errno));
-			goto fail;
 		}
+		else if (fds[0].revents)
+			error("Unexpected BlueALSA source PCM poll event: %#x", fds[0].revents);
 
-		pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
+		if (read_samples == 0)
+			continue;
 
-		/* If PCM mixer is disabled, check whether we should play audio. */
-		if (!pcm_mixer) {
-			struct pcm_worker *worker = get_active_worker();
-			if (worker != NULL && worker != w) {
-				if (pause_counter < 5 && (pause_bytes += ret) > pause_threshold) {
+		/* If current worker is not active and the single playback mode was
+		 * enabled, we have to check if there is any other active worker. */
+		if (force_single_playback && !w->active) {
+
+			/* Before checking active worker, we need to lock the single playback
+			 * mutex. It is required to lock it, because the active state is changed
+			 * in the worker thread after opening the PCM device, so we have to
+			 * synchronize all threads at this point. */
+			pthread_mutex_lock(&single_playback_mutex);
+			single_playback_mutex_locked = true;
+
+			if (get_active_io_worker() != NULL) {
+				/* In order not to flood BT connection with AVRCP packets,
+				 * we are going to send pause command every 0.5 second. */
+				if (pause_retries < 5 &&
+						(pause_retry_pcm_samples += read_samples) > pcm_1s_samples / 2) {
 					if (pause_device_player(&w->ba_pcm) == -1)
 						/* pause command does not work, stop further requests */
-						pause_counter = 5;
-					pause_counter++;
-					pause_bytes = 0;
+						pause_retries = 5;
+					pause_retry_pcm_samples = 0;
+					pause_retries++;
 					timeout = 100;
 				}
 				continue;
 			}
+
 		}
 
-		if (w->pcm == NULL) {
+		if (w->snd_pcm == NULL) {
 
 			unsigned int buffer_time = pcm_buffer_time;
 			unsigned int period_time = pcm_period_time;
-			snd_pcm_uframes_t buffer_size;
-			snd_pcm_uframes_t period_size;
+			snd_pcm_uframes_t buffer_frames;
+			snd_pcm_uframes_t period_frames;
 			char *tmp;
 
-			/* After PCM open failure wait one second before retry. This can not be
-			 * done with a single sleep() call, because we have to drain PCM FIFO. */
-			if (pcm_open_retries++ % 20 != 0) {
-				usleep(50000);
-				continue;
+			if (pcm_open_retries > 0) {
+				/* After PCM open failure wait some time before retry. This can not be
+				 * done with a sleep() call, because we have to drain PCM FIFO, so it
+				 * will not have any stale data. */
+				unsigned int interval = pcm_open_retries > ARRAYSIZE(pcm_open_retry_intervals) ?
+					pcm_open_retry_intervals[ARRAYSIZE(pcm_open_retry_intervals) - 1] :
+					pcm_open_retry_intervals[pcm_open_retries - 1];
+				if ((pcm_open_retry_pcm_samples += read_samples) <= interval * pcm_1s_samples)
+					continue;
 			}
 
-			if (alsa_pcm_open(&w->pcm, pcm_device, pcm_format, w->ba_pcm.channels,
+			debug("Opening ALSA playback PCM: name=%s channels=%u rate=%u",
+					pcm_device, w->ba_pcm.channels, w->ba_pcm.sampling);
+			if (alsa_pcm_open(&w->snd_pcm, pcm_device, pcm_format, w->ba_pcm.channels,
 						w->ba_pcm.sampling, &buffer_time, &period_time, &tmp) != 0) {
-				warn("Couldn't open PCM: %s", tmp);
+				warn("Couldn't open ALSA playback PCM: %s", tmp);
 				pcm_max_read_len = pcm_max_read_len_init;
-				usleep(50000);
+				pcm_open_retry_pcm_samples = 0;
+				pcm_open_retries++;
 				free(tmp);
 				continue;
 			}
 
-			if (alsa_mixer_open(&w->mixer, &w->mixer_elem,
+			snd_pcm_get_params(w->snd_pcm, &buffer_frames, &period_frames);
+			pcm_max_read_len = period_frames * w->ba_pcm.channels * pcm_format_size;
+
+			debug("Opening ALSA mixer: name=%s elem=%s index=%u",
+					mixer_device, mixer_elem_name, mixer_elem_index);
+			if (alsa_mixer_open(&w->snd_mixer, &w->snd_mixer_elem,
 						mixer_device, mixer_elem_name, mixer_elem_index, &tmp) != 0) {
-				warn("Couldn't open mixer: %s", tmp);
+				warn("Couldn't open ALSA mixer: %s", tmp);
 				free(tmp);
 			}
 
-			/* initial volume synchronization */
-			pcm_worker_mixer_volume_sync(w, &w->ba_pcm);
+			io_worker_mixer_volume_sync_setup(w);
 
-			snd_pcm_get_params(w->pcm, &buffer_size, &period_size);
-			pcm_max_read_len = period_size * w->ba_pcm.channels * pcm_format_size;
+			/* reset retry counters */
+			pcm_open_retry_pcm_samples = 0;
 			pcm_open_retries = 0;
 
 			if (verbose >= 2) {
-				printf("Used configuration for %s:\n"
-						"  PCM buffer time: %u us (%zu bytes)\n"
-						"  PCM period time: %u us (%zu bytes)\n"
+				info("Used configuration for %s:\n"
+						"  ALSA PCM buffer time: %u us (%zu bytes)\n"
+						"  ALSA PCM period time: %u us (%zu bytes)\n"
 						"  PCM format: %s\n"
 						"  Sampling rate: %u Hz\n"
-						"  Channels: %u\n",
+						"  Channels: %u",
 						w->addr,
-						buffer_time, snd_pcm_frames_to_bytes(w->pcm, buffer_size),
-						period_time, snd_pcm_frames_to_bytes(w->pcm, period_size),
+						buffer_time, snd_pcm_frames_to_bytes(w->snd_pcm, buffer_frames),
+						period_time, snd_pcm_frames_to_bytes(w->snd_pcm, period_frames),
 						snd_pcm_format_name(pcm_format),
 						w->ba_pcm.sampling,
 						w->ba_pcm.channels);
 			}
 
+			if (verbose >= 3)
+				alsa_pcm_dump(w->snd_pcm, stderr);
+
 		}
 
 		/* mark device as active and set timeout to 500ms */
 		w->active = true;
 		timeout = 500;
 
-		ffb_seek(&buffer, ret / pcm_format_size);
+		/* Current worker was marked as active, so we can safely
+		 * release the single playback mutex if it was locked. */
+		if (single_playback_mutex_locked) {
+			pthread_mutex_unlock(&single_playback_mutex);
+			single_playback_mutex_locked = false;
+		}
 
-		/* calculate the overall number of frames in the buffer */
-		snd_pcm_sframes_t frames = ffb_len_out(&buffer) / w->ba_pcm.channels;
+		ffb_seek(&buffer, read_samples);
+		size_t samples = ffb_len_out(&buffer);
 
-		if ((frames = snd_pcm_writei(w->pcm, buffer.data, frames)) < 0)
+		if (!w->mixer_has_mute_switch && pcm_muted)
+			snd_pcm_format_set_silence(pcm_format, buffer.data, samples);
+
+		snd_pcm_sframes_t frames;
+
+retry_alsa_write:
+		frames = samples / w->ba_pcm.channels;
+		if ((frames = snd_pcm_writei(w->snd_pcm, buffer.data, frames)) < 0)
 			switch (-frames) {
+			case EINTR:
+				goto retry_alsa_write;
 			case EPIPE:
-				debug("An underrun has occurred");
-				snd_pcm_prepare(w->pcm);
-				usleep(50000);
-				frames = 0;
-				break;
+				debug("ALSA playback PCM underrun");
+				snd_pcm_prepare(w->snd_pcm);
+				goto retry_alsa_write;
 			default:
-				error("Couldn't write to PCM: %s", snd_strerror(frames));
-				goto fail;
+				error("ALSA playback PCM write error: %s", snd_strerror(frames));
+				goto close_alsa;
 			}
 
 		/* move leftovers to the beginning and reposition tail */
 		ffb_shift(&buffer, frames * w->ba_pcm.channels);
 
+		continue;
+
+close_alsa:
+		ffb_rewind(&buffer);
+		pcm_max_read_len = pcm_max_read_len_init;
+		if (w->snd_pcm != NULL) {
+			snd_pcm_close(w->snd_pcm);
+			w->snd_pcm = NULL;
+		}
+		if (w->snd_mixer != NULL) {
+			snd_mixer_close(w->snd_mixer);
+			w->snd_mixer_elem = NULL;
+			w->snd_mixer = NULL;
+		}
 	}
 
 fail:
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
 	pthread_cleanup_pop(1);
 	pthread_cleanup_pop(1);
 	return NULL;
@@ -630,7 +783,7 @@ static bool pcm_hw_params_equal(
 
 /**
  * Stop the worker thread at workers[index]. */
-static void pcm_worker_stop(size_t index) {
+static void io_worker_stop(size_t index) {
 
 	/* Safety check for out-of-bounds read. */
 	assert(index < workers_count);
@@ -649,7 +802,7 @@ static void pcm_worker_stop(size_t index) {
 
 }
 
-static struct pcm_worker *supervise_pcm_worker_start(const struct ba_pcm *ba_pcm) {
+static struct io_worker *supervise_io_worker_start(const struct ba_pcm *ba_pcm) {
 
 	size_t i;
 	for (i = 0; i < workers_count; i++)
@@ -658,7 +811,7 @@ static struct pcm_worker *supervise_pcm_worker_start(const struct ba_pcm *ba_pcm
 			 * audio format may have changed. If it has, the worker thread
 			 * needs to be restarted. */
 			if (!pcm_hw_params_equal(&workers[i].ba_pcm, ba_pcm))
-				pcm_worker_stop(i);
+				io_worker_stop(i);
 			else
 				return &workers[i];
 		}
@@ -667,31 +820,34 @@ static struct pcm_worker *supervise_pcm_worker_start(const struct ba_pcm *ba_pcm
 
 	workers_count++;
 	if (workers_size < workers_count) {
-		struct pcm_worker *tmp = workers;
+		struct io_worker *tmp = workers;
 		workers_size += 4;  /* coarse-grained realloc */
 		if ((workers = realloc(workers, sizeof(*workers) * workers_size)) == NULL) {
-			error("Couldn't (re)allocate memory for PCM workers: %s", strerror(ENOMEM));
+			error("Couldn't (re)allocate memory for IO workers: %s", strerror(ENOMEM));
 			workers = tmp;
 			pthread_rwlock_unlock(&workers_lock);
 			return NULL;
 		}
 	}
 
-	struct pcm_worker *worker = &workers[workers_count - 1];
+	struct io_worker *worker = &workers[workers_count - 1];
 	memcpy(&worker->ba_pcm, ba_pcm, sizeof(worker->ba_pcm));
 	ba2str(&worker->ba_pcm.addr, worker->addr);
-	worker->active = false;
 	worker->ba_pcm_fd = -1;
 	worker->ba_pcm_ctrl_fd = -1;
-	worker->pcm = NULL;
+	worker->snd_pcm = NULL;
+	worker->snd_mixer = NULL;
+	worker->snd_mixer_elem = NULL;
+	worker->mixer_has_mute_switch = false;
+	worker->active = false;
 
 	pthread_rwlock_unlock(&workers_lock);
 
-	debug("Creating PCM worker %s", worker->addr);
+	debug("Creating IO worker %s", worker->addr);
 
 	if ((errno = pthread_create(&worker->thread, NULL,
-					PTHREAD_ROUTINE(pcm_worker_routine), worker)) != 0) {
-		error("Couldn't create PCM worker %s: %s", worker->addr, strerror(errno));
+					PTHREAD_FUNC(io_worker_routine), worker)) != 0) {
+		error("Couldn't create IO worker %s: %s", worker->addr, strerror(errno));
 		workers_count--;
 		return NULL;
 	}
@@ -699,17 +855,17 @@ static struct pcm_worker *supervise_pcm_worker_start(const struct ba_pcm *ba_pcm
 	return worker;
 }
 
-static struct pcm_worker *supervise_pcm_worker_stop(const struct ba_pcm *ba_pcm) {
+static struct io_worker *supervise_io_worker_stop(const struct ba_pcm *ba_pcm) {
 
 	size_t i;
 	for (i = 0; i < workers_count; i++)
 		if (strcmp(workers[i].ba_pcm.pcm_path, ba_pcm->pcm_path) == 0)
-			pcm_worker_stop(i);
+			io_worker_stop(i);
 
 	return NULL;
 }
 
-static struct pcm_worker *supervise_pcm_worker(const struct ba_pcm *ba_pcm) {
+static struct io_worker *supervise_io_worker(const struct ba_pcm *ba_pcm) {
 
 	if (ba_pcm == NULL)
 		return NULL;
@@ -737,9 +893,9 @@ static struct pcm_worker *supervise_pcm_worker(const struct ba_pcm *ba_pcm) {
 			goto start;
 
 stop:
-	return supervise_pcm_worker_stop(ba_pcm);
+	return supervise_io_worker_stop(ba_pcm);
 start:
-	return supervise_pcm_worker_start(ba_pcm);
+	return supervise_io_worker_start(ba_pcm);
 }
 
 static DBusHandlerResult dbus_signal_handler(DBusConnection *conn, DBusMessage *message, void *data) {
@@ -754,7 +910,7 @@ static DBusHandlerResult dbus_signal_handler(DBusConnection *conn, DBusMessage *
 	const char *signal = dbus_message_get_member(message);
 
 	DBusMessageIter iter;
-	struct pcm_worker *worker;
+	struct io_worker *worker;
 
 	if (strcmp(interface, DBUS_INTERFACE_OBJECT_MANAGER) == 0) {
 
@@ -764,34 +920,32 @@ static DBusHandlerResult dbus_signal_handler(DBusConnection *conn, DBusMessage *
 			struct ba_pcm pcm;
 			DBusError err = DBUS_ERROR_INIT;
 			if (!bluealsa_dbus_message_iter_get_pcm(&iter, &err, &pcm)) {
-				error("Couldn't add new PCM: %s", err.message);
+				error("Couldn't add new BlueALSA PCM: %s", err.message);
 				dbus_error_free(&err);
 				goto fail;
 			}
 			if (pcm.transport == BA_PCM_TRANSPORT_NONE)
 				goto fail;
-			struct ba_pcm *tmp = ba_pcms;
-			if ((ba_pcms = realloc(ba_pcms, (ba_pcms_count + 1) * sizeof(*ba_pcms))) == NULL) {
-				error("Couldn't add new PCM: %s", strerror(ENOMEM));
-				ba_pcms = tmp;
+			if (ba_pcm_add(&pcm) == NULL) {
+				error("Couldn't add new BlueALSA PCM: %s", strerror(errno));
 				goto fail;
 			}
-			memcpy(&ba_pcms[ba_pcms_count++], &pcm, sizeof(*ba_pcms));
-			supervise_pcm_worker(&pcm);
+			supervise_io_worker(&pcm);
 			return DBUS_HANDLER_RESULT_HANDLED;
 		}
 
 		if (strcmp(signal, "InterfacesRemoved") == 0) {
 			if (!dbus_message_iter_init(message, &iter) ||
 					dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_OBJECT_PATH) {
-				error("Couldn't remove PCM: %s", "Invalid signal signature");
+				error("Couldn't remove BlueALSA PCM: %s", "Invalid signal signature");
 				goto fail;
 			}
 			dbus_message_iter_get_basic(&iter, &path);
 			struct ba_pcm *pcm;
-			if ((pcm = get_ba_pcm(path)) == NULL)
+			if ((pcm = ba_pcm_get(path)) == NULL)
 				goto fail;
-			supervise_pcm_worker_stop(pcm);
+			supervise_io_worker_stop(pcm);
+			ba_pcm_remove(path);
 			return DBUS_HANDLER_RESULT_HANDLED;
 		}
 
@@ -799,19 +953,19 @@ static DBusHandlerResult dbus_signal_handler(DBusConnection *conn, DBusMessage *
 
 	if (strcmp(interface, DBUS_INTERFACE_PROPERTIES) == 0) {
 		struct ba_pcm *pcm;
-		if ((pcm = get_ba_pcm(path)) == NULL)
+		if ((pcm = ba_pcm_get(path)) == NULL)
 			goto fail;
 		if (!dbus_message_iter_init(message, &iter) ||
 				dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_STRING) {
-			error("Couldn't update PCM: %s", "Invalid signal signature");
+			error("Couldn't update BlueALSA PCM: %s", "Invalid signal signature");
 			goto fail;
 		}
 		dbus_message_iter_get_basic(&iter, &interface);
 		dbus_message_iter_next(&iter);
 		if (!bluealsa_dbus_message_iter_get_pcm_props(&iter, NULL, pcm))
 			goto fail;
-		if ((worker = supervise_pcm_worker(pcm)) != NULL)
-			pcm_worker_mixer_volume_update(worker, pcm);
+		if ((worker = supervise_io_worker(pcm)) != NULL)
+			io_worker_mixer_volume_sync_snd_mixer_elem(worker, pcm);
 		return DBUS_HANDLER_RESULT_HANDLED;
 	}
 
@@ -822,10 +976,11 @@ fail:
 int main(int argc, char *argv[]) {
 
 	int opt;
-	const char *opts = "hVvlLB:D:M:";
+	const char *opts = "hVSvlLB:D:M:";
 	const struct option longopts[] = {
 		{ "help", no_argument, NULL, 'h' },
 		{ "version", no_argument, NULL, 'V' },
+		{ "syslog", no_argument, NULL, 'S' },
 		{ "verbose", no_argument, NULL, 'v' },
 		{ "list-devices", no_argument, NULL, 'l' },
 		{ "list-pcms", no_argument, NULL, 'L' },
@@ -842,6 +997,12 @@ int main(int argc, char *argv[]) {
 		{ 0, 0, 0, 0 },
 	};
 
+	bool syslog = false;
+
+	/* Check if syslog forwarding has been enabled. This check has to be
+	 * done before anything else, so we can log early stage warnings and
+	 * errors. */
+	opterr = 0;
 	while ((opt = getopt_long(argc, argv, opts, longopts, NULL)) != -1)
 		switch (opt) {
 		case 'h' /* --help */ :
@@ -850,6 +1011,7 @@ int main(int argc, char *argv[]) {
 					"\nOptions:\n"
 					"  -h, --help\t\t\tprint this help and exit\n"
 					"  -V, --version\t\t\tprint version and exit\n"
+					"  -S, --syslog\t\t\tsend output to syslog\n"
 					"  -v, --verbose\t\t\tmake output more verbose\n"
 					"  -l, --list-devices\t\tlist available BT audio devices\n"
 					"  -L, --list-pcms\t\tlist available BT audio PCMs\n"
@@ -875,9 +1037,27 @@ int main(int argc, char *argv[]) {
 			printf("%s\n", PACKAGE_VERSION);
 			return EXIT_SUCCESS;
 
+		case 'S' /* --syslog */ :
+			syslog = true;
+			break;
+
 		case 'v' /* --verbose */ :
 			verbose++;
 			break;
+		}
+
+	log_open(basename(argv[0]), syslog);
+	dbus_threads_init_default();
+
+	/* parse options */
+	optind = 0; opterr = 1;
+	while ((opt = getopt_long(argc, argv, opts, longopts, NULL)) != -1)
+		switch (opt) {
+		case 'h' /* --help */ :
+		case 'V' /* --version */ :
+		case 'S' /* --syslog */ :
+		case 'v' /* --verbose */ :
+			break;
 
 		case 'l' /* --list-devices */ :
 			list_bt_devices = true;
@@ -922,7 +1102,7 @@ int main(int argc, char *argv[]) {
 			break;
 
 		case 5 /* --single-audio */ :
-			pcm_mixer = false;
+			force_single_playback = true;
 			break;
 
 		default:
@@ -930,9 +1110,6 @@ int main(int argc, char *argv[]) {
 			return EXIT_FAILURE;
 		}
 
-	log_open(argv[0], false, false);
-	dbus_threads_init_default();
-
 	DBusError err = DBUS_ERROR_INIT;
 	if (!bluealsa_dbus_connection_ctx_init(&dbus_ctx, dbus_ba_service, &err)) {
 		error("Couldn't initialize D-Bus context: %s", err.message);
@@ -971,15 +1148,15 @@ int main(int argc, char *argv[]) {
 		for (i = 0; i < ba_addrs_count; i++, tmp += 19)
 			ba2str(&ba_addrs[i], stpcpy(tmp, ", "));
 
-		printf("Selected configuration:\n"
+		info("Selected configuration:\n"
 				"  BlueALSA service: %s\n"
-				"  PCM device: %s\n"
-				"  PCM buffer time: %u us\n"
-				"  PCM period time: %u us\n"
+				"  ALSA PCM device: %s\n"
+				"  ALSA PCM buffer time: %u us\n"
+				"  ALSA PCM period time: %u us\n"
 				"  ALSA mixer device: %s\n"
 				"  ALSA mixer element: '%s',%u\n"
 				"  Bluetooth device(s): %s\n"
-				"  Profile: %s\n",
+				"  Profile: %s",
 				dbus_ba_service,
 				pcm_device, pcm_buffer_time, pcm_period_time,
 				mixer_device, mixer_elem_name, mixer_elem_index,
@@ -1007,9 +1184,8 @@ int main(int argc, char *argv[]) {
 	if (!bluealsa_dbus_get_pcms(&dbus_ctx, &ba_pcms, &ba_pcms_count, &err))
 		warn("Couldn't get BlueALSA PCM list: %s", err.message);
 
-	size_t i;
-	for (i = 0; i < ba_pcms_count; i++)
-		supervise_pcm_worker(&ba_pcms[i]);
+	for (size_t i = 0; i < ba_pcms_count; i++)
+		supervise_io_worker(&ba_pcms[i]);
 
 	struct sigaction sigact = { .sa_handler = main_loop_stop };
 	sigaction(SIGTERM, &sigact, NULL);
@@ -1036,5 +1212,10 @@ int main(int argc, char *argv[]) {
 
 	}
 
+	for (size_t i = 0; i < workers_count; i++)
+		io_worker_stop(i);
+	free(workers);
+
+	bluealsa_dbus_connection_ctx_free(&dbus_ctx);
 	return EXIT_SUCCESS;
 }
